(** printing ⋈ %\ensuremath{\bowtie}% *)
(** printing ⋆ %\ensuremath{^{-1}}% *)
(** printing ≌ %\ensuremath{\cong}% *)
Require Import tools.
 
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

(** * Set of atoms. *)
(** Type of atoms *)

Parameter Atom : Set.

(** We require [Atom] to be decidable, by a boolean function. *)

Parameter beq_atom : Atom -> Atom -> bool.

Axiom beq_atom_refl : forall a b, reflect (a=b) (beq_atom a b).

(** The following trivial lemmas will be usefull in the rest *)
(** of the proofs to simplify statements. *)

Lemma beq_atom_spec : forall a b, beq_atom a b = true <-> a = b.
Proof.
  intros a b;split; destruct (beq_atom_refl a b);
  firstorder discriminate.
Qed.
Lemma Atom_eq_dec : forall x y : Atom, {x = y} + {x <> y}.
Proof. intros x y;destruct (beq_atom_refl x y);firstorder. Qed.
Ltac case_in x l :=
  let h := fresh "h" in case (in_dec Atom_eq_dec x l ) as [h|h].
Lemma beq_eq_atom a : beq_atom a a = true.
Proof. rewrite beq_atom_spec;auto. Qed.
Lemma beq_neq_atom1 a b : a <> b -> beq_atom a b = false.
Proof. destruct (beq_atom_refl a b); firstorder. Qed.
Lemma beq_neq_atom a b : a <> b <-> beq_atom a b = false.
Proof. destruct (beq_atom_refl a b); firstorder. Qed.
Hint Resolve Atom_eq_dec beq_atom_spec.
Ltac case_eq_at x y :=
  let h := fresh "h" in
  case (Atom_eq_dec x y ) as [h|h];[try rewrite h in *|].

(** Simplification of [if then else] statements. *)
Lemma if_Atom_eq_dec_eq {A} : forall a1 a2 (b c : A),
    a1 = a2 -> (if Atom_eq_dec a1 a2 then b else c) = b.
Proof. intros a1 a2 b c;case_eq_at a1 a2;tauto. Qed.

Lemma if_Atom_eq_dec_neq {A} : forall a1 a2 (b c : A),
    a1<>a2 -> (if Atom_eq_dec a1 a2 then b else c) = c.
Proof. intros a1 a2 b c;case_eq_at a1 a2;tauto. Qed.

(** Boolean equality predicate over lists. *)
Fixpoint beq_list l m :=
  match (l,m) with
  | ([],[]) => true
  | (a::l,b::m) => (beq_atom a b)&&(beq_list l m)
  | _ => false
  end.

Lemma beq_list_spec l m : reflect (l=m) (beq_list l m).
Proof.
  apply iff_reflect;revert m;induction l;intro m;
  destruct m;simpl;try simpl_bool;try firstorder congruence.
  rewrite <- (reflect_iff _ _ (beq_atom_refl _ _)),<-IHl.
  firstorder congruence.
Qed.
  
(** Boolean [In] predicate *)

Definition inb a l := (existsb (fun y => beq_atom y a) l).

Lemma inb_spec x l : reflect (In x l) (inb x l).
Proof.
  apply iff_reflect;induction l;simpl;try simpl_bool;
  firstorder auto with bool using beq_atom_spec.
Qed.
Ltac case_ex a l :=
  let h := fresh "h" in
  case (inb_spec a l) as [h|h].

Lemma in_double_neg : forall (x : Atom) l, ~~In x l -> In x l.
Proof. intros x l; case_in x l;tauto. Qed.

Ltac simpl_beq :=
  ( rewrite beq_eq_atom in * )
  || (match goal with
      | h:?a<>?b |- _ =>
        let hh := fresh "h" in
        assert (hh:b<>a) by (now firstorder);
          (rewrite (beq_neq_atom1 h) in * )
          || (rewrite (beq_neq_atom1 hh) in * );
          clear hh
      end)
  || (rewrite if_Atom_eq_dec_eq ;[|now firstorder])
  || (rewrite if_Atom_eq_dec_neq ;[|now firstorder])
  || (match goal with
      | h:_|-_ =>
        (rewrite if_Atom_eq_dec_eq in h;[|now firstorder])
        ||(rewrite if_Atom_eq_dec_neq in h;[|now firstorder])
      end).
Hint Resolve in_double_neg.
(** * Permutations *)

(** Permutations are defined as lists of pairs. This is general enough *)
(** as every permutation (on an infinite set) can be factored as a product *)
(** of cycles of length 2. *)

Definition perm := list (Atom * Atom).

(** Boolean equality predicate over permutations. *)

Fixpoint beq_perm l m :=
  match (l,m) with
  | ([],[]) => true
  | ((a,a')::l,(b,b')::m) =>
    (beq_atom a b)&&(beq_atom a' b')&&(beq_perm l m)
  | _ => false
  end.

Lemma beq_perm_spec l m : reflect (l=m) (beq_perm l m).
Proof.
  apply iff_reflect;revert m;induction l as [|(a&b) l IHl];intro m;
  destruct m as [|(x&y) m];simpl;repeat simpl_bool;
  try firstorder congruence.
  repeat rewrite <- (reflect_iff _ _ (beq_atom_refl _ _));
    repeat rewrite <-IHl.
  firstorder congruence.
Qed.

(** [act p a] computes the effect of the permutation [p] on the atom [a]. *)

Definition act (p : perm) a :=
  fold_right (fun t a =>
               match t with
                 | (x,y) => if beq_atom a x
                            then y
                            else if beq_atom a y
                                 then x
                                 else a
               end)
            a p.
Notation "p ⋈ x" := (act p x) (at level 20).

(** The effect of the empty permutation is the identity. *)

Lemma act_nil : forall a, [] ⋈ a = a.
Proof. now simpl. Qed.

(** List concatenation corresponds to functionnal composition of permutations. *)

Lemma act_comp_app (p q : perm) x : p ⋈ (q ⋈ x) = (p++q) ⋈ x.
Proof.
  unfold act;rewrite fold_right_app;auto.
Qed.

(** The inverse of a permutation is computed by reversing the list. *)

Definition inverse (p : perm) : perm :=
  rev p.

Notation "p ⋆" := (inverse p) (at level 35).
Hint Unfold inverse.

(** The inverse is an involutive operation. *)

Lemma inverse_inv (p : perm) : (p⋆)⋆ = p.
   unfold inverse;apply rev_involutive.
Qed.

(** Application of the inverse of a permutation is the functionnal left inverse of the application of the same permutation. *)

Lemma inverse_comp_l (p : perm) : forall x,  p⋆ ⋈ (p ⋈ x) = x.
Proof.
  intro x; induction p;simpl;auto.
  destruct a as (a&b).
  rewrite <- act_comp_app;simpl;
  case (Atom_eq_dec (p ⋈ x) a) as [ha|ha];
    [ rewrite ha in *;rewrite beq_eq_atom
    | apply beq_neq_atom in ha;repeat rewrite ha;
      case (Atom_eq_dec (p ⋈ x) b) as [hb|hb];
      [ rewrite hb in *;rewrite beq_eq_atom
      | apply beq_neq_atom in hb;repeat rewrite hb;repeat rewrite ha]];
    simpl; try rewrite beq_eq_atom;auto.
  case (Atom_eq_dec b a) as [->|hba];
    [rewrite beq_eq_atom|apply beq_neq_atom in hba;rewrite hba];auto.
Qed.

Ltac inv1 p x :=
  replace x with (p⋆⋈(p⋈x));[|apply inverse_comp_l].
(** Application of the inverse of a permutation is the functionnal right inverse of the application of the same permutation. *)

Lemma inverse_comp_r (p : perm) : forall x, p ⋈ (p⋆ ⋈ x) = x.
Proof.
  intro x;rewrite <- (inverse_inv p) at 1.
  apply inverse_comp_l.
Qed.
Ltac inv2 p x :=
  replace x with (p⋈(p⋆⋈x));[|apply inverse_comp_r].

(** The inverse of a composite permutation is the reverse composition of the inverses of its parts. *)

Lemma inverse_app p q : (p++q)⋆ = q⋆++p⋆.
Proof. unfold inverse;apply rev_app_distr;reflexivity. Qed.

Hint Resolve inverse_inv inverse_comp_l inverse_comp_r.
Ltac aux_perm :=
  (rewrite inverse_comp_l in *)
    || (rewrite inverse_comp_r in *)
    || (rewrite inverse_inv in *).
(** Permutations are injective. *)

Lemma perm_inj p x y : p ⋈ x = p ⋈ y -> x = y.
  intro h;induction p.
  simpl in *;auto.
  destruct a as (a&b).
  simpl in *;
    case (Atom_eq_dec (p ⋈ x) a) as [hxa|hxa];
    [ rewrite hxa in *;rewrite beq_eq_atom in *
    | apply beq_neq_atom in hxa;repeat rewrite hxa in *;
      case (Atom_eq_dec (p ⋈ x) b) as [hxb|hxb];
      [ rewrite hxb in *;rewrite beq_eq_atom in *
      | apply beq_neq_atom in hxb;repeat rewrite hxb in *;
        repeat rewrite hxa in *]];
    (case (Atom_eq_dec (p ⋈ y) a) as [hya|hya];
     [ rewrite hya in *;rewrite beq_eq_atom in *
     | apply beq_neq_atom in hya;repeat rewrite hya in *;
       case (Atom_eq_dec (p ⋈ y) b) as [hyb|hyb];
       [ rewrite hyb in *;rewrite beq_eq_atom in *
       | apply beq_neq_atom in hyb;repeat rewrite hyb in *;
         repeat rewrite hya in *]]);
    apply IHp;auto.
  rewrite h in hyb;rewrite beq_eq_atom in hyb;discriminate. 
  rewrite h in hya;rewrite beq_eq_atom in hya;discriminate. 
  rewrite h in hxb;rewrite beq_eq_atom in hxb;discriminate. 
  rewrite h in hxa;rewrite beq_eq_atom in hxa;discriminate. 
Qed.

(** Permutations are bijective. *)

Lemma perm_bij : forall p x y, p ⋈ x = p ⋈ y <-> x = y.
Proof.
  split;intro h.
  eapply perm_inj;eauto.
  rewrite h;tauto.
Qed.

(** Two permutations [p] and [q] are considered equivalent when the functions [p⋈_] and [q⋈_] are extensionnally equal. *)

Definition equiv_perm p q := forall a, p ⋈ a = q ⋈ a.
Notation " p ≌ q " := (equiv_perm p q) (at level 60).

(** The relation [≌] is an equivalence relation. *)

Lemma equiv_perm_refl p : p ≌ p.
Proof. intros a;reflexivity. Qed.
Lemma equiv_perm_sym p q : p ≌ q -> q ≌ p.
Proof. intros E a ;rewrite E;auto. Qed.
Lemma equiv_perm_trans p q r : p ≌ q -> q ≌ r -> p ≌ r.
Proof. intros E1 E2 a;rewrite E1,E2;auto. Qed.
Hint Resolve equiv_perm_refl equiv_perm_sym equiv_perm_trans.
Global Instance equiv_perm_Equivalence : Equivalence equiv_perm.
Proof.
  split.
  - intros p a;reflexivity.
  - intros p q E a ;rewrite E;auto.
  - intros p q r E1 E2 a;rewrite E1,E2;auto.
Qed.

(** If [p≌q] then [p⋆≌q⋆]. *)

Global Instance equiv_perm_inverse :
  Proper (equiv_perm ==> equiv_perm) inverse.
Proof. intros p q H a;inv1 p ((q⋆)⋈a); rewrite H,inverse_comp_r;auto.
Qed.

(** If [p≌q] and [p'≌q'] then [p++p'≌q++q']. *)

Global Instance equiv_perm_app :
  Proper (equiv_perm ==> equiv_perm ==> equiv_perm)
         (@app (Atom*Atom)%type).
Proof.
  intros p p' E q q' F a;repeat rewrite <- act_comp_app;
  rewrite E,F;auto.
Qed.

(** Paraphrasing of the definition of [≌]. *)

Global Instance equiv_perm_spec a :
  Proper (equiv_perm ==> Logic.eq) (fun p => p ⋈ a).
Proof. intros p q e; rewrite e ;reflexivity. Qed.

(** For every atom [a], the permutations [[a,a]] and [[]] are equivalent. *)

Lemma perm_a_a_eq_nil {a} : [a,a] ≌ [].
Proof. intro x;simpl;case_eq_at x a;simpl_beq;tauto. Qed.

(** For every permutation [p], [p⋆++p] and [p++p⋆] are also equivalent to [[]]. *)

Lemma perm_pinv_p_eq_nil {p} : (p⋆++p) ≌ [].
Proof. intro x; rewrite <-act_comp_app,inverse_comp_l;reflexivity. Qed.

Lemma perm_p_pinv_eq_nil {p} : (p++p⋆) ≌ [].
Proof. intro x; rewrite <-act_comp_app,inverse_comp_r;reflexivity. Qed.