(** printing -- %\ensuremath{\setminus}% *)
(** printing 〈 %\ensuremath{\langle}% *)
(** printing 〉 %\ensuremath{\rangle}% *)
(** printing ∩ %\ensuremath{\cap}% *)
(** printing ⧓ %\ensuremath{\underline\bowtie}% *)
Require Import tools atoms.
 
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

(** Lists of Atoms. *)

(** * Operator definitions *)
(** [l -- m] removes from the list [l] elements that appear *)
(** in [m]. *)

Definition purge l m := filter (fun a => negb (inb a m)) l.
Hint Unfold purge.

Notation "l -- m" := (purge l m) (at level 40).

(** [〈l〉] is a list containing the same elements as [l], but without *)
(** duplication. *)

Definition undup l :=
  fold_right (fun a acc =>
                if inb a acc
                then acc
                else a::acc)
             nil
             l.

Notation " 〈 l 〉 " := (undup l) (at level 0).

(** The intersection of two lists. *)

Definition inter l m := filter (fun a => inb a m) l.

Notation "l ∩ m" := (inter l m) (at level 40).

(** Removing an element from a list. *)

Definition rm := (@remove Atom Atom_eq_dec).
  
(** Applying a permutation to a list. *)

Definition mact p l :=map (act p) l.

Notation "p ⧓ l" := (mact p l) (at level 30).

(** If two permutations are equivalent, then they have the same effect on lists. *)

Global Instance equiv_perm_mact {a} :
  Proper (equiv_perm ==> Logic.eq) (fun p => p ⧓ a).
Proof. intros p q E;induction a;simpl;auto;rewrite E,IHa;auto. Qed.

(** * [length] of the operators. *)
Create HintDb length.
(** Filtering out elements of a list yields a smaller list. *)

Lemma filter_length {A} (f : A -> bool) l :
 length (filter f l) <= length l.
Proof. induction l;simpl;[|destruct (f a);simpl];lia. Qed.

(** [l--m], [l∩m], [rm a l], and [〈l〉] are all smaller than [l]. *)

Lemma purge_length l m : length (l--m) <= length l.
Proof. unfold purge;rewrite filter_length;auto. Qed.
Lemma inter_length l m : length (l ∩ m) <= length l.
Proof. unfold inter;rewrite filter_length;auto. Qed.
Lemma rm_length a l : length (rm a l) <= length l.
Proof. induction l;simpl;[|case_eq_at a a0;simpl];lia. Qed.
Lemma undup_length l : length 〈l〉 <= length l.
Proof. induction l;simpl; [|destruct (inb a _);simpl];lia. Qed.

(** Whatever permutation [p], the lists [p⧓l] and [l] always have the same length. *) 

Lemma mact_length p l : length (p⧓l) = length l.
Proof. unfold mact;rewrite map_length;auto. Qed.
Hint Resolve filter_length purge_length inter_length
     rm_length mact_length app_length undup_length : length.

(** * [In] predicate over operators *)

Lemma In_cons {A} : forall (a b : A) l,
    In b (a::l) <-> a=b \/ In b l.
Proof. simpl;tauto. Qed.

Lemma In_app_iff {A} : forall (a : A) l m,
    In a (l ++ m) <-> In a l \/ In a m.
Proof.
  intros a l m;induction l;split;simpl;intro h;auto.
  case h;auto.
  intro;contradiction.
  rewrite IHl in h;case h as [h|[h|h]];auto.
  rewrite IHl;case h as [[h|h]|h];auto.
Qed.

Lemma In_purge :forall a l m,
    In a (l -- m) <-> In a l /\ ~ In a m.
Proof.
  intros a l m; unfold purge; rewrite filter_In.
  rewrite (reflect_iff _ _ (inb_spec a m));
  destruct (inb _ _);firstorder.
Qed.

Lemma In_undup : forall a l,
    In a 〈l〉 <-> In a l.
Proof.
  intros a l;revert a; induction l;simpl;auto;try tauto.
  intro b;destruct (inb_spec a 〈l〉);
  rewrite IHl in *;case_eq_at a b;firstorder.
Qed.

Lemma In_inter : forall l m a,
  In a (l ∩ m) <-> In a l /\ In a m.
Proof.
  intros l m a;unfold inter;
  rewrite filter_In,<-(reflect_iff _ _ (inb_spec _ _));tauto. 
Qed.

Lemma In_rm : forall a b l,
  In a (rm b l)<-> In a l /\ b <> a.
Proof.
  intros a b l;unfold rm;induction l;simpl;try tauto.
  destruct Atom_eq_dec;simpl;rewrite IHl;firstorder congruence.
Qed.

Lemma In_mact : forall a p l,
    In a (p ⧓ l) <-> In (p⋆ ⋈ a) l.
Proof.
  intros a p l; unfold mact;rewrite in_map_iff;split;
  [intros (y & <- & h)|intro h; eexists;split];
  now eauto;aux_perm.
Qed.

Ltac simpl_in :=
  repeat 
    ((rewrite In_mact in * )
     || (rewrite In_rm in * )
     || (rewrite In_cons in * )
     || (rewrite in_app_iff in * )
     || (rewrite In_purge in * )
     || (rewrite In_inter in * )
     || (rewrite In_undup in * )).
(** * [NoDup] predicate over operators *)
(** The [〈〉] operator always produces a duplication free list. *)

Lemma NoDup_undup l : NoDup 〈l〉.
Proof.
  induction l;simpl;auto using NoDup_nil.
  case (inb_spec);auto using NoDup_cons.
Qed.

(** The list [a::l] is duplication free if and only if [a] doesn't appear in [l] and [l] is itself duplication free. *)

Lemma NoDup_cons_iff {A : Type} (a : A) l :
  NoDup (a::l) <-> ~ In a l /\ NoDup l.
Proof.
  split.
  replace l with (nil++l);auto.
  now split;[apply NoDup_remove_2|apply NoDup_remove_1 with (a:=a)].
  now intros (I&h);apply NoDup_cons.
Qed.  

(** The functions [filter f], [_ -- m], [_ ∩ m], and [rm a] preserve the NoDup predicate. *)

Lemma NoDup_filter {A} f (l : list A) : NoDup l -> NoDup (filter f l).
Proof.
  induction l.
  simpl;auto.
  intro h;inversion h.
  simpl.
  case (f a);auto.
  apply NoDup_cons.
  rewrite filter_In.
  intro h1;apply H1;apply h1.
  tauto.
Qed.

Lemma NoDup_purge l m : NoDup l -> NoDup (l -- m).
Proof.
  intro h;apply NoDup_filter;auto.
Qed.  

Lemma NoDup_inter l m: NoDup l -> NoDup (l ∩ m).
Proof.
  intro h;unfold inter; now apply NoDup_filter.
Qed.

Lemma NoDup_rm a l : NoDup l -> NoDup (rm a l).
Proof.
  unfold rm;induction l;simpl;auto;destruct Atom_eq_dec;
  repeat rewrite NoDup_cons_iff;firstorder.
  now rewrite In_rm.
Qed.

(** If [l] and [m] don't share any element, then if both of them don't have duplications their concatenation won't either. *)

Lemma NoDup_app (l m : list Atom):
  (forall a, In a l -> ~ In a m) -> NoDup l ->  NoDup m -> NoDup (l++m).
Proof.
  intros h1 h2 h3;revert h2 m h3 h1; induction l.
  - simpl;intros;auto.
  - intros hl m hm h; inversion hl.
    simpl;apply NoDup_cons;auto.
    rewrite In_app_iff;firstorder.
Qed.

(** If the concatenation of two lists is duplication free, then so are they. *)

Lemma NoDup_app_inv {A} : forall (l m : list A),
    NoDup (l++m) -> NoDup l /\ NoDup m.
Proof.
  induction l;simpl;intros m nd.
  - split;auto using NoDup_nil.
  - apply NoDup_cons_iff in nd;simpl_in;split;
    [apply NoDup_cons|];firstorder.
Qed.

(** Applying a permutation to a list doesn't change its freeness from duplications. *)

Lemma NoDup_mact p l : NoDup l <-> NoDup (p ⧓ l).
Proof.
  induction l;simpl;split;intro h;auto;
  apply NoDup_cons_iff in h as (h1 & h2);
  now apply NoDup_cons;[rewrite In_mact in *;aux_perm
                       |rewrite<- IHl in *].
Qed.

(** The [〈〉] operator doesn't change duplication free lists. *)

Lemma NoDup_undup_eq l : NoDup l -> l = 〈l〉.
Proof.
  induction l;simpl;auto.
  intro h;apply NoDup_cons_iff in h as (I&h).
  case inb_spec.
  rewrite In_undup at 1;tauto.
  intuition;congruence.
Qed.

(** * Simple properties of the list operators. *)

Lemma filter_id {A} f (l : list A) :
  (forall a, In a l -> f a = true) -> filter f l = l.
Proof.
  intro h;induction l;auto.
  simpl;rewrite h;auto.
  rewrite IHl;auto.
Qed.

Lemma rm_comm : forall a b l, rm a (rm b l) = rm b (rm a l).
Proof.
  intros a b l;induction l;simpl;auto.
  repeat (destruct Atom_eq_dec;simpl);congruence.
Qed.

Lemma purge_nil l : l -- [] = l.
Proof.  unfold purge;rewrite filter_id;auto. Qed.

Lemma purge_cons a l m : l -- (a::m) = (rm a l) -- m.
Proof.
  induction l;simpl;auto.
  case_eq_at a a0;simpl_beq;simpl;auto.
  now rewrite<- IHl.
Qed.

Lemma rm_app a l m : rm a (l++m) = rm a l ++ rm a m.
Proof.
  induction l;simpl;auto.
  destruct Atom_eq_dec;simpl;congruence.
Qed.

Lemma rm_purge a l m : (rm a l) -- m = rm a (l--m).
Proof.
  unfold rm;induction l;simpl in *;auto.
  case_eq_at a a0;simpl.
  - rewrite <-h in *;destruct (inb_spec a m);simpl;try simpl_beq;auto.
  - destruct (inb_spec a0 m);simpl;try simpl_beq;simpl;f_equal;auto.
Qed.
Lemma purge_l_l : (forall l, l--l = []).
Proof.
  induction l;simpl;auto;simpl_beq;simpl.
  rewrite purge_cons,rm_purge,IHl;reflexivity.
Qed.
Hint Resolve purge_nil purge_l_l.

Lemma rm_id a l : ~ In a l -> rm a l = l.
Proof.
  induction l;simpl;auto;destruct Atom_eq_dec;firstorder congruence.
Qed.

Lemma rm_cons a l : ~ In a l -> rm a (a::l) = l.
Proof.
  induction l;simpl;auto;repeat (destruct Atom_eq_dec;simpl in *);
                                 firstorder congruence.
Qed.
Hint Resolve rm_id rm_cons.

Lemma filter_app {A} : forall f (l m : list A),
    filter f (l ++ m) = filter f l ++ filter f m.
Proof.
  intros f;induction l;intros m;simpl;auto.
  destruct (f a);rewrite IHl;simpl;auto.
Qed.

Lemma purge_app: forall l m n, (l ++ m) -- n = (l -- n) ++ (m -- n).
Proof. intros;unfold purge;rewrite filter_app;auto. Qed.

Lemma disjoint_In : forall l m,
    l ∩ m = [] <-> (forall a, ~ (In a l /\ In a m)).
Proof.
  intros l m;split;[intros C x (I1 & I2)|intro I].
  cut (In x []);[simpl|rewrite<- C;simpl_in];tauto.
  cut (forall x, ~In x (l∩m));[| intro;simpl_in;apply I].
  clear I;destruct (l∩m);intro I;auto.
  cut (In a (a::l0));[|simpl];firstorder.
Qed.

Lemma disjoint_com : forall l m, l ∩ m=[] <-> m ∩ l=[].
Proof. intros;repeat rewrite disjoint_In;firstorder. Qed.

Lemma disjoint_purge : forall l m, l ∩ m = [] -> l -- m = l.
Proof.
  intros l m;rewrite disjoint_In;intro C;apply filter_id.
  intros x I;apply Bool.eq_true_not_negb; case inb_spec;firstorder.
Qed.

Lemma purge_rm l m a : ~ In a l -> l -- (rm a m) = l -- m.
Proof.
  intro h;induction l;simpl;auto.
  rewrite IHl;simpl in *;auto.
  case_eq_at a a0;[tauto|].
  destruct (inb_spec a0 m);
    destruct (inb_spec a0 (rm a m));simpl_in;try tauto;simpl.
Qed.

Lemma undup_purge l m : 〈l -- m〉 = 〈l〉 -- m.
Proof.
  revert m;induction l;firstorder;simpl.
  case_ex a m;simpl.
  - case_ex a 〈l〉;simpl;auto.
    case_ex a m;simpl;firstorder congruence.
  - case_ex a 〈l〉;simpl;auto.
    * case_ex a 〈l--m〉;simpl;simpl_in;
       firstorder congruence.
    * case_ex a m;simpl;case_ex a 〈l--m〉;simpl;simpl_in;
       firstorder congruence.
Qed.

Lemma undup_app l m : 〈l ++ m〉 = 〈l -- m〉 ++ 〈m〉.
Proof.
  revert m;induction l;firstorder; simpl.
  case_ex a m;simpl;auto.
  - case_ex a 〈l++m〉;simpl;simpl_in;firstorder.
  - case_ex a 〈l++m〉;simpl;simpl_in;
    case_ex a 〈l--m〉;simpl;simpl_in;firstorder congruence.
Qed.

Lemma undup_rm l a : 〈rm a l〉 = rm a 〈l〉.
Proof.
  induction l;firstorder.
  simpl.
  case_eq_at a a0;case_ex a0 〈l〉.
  firstorder congruence.
  simpl;simpl_beq; firstorder congruence.
  simpl;case_ex a0 〈rm a l〉;simpl_in;firstorder congruence.
  simpl_in;simpl; simpl_beq.
  simpl;case_ex a0 〈rm a l〉;simpl_in;firstorder congruence.
Qed.

(** * Simple properties of [⧓]. *)

Lemma mact_comp_app p q l : p ⧓ (q ⧓ l) = (p ++ q) ⧓ l.
  etransitivity;[apply map_map|apply map_ext];
  intro x;apply act_comp_app.
Qed.

Lemma mact_inv_r p l : p ⧓ (p⋆  ⧓  l) = l.
  etransitivity;[apply mact_comp_app
                |etransitivity;
                  [eapply map_ext
                  |apply map_id]].
  intro x;rewrite <-act_comp_app;aux_perm;auto.
Qed.

Lemma mact_inv_l p l : p⋆ ⧓ (p ⧓ l) = l.
Proof.
  rewrite<- (mact_inv_r (p⋆) l) at 2;repeat f_equal.
  now rewrite inverse_inv.
Qed.
Hint Resolve mact_inv_r mact_inv_l.

Lemma mact_rm p a l: rm a (p ⧓ l) = p ⧓ (rm (p⋆ ⋈ a) l).
Proof.
  induction l;simpl;auto.
  repeat (destruct Atom_eq_dec;simpl in * );try congruence.
  rewrite e,inverse_comp_l in n;tauto.
  rewrite <-e,inverse_comp_r in n;tauto.
Qed.

Lemma mact_purge p m l: (p ⧓ l) -- m = p ⧓ (l -- (p⋆ ⧓ m)).
  induction l;simpl;auto.
  case_ex a (p⋆⧓m);simpl;
  rewrite In_mact in h;aux_perm;
  case_ex (p⋈a) m;simpl;try tauto.
  now f_equal.
Qed.  

Lemma mact_id2 a b l : ~ In a l -> ~ In b l -> [a,b] ⧓ l = l.
  induction l;intros h1 h2;simpl;auto.
  case (Atom_eq_dec a0 a) as [->|h];simpl in *;try tauto.
  rewrite beq_neq_atom in h;rewrite h;f_equal;auto.
  case (Atom_eq_dec a0 b) as [->|hh];simpl in *;try tauto.
  now rewrite beq_neq_atom in hh;rewrite hh.
Qed.
Hint Resolve mact_id2.

Lemma mact_nil : forall l, [] ⧓ l = l.
Proof. now induction l;simpl;[|rewrite IHl]. Qed.
Hint Resolve mact_nil act_nil.

Lemma mact_id1 a l : [a,a] ⧓ l = l.
Proof.
  rewrite (equiv_perm_mact _ _ perm_a_a_eq_nil);auto.
Qed.
Hint Resolve mact_id1.
Ltac simpl_perm :=
  repeat
    (rewrite inverse_comp_l in * )
  || (rewrite inverse_comp_r in * )
  || (rewrite inverse_inv in * )
  || (rewrite inverse_app in * )
  || (rewrite mact_id1 in * )
  || (rewrite mact_nil in * )
  || (rewrite act_nil in * )
  || (rewrite mact_inv_r in * )
  || (rewrite mact_inv_l in * ).

Lemma mact_app_distr : forall p l m, p ⧓ (l ++ m) = p ⧓ l ++ p ⧓ m.
Proof.
  intros p l; revert p;induction l;firstorder (simpl;congruence).
Qed.

Lemma mact_undup p l : p ⧓ 〈l〉 = 〈p ⧓ l〉.
Proof.
  revert p; induction l;intros;simpl;auto.
  case_ex a 〈l〉; case_ex (p⋈a) 〈p⧓l〉;simpl_in;simpl_perm;
  firstorder(simpl;congruence).
Qed.
