(** printing ≈ %\ensuremath{\approx}% *)
(** printing ⊆ %\ensuremath{\subseteq}% *)
Require Import tools atoms atoms_ops.
 
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.
(** Equivalence of lists *)
(** * Definitions *)
Definition equiv_lst l m := forall (x : Atom), In x l <-> In x m.
Notation "l ≈ m" := (equiv_lst l m)(at level 60).

Global Instance equiv_lst_refl : Reflexive equiv_lst.
Proof.  intros l x;tauto. Qed.
Global Instance equiv_lst_sym : Symmetric equiv_lst.
Proof. intros l m h x;rewrite (h x);tauto. Qed.
Global Instance equiv_lst_trans : Transitive equiv_lst.
Proof. intros l m n h1 h2 x;rewrite (h1 x);rewrite (h2 x);tauto. Qed.
Hint Resolve  equiv_lst_refl equiv_lst_sym equiv_lst_trans.
Global Instance equiv_lst_Equiv : Equivalence equiv_lst.
Proof. split;eauto. Qed.
Hint Resolve  equiv_lst_Equiv.

Definition incl_lst l m := forall (x : Atom), In x l -> In x m.

Global Instance pre_incl_lst : PreOrder incl_lst.
Proof.
  split.
  intros l x;tauto.
  intros l m n h1 h2 x h3;apply (h2 x);apply (h1 x);tauto.
Qed.
  
Global Instance anti_incl_lst : PartialOrder equiv_lst incl_lst.
Proof.
  split.
  intro h; split;intro y;apply (h y).
  intros (h1 & h2); split;intro h;
  [apply (h1 x1)|apply (h2 x1)];auto.
Qed.
  
Global Instance antisym_incl_lst : Antisymmetric (list Atom)
                                                 equiv_lst incl_lst.
Proof.
  intros x y h1 h2 a; split;intro h;
  [apply (h1 a)|apply (h2 a)];auto.
Qed.
Notation "l ⊆ m" := (incl_lst l m)(at level 60).
Global Instance In_equiv a :
  Proper (equiv_lst ==> Basics.impl) (In a).
Proof. intros l m h1 h2. now apply h1. Qed.
Global Instance In_incl a :
  Proper (incl_lst ==> Basics.impl) (In a).
Proof. intros l m h1 h2. now apply h1. Qed.

Definition incl_lst_b l m :=
  forallb (fun x => inb x m) l.
Lemma incl_lst_b_spec l m : reflect (l ⊆ m) (incl_lst_b l m).
Proof.
  unfold incl_lst; apply iff_reflect;induction l;simpl;try simpl_bool.
  - intuition.
  - rewrite <- (reflect_iff _ _ (inb_spec _ _));firstorder congruence.
Qed.

Definition equiv_lst_b l m := (incl_lst_b l m && incl_lst_b m l)%bool.
Lemma equiv_lst_b_spec l m : reflect (l ≈ m) (equiv_lst_b l m).
Proof.
  unfold equiv_lst_b; apply iff_reflect;simpl;simpl_bool.
  repeat rewrite <-(reflect_iff _ _ (incl_lst_b_spec _ _)).
  pose proof (anti_incl_lst l m); auto.
Qed.


(** * Simple Properties *)

Lemma cons_comm a b l m :
  l ≈ m -> (a :: b :: l) ≈ (b :: a :: m).
Proof. unfold equiv_lst; firstorder. Qed.
Lemma purge_cong l1 l2 m1 m2 :
  l1 ≈ l2 -> m1 ≈ m2 -> l1 -- m1 ≈ l2 -- m2.
Proof.
  intros h1 h2 x;repeat rewrite In_purge;
  rewrite (h1 x);rewrite (h2 x);tauto.
Qed.
Lemma app_cong l1 l2 m1 m2 :
  l1 ≈ l2 -> m1 ≈ m2 -> (l1 ++ m1) ≈ (l2 ++ m2).
Proof.
  intros h1 h2 x;repeat rewrite in_app_iff;
  rewrite (h1 x);rewrite (h2 x);tauto.
Qed.
Lemma undup_equiv_lst l : 〈l〉 ≈ l.
Proof. intro x;simpl_in;tauto. Qed.
Lemma inter_cong l1 l2 m1 m2 :
  l1 ≈ l2 -> m1 ≈ m2 -> l1 ∩ m1 ≈ l2 ∩ m2.
Proof.
  intros h1 h2 x;repeat rewrite In_inter;
  rewrite (h1 x);rewrite (h2 x);tauto.
Qed.

Lemma cons_cong a l m : l≈m -> (a::l) ≈ (a::m).
Proof. now intros h x;simpl;rewrite (h x). Qed.

Lemma rm_cong a l m : l≈m -> (rm a l) ≈ (rm a m).
Proof. now intros h x;repeat rewrite In_rm;simpl;rewrite (h x). Qed.
Lemma mact_cong p l m : p ⧓ l ≈ p ⧓ m <-> l ≈ m.
Proof.
  unfold equiv_lst;repeat setoid_rewrite In_mact;
  split;intros h x;[inv1 p x|];apply h.
Qed.

(** * The operators are morphims with respect to ≈ and ⊆. *)
Global Instance rm_equiv a : Proper (equiv_lst ==> equiv_lst) (rm a).
Proof. now intros l m h;apply rm_cong. Qed.
Global Instance cons_equiv a :
  Proper (equiv_lst ==> equiv_lst) (cons a).
Proof. now intros l m h;apply cons_cong. Qed. 
Global Instance purge_equiv :
  Proper (equiv_lst ==> equiv_lst ==> equiv_lst) purge.
Proof. now intros l m l' m' h h';apply purge_cong. Qed. 
Global Instance purge_eq :
  Proper (Logic.eq ==> equiv_lst ==> Logic.eq) purge.
Proof.
  intros l ? -> m m' h.
  induction y;simpl;auto;pose proof (h a).
  case_ex a m;simpl;case_ex a m';simpl;firstorder congruence.
Qed. 
Global Instance inter_equiv :
  Proper (equiv_lst ==> equiv_lst ==> equiv_lst) inter.
Proof. now intros l m l' m' h h';apply inter_cong. Qed. 
Global Instance rm_incl a : Proper (incl_lst ==> incl_lst) (rm a).
Proof. unfold incl_lst;intros l m h x h';simpl_in;firstorder. Qed.
Global Instance cons_incl a : Proper (incl_lst ==> incl_lst) (cons a).
Proof. unfold incl_lst;intros l m h x h';simpl_in;firstorder. Qed.
Global Instance purge_incl :
  Proper (incl_lst ==> equiv_lst ==> incl_lst) purge.
Proof.
  unfold incl_lst;intros l m h l' m' h' x h'';simpl_in ;firstorder.
Qed.
Global Instance inter_incl :
  Proper (incl_lst ==> incl_lst ==> incl_lst) inter.
Proof.
  unfold incl_lst;intros l m h l' m' h' x h'';simpl_in ;firstorder.
Qed.
Global Instance mact_equiv  p :
  Proper (equiv_lst ==> equiv_lst) (mact p).
Proof. now intros l m h;apply mact_cong. Qed.
Global Instance mact_incl  p : Proper (incl_lst ==> incl_lst) (mact p).
Proof. unfold incl_lst; intros l m h x h'; simpl_in;firstorder. Qed.
Hint Resolve mact_equiv mact_incl.
Global Instance undup_equiv : Proper (equiv_lst ==> equiv_lst) undup. 
Proof. now intros l m h; repeat rewrite undup_equiv_lst. Qed.
Global Instance app_equiv :
  Proper (equiv_lst ==> equiv_lst ==> equiv_lst) (@app Atom).
Proof. intros m n E l k F x;simpl_in;rewrite (E x),(F x);tauto. Qed.
Global Instance app_incl :
  Proper (incl_lst ==> incl_lst ==> incl_lst) (@app Atom).
Proof.
  unfold Proper,respectful,incl_lst;setoid_rewrite In_app_iff;
  firstorder.
Qed.


(** * More Properties. *)
Lemma cons_rm_eq a l m : ~ In a m -> l ≈ (a::m) -> In a l /\ rm a l ≈ m.
Proof.
  intros h1 h2;unfold equiv_lst;setoid_rewrite In_rm;split.
  now apply (h2 a);left.
  intros x;split;[intros (h3 & h4)|intro h3;split].
  apply h2 in h3 as [->|h3];tauto.
  now apply h2;right.
  now intros ->.
Qed.  
  
Lemma rm_cons_eq a l m : In a l -> rm a l ≈ m -> ~ In a m /\ l ≈ (a::m).
Proof.
  unfold equiv_lst;setoid_rewrite In_rm;
  intuition.
  apply H0 in H1 as (h1 & h2);tauto.
  case_eq_at a x; try simpl_beq;simpl;[|rewrite<-H0];tauto.
  now case H1 as [->|h];[|apply H0 in h as (h&h')].
Qed.

Lemma rm_cons_eq' a l : In a l -> l ≈ (a :: rm a l).
Proof. intros; apply rm_cons_eq; auto. Qed.
Lemma cons_rm_eq' a m : ~ In a m -> rm a (a::m) ≈ m.
Proof. intros; apply cons_rm_eq; auto. Qed.
Hint Resolve rm_cons_eq' cons_rm_eq'.
Lemma purge_l_m_nil l m : l ⊆ m -> l -- m = [].
Proof.
  revert m;induction l;intros m h;simpl;auto;simpl;rewrite <- (IHl m).
  case_ex a m;simpl;auto;exfalso;apply h0;apply h;left;auto.
  intros x h0;apply h;right;auto.
Qed.
Lemma app_l_l l : (l++l) ≈ l.
Proof. intro x;simpl_in;tauto. Qed.

Hint Resolve app_l_l.
Lemma map_act_inv_eq p l m : p ⧓ l ≈ m <-> l ≈ p⋆ ⧓ m.
Proof.
  unfold equiv_lst;repeat setoid_rewrite In_mact;
  now split;intros h x;[inv1 p x|]; rewrite h;aux_perm.
Qed.

Lemma map1 p l m : l ⊆ (p⋆) ⧓ m <-> p ⧓ l ⊆ m.
  unfold incl_lst;repeat setoid_rewrite In_mact;aux_perm;
  split;intros h x h'.
  now inv2 p x;apply h.
  now apply h;aux_perm.
Qed.

Lemma map2 p l m : l ⊆ p ⧓ m <-> (p⋆) ⧓ l ⊆ m.
  unfold incl_lst;repeat setoid_rewrite In_mact.
  now split;intros h x h'; [apply h in h'|apply h];aux_perm.
Qed.
Lemma app_com l m : (l ++ m) ≈ (m ++ l).
Proof. intro;simpl_in;tauto. Qed.

Hint Resolve rm_equiv cons_equiv purge_equiv inter_equiv.


(** To be sorted. *)

Fixpoint elements (p : perm) :=
  match p with
  | [] => []
  | (a,b)::p => a::b::(elements p)
  end.
Lemma elements_inv_act x p :
  ~ In x (elements p) -> p⋈x = x.
Proof.
  revert x;induction p;simpl;auto.
  destruct a;intros.
  assert (~ In x (elements p)) as I by
        (now intro;apply H;right;right).
  apply IHp in I;rewrite I.
  case_eq_at x a;simpl_beq.
  now exfalso;apply H;left.
  case_eq_at x a0;simpl_beq.
  now exfalso;apply H;right;left.
  reflexivity.
Qed.
Lemma elements_inv_mact x p :
  (forall y, In y x -> ~ In y (elements p)) -> p⧓x = x.
Proof.
  revert p;induction x;simpl;intros;auto.
  rewrite IHx by (now intros; apply H;right).
  now f_equal;apply elements_inv_act,H;left.
Qed.
Lemma elements_app : forall p q,
    elements (p++q) ≈ (elements p ++ elements q).
Proof.
  induction p;destruct q;simpl;auto;destruct a;simpl.
  now repeat rewrite app_nil_r.
  now destruct p0;rewrite IHp;simpl.
Qed.
Lemma elements_inv_eq :forall q, elements q ≈ elements (q⋆).
Proof.
  induction q;simpl;auto.
  destruct a;rewrite elements_app,IHq;simpl;intro x;simpl_in;firstorder.
Qed.

Lemma elements_act_in_equiv : forall q a,
    In a (elements q) <-> In (q⋈a) (elements q).
Proof.
  induction q;simpl;intuition.
  - case_in a (elements q).
    * case_eq_at (q⋈a) a0;simpl_beq;auto.
      case_eq_at (q⋈a) b;simpl_beq;auto.
      now right;right;rewrite <-IHq.
    * pose proof (elements_inv_act h) as ->.
      case H as [->|[->|I]];auto.
      simpl_beq;auto.
      case_eq_at a a0;repeat simpl_beq;auto.
  - case_in a (elements q).
    * case_eq_at (q⋈a) a0;simpl_beq;auto.
    * simpl in *; pose proof (elements_inv_act h) as E;rewrite E in *.
      case_eq_at a a0;simpl_beq;auto.
      case_eq_at a b;simpl_beq;auto.
Qed.    
    
Lemma elements_in_mact : forall l y q,
    In y (q ⧓ l) -> In y l \/ In y (elements q).
Proof.
  induction l;intros;simpl in *;auto.
  case H as [<-|H].
  case_in a (elements q);auto.
  apply elements_act_in_equiv in h;auto.
  rewrite elements_inv_act;auto.
  firstorder.
Qed.

Lemma mact_elements_equiv_perm p q :
  p ≌ q <->
  (forall x, In x (elements p ++ elements q) -> p⋈x = q⋈x).
Proof.
  split.
  * now intros E x _;rewrite E.
  * intros h x.
    case_in x (elements p ++ elements q);intuition.
    rewrite In_app_iff in h0;repeat rewrite elements_inv_act;
    intuition.
Qed.

Definition bequiv_perm p q :=
  forallb (fun x => beq_atom (p⋈x) (q⋈x)) (elements p ++ elements q).
Lemma bequiv_perm_spec p q : reflect (p ≌ q) (bequiv_perm p q).
Proof.
  apply iff_reflect; unfold bequiv_perm;
  rewrite (mact_elements_equiv_perm p q), forallb_forall;
  setoid_rewrite beq_atom_spec;intuition.
Qed.