(** printing ⊥ %\ensuremath{\bot}% *)
(** printing · %\ensuremath{\cdot}% *)
(** printing ^* %\ensuremath{^\star}% *)
(** printing ‹ %\ensuremath{\lgroup}% *)
(** printing › %\ensuremath{\rgroup}% *)
(** printing θ %\ensuremath{\theta}% *)
(** printing ν %\ensuremath{\nu}% *)
(** printing ω %\ensuremath{\omega}% *)
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

Require Export tools atoms atoms_ops equiv_lst letter.

(** * Definitions, notations and simplification tactics. *)
(** Type of expressions *)
Inductive expr : Type :=
| bot : (list Atom) -> expr
| id : (list Atom) -> expr
| zero : expr
| un : expr
| var : Letter -> expr
| vat : Atom -> expr
| plus : expr -> expr -> expr
| times : expr -> expr -> expr
| star : expr -> expr
| ν : Atom -> expr -> expr
| ω : Atom -> expr -> expr
| θ : perm -> expr -> expr.
Notation " ⊥ A " := (bot A) (at level 20).
Notation " 0 " := zero (at level 20).
Notation " 1 " := un (at level 20).
Notation "x + y" := (plus x y) (left associativity, at level 50).
Notation "x · y" := (times x y) (left associativity, at level 40).
Notation "x ^*" := (star x) (at level 20).
Notation " ‹ p › x" := (θ p x) (at level 20).

(** The size of an expression is its number
 of constructors. *)
Fixpoint size e : nat :=
  match e with
    | 0 | 1 | var _ | vat _ | bot _ | id _  => (S O)%nat
    | e + f | e · f => S (size e + size f)
    | e ^* | ν _ e | ‹ _ › e | ω _ e => S (size e)
  end.
Hint Unfold size.

Lemma size_non_zero : forall e, O < size e.
Proof. destruct e;simpl;lia. Qed.
  
(** It is convenient to write weakennings in bulk. *)
Definition W l e := fold_right (fun a e => ω a e) e l.
(** Compositions of [W] may be factored with list concatenation. *)
Lemma W_app : forall l1 l2 e, W l1 (W l2 e) = W (l1++l2)e.
Proof.
  unfold W;intros;rewrite <-fold_right_app;auto.
Qed.
(** The size of [W l e] is the sum of the size of [e] and *)
(** the length of [l]. *)
Lemma size_W : forall e l, size (W l e) = (length l + size e)%nat.
Proof. induction l;simpl;intuition. Qed.

(** * The support of an expression computes the set of
       its free variables. *)
Fixpoint support e :=
  match e with
  | 0 | 1 => []
  | bot _ | id _ => []
  | var x => τ x
  | vat a => [a]
  | e + f | e · f => (support e) ++ (support f)
  | e ^* => (support e)
  | ‹ p › e => p ⧓ (support e)
  | ν a e => rm a (support e)
  | ω a e => (support e)
  end.
Definition Bs A e f := (support e ≈ support f) /\ A e f.
Hint Unfold support Bs.

Section poly.
  Parameter K : nat.
  Hypothesis K_not_zero : O < K.
  Hypothesis bounded_τ : forall x, length (τ x) <= K.
  Lemma length_support_size e : length (support e) <= (K * size e)%nat.
  Proof.
    induction e;simpl;auto with length;try lia.
    - rewrite bounded_τ;lia.
    - rewrite app_length;lia.
    - rewrite app_length;lia.
    - transitivity (length (support e));auto with length;try lia.
    - rewrite mact_length;lia.
  Qed.
End poly.

(** * The stage of an expression computes its sheaf. *)
Fixpoint stage e :=
  match e with
  | 0 | 1 => []
  | bot A | id A => A
  | var x => τ x
  | vat a => [a]
  | e + f | e · f => (stage e) ++ (stage f)
  | e ^* => (stage e)
  | ‹p› e => p ⧓ (stage e)
  | ν a e => rm a (stage e)
  | ω a e => a::(stage e)
  end.
Hint Unfold stage.


(** * [eq0] tests if an expression is equivalent to [0] or [⊥ _]. *)
Fixpoint eq0 e : bool :=
  match e with
    | 0 | bot _ => true
    | _^* | 1 | id _ | var _ | vat _ => false
    | e + f => eq0 e && eq0 f
    | e · f => eq0 e || eq0 f
    | ν _ e | ‹_› e | ω _ e => eq0 e
  end.
Hint Unfold eq0.
(** * [eq0'] tests if an expression is equal to [0] or [⊥ _]. *)
Definition eq0' e : bool :=
  match e with
  | 0 | bot _ => true
  | _ => false
  end.
Hint Unfold eq0'.
Lemma eq0'_spec e : reflect (e = 0 \/ (exists B, e = ⊥ B)) (eq0' e).
Proof.
  destruct e;simpl;constructor; firstorder congruence
                                || (right;eexists;reflexivity).
Qed.

Definition get_bot e :=
  match e with
  | ⊥ A => inr A
  | 0 => inl false
  | _ => inl true
  end.
Hint Unfold get_bot.

Lemma get_bot_spec e : reflect3 (e = 0) (fun B => e = ⊥ B)
                                (eq0' e = false) (get_bot e).
Proof.
  destruct e;simpl;constructor; firstorder congruence.
Qed.

Lemma eq0'_get_bot_positive e :
  ~ (e = 0 \/ (exists B : list Atom, e = ⊥ B)) ->
  eq0' e = false /\ get_bot e = inl true.
Proof.
  case eq0'_spec; [intros [->|(? & ->)]|intros _ ];simpl;auto.
  intros h;exfalso;apply h;left;auto.
  intros h;exfalso;apply h;right;eauto.
  case get_bot_spec;[intros ->|intros ? ->|].
  intros h;exfalso;apply h;left;auto.
  intros h;exfalso;apply h;right;eauto.
  intros;split;reflexivity.
Qed.

Ltac destr_pos e :=
  let hf := fresh "h" in
  let eqf := fresh "eq" in
  let gbf := fresh "gb" in
  case (@eq0'_spec e);
    [intros [hf|(? & hf)]
    |intro hf;pose proof (eq0'_get_bot_positive hf) as (eqf&gbf)];
    repeat rewrite hf in * || rewrite eqf in * || rewrite gbf in *;
    simpl;auto.
Ltac simpl_eq0' :=
  let h := fresh "h" in
  let A := fresh "A" in
  case eq0'_spec;[intros [h|(A & h)];try rewrite h in * |intro h].
Ltac simpl_get_bot :=
  let h := fresh "h" in
  let A := fresh "A" in
  case get_bot_spec; [intro h;try rewrite h in *|intros A h;try rewrite h in *|intro h].
