(** printing ⎣ %\ensuremath{\lfloor}% *)
(** printing ⎦ %\ensuremath{\rfloor}% *)
(** printing ⎡ %\ensuremath{\lceil}% *)
(** printing ⎤ %\ensuremath{\rceil}% *)
(** printing • %\ensuremath{\bullet}% *)
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

Require Export expr predicates.

(** * Coertions between untyped and typed expressions. *)

(** Produces an untyped expression by removing all [ω]. *)
Fixpoint erase (e : expr) : expr :=
  match e with
    | 0 | ⊥ _ => 0
    | 1 | id _ => 1
    | var x => var x
    | vat x => vat x
    | e + f => erase e + erase f
    | e · f => erase e · erase f
    | e ^* => (erase e) ^*
    | ‹ p › e => ‹ p › (erase e)
    | ν a e => ν a (erase e)
    | ω a e => erase e
  end.
Hint Unfold erase.
Notation " ⎣ e ⎦ " := (erase e).
(** [erase] perserves [no_perm], [single], [multi], [str_positive] *)
(** and [positive]. *)
(** An atom is fresh for [⎣e⎦] if it doesn't appear in 
    the type of [e]. *)
Lemma fresh_type : forall A e, A ⊨ e -> forall a, ~ In a A -> a # ⎣e⎦.
Proof.
  intros A e;revert A;induction e;simpl;intros A T b P;intros;simpl_in;
  try firstorder subst.
  - firstorder.
  - case_eq_at a b;[left|right];firstorder.
  - apply (IHe _ H0);simpl_in;firstorder.
  - apply (IHe _ T) ;simpl_in;simpl_perm;auto.
Qed.

(** The support of [§e] is included in its type. *)
Lemma support_min_type : forall e A, A ⊨ e -> support ⎣e⎦ ⊆ A.
Proof.
  unfold incl_lst;
  induction e;intro A;simpl;intros T x h;simpl in *;simpl_in;auto;
  try (rewrite T;simpl);firstorder auto.
  - eapply IHe in H0;simpl_in;eauto;tauto.
  - apply (IHe _ T) in h;simpl_in;simpl_perm;tauto.
Qed.

(** [W] gets obliterated by [erase]. *)
Lemma W_erase l e : ⎣W l e⎦ = ⎣e⎦.
Proof.
  induction l;auto.
Qed.
Lemma no_perm_erase : forall e, no_perm e -> no_perm ⎣e⎦.
Proof. induction e;simpl;intuition. Qed.
Lemma single_erase : forall e, single e -> single ⎣e⎦.
Proof. induction e;simpl;intuition. Qed.
Lemma multi_erase : forall e, multi e -> multi ⎣e⎦.
Proof. induction e;simpl;intuition. Qed.
Lemma str_positive_erase :
  forall e, str_positive e -> str_positive ⎣e⎦.
Proof. induction e;simpl;intuition. Qed.
Lemma positive_erase : forall e, positive e -> positive ⎣e⎦.
Proof.
  intros e h;remember e as E;destruct E;try (simpl;tauto);
  apply str_positive_positive,str_positive_erase;auto.
Qed.
(** The size of [⎣e⎦] is smaller than that of [e]. *)
Lemma size_erase : forall e, size ⎣e⎦ <= size e.
Proof. induction e;simpl;intuition. Qed.
(** [⎣e⎦] is always untyped. *)
Lemma untyped_erase : forall e, untyped ⎣e⎦.
Proof. induction e;simpl;intuition. Qed.
(** [⎣⎦] preserves [==]. *)
Lemma eq_expr_erase : Proper (eq_expr ==> eq_expr) erase.
Proof. intro e;induction e;intro f;destruct f;simpl;intuition. Qed.
(** Produces a typed expression, by using [W] and [support] heavily. *)
Fixpoint typed_vers (e : expr) :=
  match e with
    | 0 => ⊥ []
    | 1 => id []
    | var x => var x
    | vat x => vat x
    | e + f =>
      W (〈support f〉--(support e)) (typed_vers e)
      + W (〈support e〉--(support f)) (typed_vers f)
    | e · f =>
      W (〈support f〉--(support e)) (typed_vers e)
      · W (〈support e〉--(support f)) (typed_vers f)
    | e ^* => (typed_vers e)^*
    | ‹ p › e => ‹ p › (typed_vers e)
    | ν a e =>
      if inb a (support e)
      then ν a (typed_vers e)
      else ν a (ω a (typed_vers e))
    | _ => e
  end.
Hint Unfold typed_vers.
Notation " ⎡ e ⎤ " := (typed_vers e).

Section poly.
  Hypothesis K_not_zero : O < K.
  Hypothesis bounded_τ : forall x, length (τ x) <= K.
  Lemma typed_vers_size e : size (⎡e⎤) <= (K*(size e)*(size e))%nat.
  Proof.
    assert (h1 : forall n m, ((n + m) * (n + m) = n*n + m*m + n * m + n * m)%nat)
      by (intros;lia).
    assert (h2 : forall n, (S n = S O + n)%nat)
      by (intros;lia).
    assert (tech:forall n m,
               (S
                  (K * m + K * n * n +
                   (K * n + K * m * m)) <=
                K * S (n + m) * S (n + m))%nat).
    clear bounded_τ .
    intros n m;
      transitivity (K * ((S 0 + n + m) * (S 0 + n + m)))%nat;
      try lia;rewrite h2,h1,h1.
    repeat rewrite Nat.mul_add_distr_l.
    transitivity (K+ K * (n * n) + K * n  + K * n +
                  K * (m * m) + K * m  +
                  K * m + K * n * m + K * n * m)%nat;try lia.
    pose proof (length_support_size K_not_zero bounded_τ) as h.
    induction e;simpl;firstorder (try discriminate || lia).
    - repeat rewrite size_W,purge_length,undup_length;
      rewrite IHe1,IHe2,h,h;auto. 
    - repeat rewrite size_W,purge_length,undup_length;
      rewrite IHe1,IHe2,h,h;auto. 
    - destruct (inb _ _);simpl;rewrite IHe;try lia.
      pose proof (size_non_zero e) as h0.
      rewrite (h2 (size e)),<-(Nat.mul_assoc K),<-(Nat.mul_assoc K),h1.
      transitivity
        (K + K * size e * size e + K * size e + K * size e)%nat;try lia.
      transitivity (K +(K * size e * size e)+S O + S O)%nat;try lia.
      assert (h3 : S O <= K * size e)
        by (cut (O < K * size e);auto;apply Nat.mul_pos_pos;auto).
      lia.
    - transitivity (S (size e) * (S (size e)))%nat.
      repeat rewrite h2;rewrite h1;lia.
      destruct K. lia.
      rewrite (h2 n),<-Nat.mul_assoc, Nat.mul_add_distr_r. lia.
  Qed.
End poly.   
  
(** First half of a bijection between typed and untyped expressions. *)
Lemma typed_version_erase e : untyped e -> ⎣⎡e⎤⎦ = e.
Proof.
  induction e;simpl;intro U;
  try intuition (repeat rewrite W_erase;congruence).
  case_ex a (support e);simpl;intuition congruence.
Qed.

(** [typed_vers] perserves [no_perm], [single], [multi], *)
(** [str_positive] and [positive]. *)
Lemma no_perm_typed_vers : forall e, no_perm e -> no_perm ⎡e⎤.
Proof.
  induction e;simpl;intuition (try apply no_perm_W;auto).
  case_ex a (support e);intuition.
Qed.
Lemma single_typed_vers : forall e, single e -> single ⎡e⎤.
Proof.
  induction e;simpl;intuition (try apply single_W;auto).
  case_ex a (support e);intuition.
Qed.
Lemma multi_typed_vers : forall e, multi e -> multi ⎡e⎤.
Proof. 
  induction e;simpl;intuition (try apply multi_W;auto).
  case_ex a (support e);intuition.
Qed.
Lemma str_positive_typed_vers :
  forall e, str_positive e -> str_positive ⎡e⎤.
Proof.
  induction e;simpl;intuition (try apply str_positive_W;auto).
  case_ex a (support e);intuition.
Qed.
Lemma positive_typed_vers :
  forall e, positive e -> positive ⎡e⎤.
Proof.
  intros e h;remember e as E;destruct E;try (simpl;tauto);
  apply str_positive_positive,str_positive_typed_vers;auto.
Qed.
(** The size of [⎡e⎤] is larger than that of [e]. *)
Lemma size_typed_vers : forall e, size e <= size ⎡e⎤.
Proof.
  induction e;simpl;intuition repeat rewrite size_W;intuition.
  case_ex a (support e);simpl;intuition.
Qed.
(** The support of [e] can type [⎡e⎤], if [e] is untyped. *)
Lemma typed_version_support_type e : untyped e -> (support e) ⊨ ⎡e⎤.
Proof. 
  induction e;simpl;intro U;try tauto || reflexivity.
  - intuition (apply W_type; [prove_type| |repeat solve_no_dup];
               intro x;simpl_in;firstorder).
  - intuition (apply W_type; [prove_type| |repeat solve_no_dup];
               intro x;simpl_in;firstorder).
  - case_ex a (support e);simpl;simpl_in;try simpl_beq;simpl;
    intuition prove_type;auto;intro x;simpl_in.
    firstorder (intros ->;auto).
  - simpl_perm;intuition.
Qed.

(** * [rm0 e] removes [0] and [⊥ _] from [e] if it is possible,
    and returns [0] or [⊥ _] otherwise. *)
Fixpoint rm0 e :=
  match e with
    | e + f =>
      let e := rm0 e in
      let f := rm0 f in     
      if (eq0' e)
      then f
      else if eq0' f
           then e
           else (e + f)
    | e · f =>
      let e := rm0 e in
      let f := rm0 f in  
      if (eq0' e)
      then e
      else if eq0' f
           then f
           else (e · f)
    | e^* =>
      let e := rm0 e in
      match get_bot e with
      | inl false => 1
      | inr A => id A
      | inl true => e^*
      end
    | ν a e =>
      let e := rm0 e in
      match get_bot e with
      | inl false => 0
      | inr A => ⊥ (rm a A)
      | inl true => ν a e
      end
    | ‹ p › e =>
      let e := rm0 e in
      match get_bot e with
      | inl false => 0
      | inr A => ⊥ (p⧓ A)
      | inl true => ‹ p › e
      end
    | ω a e =>
      let e := rm0 e in
      match get_bot e with
      | inl false => 0
      | inr A => ⊥ (a :: A)
      | inl true => ω a e
      end
    | e => e
  end.
(** [rm0] perserves [no_perm],[single],[multi] and [str_positive]. *)
Lemma no_perm_rm0 : forall e, no_perm e -> no_perm (rm0 e).
Proof.
  induction e;simpl;intuition.
  destr_pos (rm0 e1);destr_pos (rm0 e2).
  destr_pos (rm0 e1);destr_pos (rm0 e2).
  destr_pos (rm0 e).
  destr_pos (rm0 e).
  destr_pos (rm0 e).
Qed.
Lemma single_rm0 : forall e, single e -> single (rm0 e).
Proof.
  induction e;simpl;intuition.
  destr_pos (rm0 e1);destr_pos (rm0 e2).
  destr_pos (rm0 e1);destr_pos (rm0 e2).
  destr_pos (rm0 e).
  destr_pos (rm0 e).
  destr_pos (rm0 e).
  destr_pos (rm0 e).
Qed.
Lemma multi_rm0 : forall e, multi e -> multi (rm0 e).
Proof.
  induction e;simpl;intuition.
  destr_pos (rm0 e1);destr_pos (rm0 e2).
  destr_pos (rm0 e1);destr_pos (rm0 e2).
  destr_pos (rm0 e).
  destr_pos (rm0 e).
  destr_pos (rm0 e).
  destr_pos (rm0 e).
Qed.
Lemma str_positive_rm0 : forall e, str_positive e ->
                                   str_positive (rm0 e).
Proof.
  induction e;simpl;intuition.
  destr_pos (rm0 e1);destr_pos (rm0 e2).
  destr_pos (rm0 e1);destr_pos (rm0 e2).
  destr_pos (rm0 e).
  destr_pos (rm0 e).
  destr_pos (rm0 e).
  destr_pos (rm0 e).
Qed.
(** [rm0] preserves untypedness. *)
Lemma untyped_rm0 : forall e, untyped e -> untyped (rm0 e).
Proof.
  intros e;induction e;simpl;intro U;auto;
  repeat simpl_eq0'|| simpl_get_bot;simpl;intuition.
Qed.
(** [rm0] reduces the size of the expression. *)
Lemma size_rm0 : forall e, size (rm0 e) <= size e.
Proof.
  induction e;simpl;auto;
  (destruct (eq0' (rm0 e1));simpl;destruct (eq0' (rm0 e2));simpl)
  ||(repeat destruct (get_bot _);simpl);
  try (destruct b;simpl);lia.
Qed.
(** [rm0] preserves types. *)
Lemma typed_rm0 : forall e A, A ⊨ e -> A ⊨ (rm0 e).
Proof.
  intros e;induction e;simpl;intros A T;
  repeat simpl_eq0'|| simpl_get_bot;
  simpl in *;intuition;eauto.
  now apply IHe in H0 as <-;auto.
  now apply IHe in H0 as <-;auto.
  now apply IHe in T as <-;simpl_perm.
Qed.

(** [rm0] and [eq0] agree on untyped expressions. *)
Lemma eq0_rm0_u : forall e, eq0 e = eq0' (rm0 e).
Proof.
  induction e;simpl;auto.
  - rewrite IHe1,IHe2;clear IHe1 IHe2.
    repeat (case eq0'_spec;auto).
  - rewrite IHe1,IHe2;clear IHe1 IHe2.
    repeat (case eq0'_spec;tauto||auto).
  - repeat (case get_bot_spec;tauto||auto).
  - rewrite IHe;clear IHe; destr_pos (rm0 e).
  - rewrite IHe;clear IHe; destr_pos (rm0 e).
  - rewrite IHe;clear IHe; destr_pos (rm0 e).
Qed.
(** [rm0 e] is either syntactically null or [str_positive]. *)
Lemma positive_or_nul_rm0 e : eq0' (rm0 e) = true \/
                              (str_positive (rm0 e)).
Proof.
  assert (aux:forall f, str_positive f -> get_bot f = inl true)
    by (destruct f;simpl;tauto).
  induction e; simpl;auto;
  revert IHe1 IHe2 || revert IHe ;repeat case eq0'_spec;simpl;
  try (intros;left;reflexivity);
  repeat (let h := fresh "h" in
          (intros [h|(?&h)];try rewrite h in * )
          || (intros [h|h];[try discriminate
                           |try rewrite (aux _ h) in *])
          || intros h);
  simpl in *;try tauto; eauto.
Qed.
(** [rm0 e] is always [positive]. *)
Lemma positive_rm0 e : positive (rm0 e).
Proof.
  destruct (positive_or_nul_rm0 e) as [E|E].
  revert E;case eq0'_spec;[intros [->|(?&->)]|discriminate];auto.
  now apply str_positive_positive.
Qed.
(** [rm0] is the identity over [str_positive] expressions. *)
Lemma str_positive_rm0_id : forall e, str_positive e -> rm0 e = e.
Proof.
  induction e;simpl;tauto
                    || (intros (Pe&Pf);rewrite IHe1,IHe2;
                        destr_pos (e1);[| |destr_pos (e2)];
                        simpl in *; tauto)
                    || (intros Pe;rewrite IHe; destr_pos e;
                        simpl in *; tauto).
Qed.
(** [positive] expressions compose the kernel of [rm0]. *)
Lemma positive_rm0_id : forall e, positive e <-> rm0 e = e.
Proof.
  intros e;split.
  intro h;remember e as E;destruct E;try (simpl;tauto);
  apply str_positive_rm0_id;auto.
  intros <-;apply positive_rm0. 
Qed.
(** [rm0] is involutive. *)
Lemma rm0_involutive : forall e, rm0 (rm0 e) = rm0 e.
Proof. intro e;rewrite <-positive_rm0_id;apply positive_rm0. Qed.
(** [rm0] preserves freshness for untyped expressions. *)
Lemma fresh_rm0 a e : untyped e -> fresh a e -> fresh a (rm0 e).
Proof.
  revert a;induction e;simpl;try tauto;intuition.
  - apply (IHe1 _ H1) in H;apply (IHe2 _ H2) in H3.
    destr_pos (rm0 e1); destr_pos (rm0 e2).
  - apply (IHe1 _ H1) in H;apply (IHe2 _ H2) in H3.
    destr_pos (rm0 e1); destr_pos (rm0 e2).
  - apply (IHe _ H) in H0; destr_pos (rm0 e).
  - rewrite H1;apply untyped_rm0 in H; destr_pos (rm0 e).
  - apply (IHe _ H) in H1;apply untyped_rm0 in H; destr_pos (rm0 e).
  - apply (IHe _ H) in H0;apply untyped_rm0 in H; destr_pos (rm0 e).
Qed.
(** [rm0] preserves [==]. *)
Lemma eq_expr_rm0 : Proper (eq_expr ==> eq_expr) rm0.
Proof.
  intro e;induction e;intro f;destruct f;simpl;intuition.
  - pose proof (IHe1 _ H0) as E1; pose proof (IHe2 _ H1) as E2.
    rewrite (eq_expr_eq0' E1), (eq_expr_eq0' E2).
    destruct (eq0' _);auto;destruct (eq0' _);simpl;auto.
  - pose proof (IHe1 _ H0) as E1; pose proof (IHe2 _ H1) as E2.
    rewrite (eq_expr_eq0' E1), (eq_expr_eq0' E2).
    destruct (eq0' _);auto;destruct (eq0' _);simpl;auto.
  - pose proof (IHe _ H) as E.
    destruct (eq_expr_get_bot E) as [->| (a & b & -> & -> & L)];
      simpl;auto.
    destruct (get_bot _);simpl;auto;destruct b;simpl;auto.
  - pose proof (IHe _ H0) as E;subst.
    destruct (eq_expr_get_bot E) as [->| (a & b & -> & -> & L)].
    * destruct (get_bot _);simpl;auto;destruct b;simpl;auto.
    * now simpl;rewrite L.
  - pose proof (IHe _ H0) as E;subst.
    destruct (eq_expr_get_bot E) as [->| (a & b & -> & -> & L)].
    * destruct (get_bot _);simpl;auto;destruct b;simpl;auto.
    * now simpl;rewrite L.
  - pose proof (IHe _ H0) as E;subst.
    destruct (eq_expr_get_bot E) as [->| (a & b & -> & -> & L)].
    * destruct (get_bot _);simpl;auto; try destruct b;simpl;auto.
      now rewrite (equiv_perm_mact _ _ H1).
    * now simpl;rewrite (equiv_perm_mact _ _ H1);rewrite L.
Qed.
    

(** * Applying permutations. *)
Reserved Notation " p • e " (at level 19).
Lemma  apply_perm_var_bis : forall p x l,  is_perm_bool (x,l) = true ->
    is_perm_bool (x,p ⧓ l) = true.
Proof.
  intros p x l h; simpl_is_perm in h as q;simpl_is_perm.
  exists (q++p⋆);rewrite <-mact_comp_app;simpl;  
  repeat simpl_perm;simpl in *;auto.
Qed.

Definition apply_perm_var (p : perm) (n : Letter) : Letter :=
  match n with
  | exist _ (x,l) h => (exist (fun x => is_perm_bool x = true) (x,p⧓l)
                              (@apply_perm_var_bis p x l h))
  end.

Lemma apply_perm_var_τ (p : perm) (n : Letter) :
  τ (apply_perm_var p n) = p ⧓ (τ n).
Proof. now destruct n as ((x,l),m);unfold τ;simpl. Qed.
Lemma apply_perm_var_app (p q : perm) (n : Letter) :
  (apply_perm_var q (apply_perm_var p n)) ≃ apply_perm_var (q++p) n.
Proof. now destruct n as ((x,l),m);unfold apply_perm_var, eq_letter; 
       simpl;rewrite mact_comp_app.
Qed.
Lemma apply_perm_var_equiv_perm n :
  Proper (equiv_perm ==> eq_letter) (Basics.flip apply_perm_var n).
Proof.
  intros p q E;unfold eq_letter ;destruct n as ((?,?),?);simpl.
  now rewrite (equiv_perm_mact _ _ E).
Qed.
Lemma apply_perm_var_eq_letter p :
  Proper (eq_letter ==> eq_letter) (apply_perm_var p).
Proof.
  intros ((?,?),?) ((?,?),?) E;inversion E as [[ h1 h2 ]];
  unfold eq_letter in *;simpl in *;rewrite h1,h2;reflexivity.
Qed.
Lemma apply_perm_var_equiv_perm_eq_letter :
  Proper (equiv_perm ==> eq_letter ==> eq_letter) apply_perm_var.
Proof.
  intros p q E x y F;
  rewrite (apply_perm_var_eq_letter _ F).
  pose proof (apply_perm_var_equiv_perm y E) as H;
    unfold Basics.flip in H;auto.
Qed.

Fixpoint apply_perm p e :=
  match e with
  | id A => id (p⧓A)
  | ⊥ A => ⊥ (p⧓A)
  | 0 => 0
  | 1 => 1
  | var x => var (apply_perm_var p x)
  | vat a => vat (p⋈a)
  | e + f => p • e + p • f
  | e · f => p • e · p • f
  | e^* => p • e ^*
  | ‹ q › e => ‹ [] › (p++q) • e
  | ν a e => ν (p⋈a) (p • e)
  | ω a e => ω (p⋈a) (p • e)
  end
where " p • e " := (apply_perm p e).
Hint Unfold apply_perm.
Global Instance apply_perm_equiv_perm :
  Proper (equiv_perm ==> eq_expr ==> eq_expr) apply_perm.
Proof.
  intros p q E e;revert p q E; induction e;intros θ θ' E f h;
  destruct f;simpl in *;intuition.
  - now rewrite h,(equiv_perm_mact _ _ E).
  - now rewrite h,(equiv_perm_mact _ _ E).
  - now f_equal;apply apply_perm_var_equiv_perm_eq_letter.
  - rewrite h;apply E.
  - rewrite H0;apply E.
  - rewrite H0;apply E.
  - erewrite IHe; try reflexivity;auto.
    apply equiv_perm_app;auto.
Qed.

(** [p•e] is not bigger than twice the size of [e]. *)
Lemma apply_perm_size e p : size (p•e) <= 2 * size e. 
Proof.
  revert p;induction e;intro θ;simpl;try lia.
  rewrite IHe1,IHe2;lia.
  rewrite IHe1,IHe2;lia.
  rewrite IHe;lia.
  rewrite IHe;lia.
  rewrite IHe;lia.
  rewrite IHe;lia.
Qed.
  
(** [apply_perm] perserves [single], [multi] and [eq0].*)
Lemma multi_apply_perm : forall e p, multi e <-> multi (p•e).
Proof. induction e;intro q;simpl;firstorder. Qed.
  
Lemma single_apply_perm : forall e p, single e <-> single (p•e).
Proof. induction e;intro q;simpl;firstorder. Qed.

Lemma eq0_apply_perm : forall e p, eq0 (p•e) = eq0 e.
Proof. induction e;intro q;simpl;firstorder congruence. Qed.

(** [apply_perm] doesn't change [no_perm]. *)
Lemma no_perm_apply_perm :
  forall e p,  (no_perm e <-> no_perm (p•e)).
Proof. induction e;intro q ;simpl; firstorder. Qed.

(** [apply_perm] isn't a problem for typing. *)
Lemma has_type_apply_perm : forall e A p, (p⋆⧓A) ⊨ e -> A ⊨ p • e.
Proof.
  induction e;intros X q;simpl;try simpl_perm;intuition;try tauto.
  now rewrite <-H;simpl_perm.
  now rewrite <-H;simpl_perm.
  destruct l as ((x,l),m).
  now unfold τ,apply_perm_var in *;simpl in *;
  rewrite <-H;simpl_perm.
  replace [q⋈a] with (q⧓[a]) by (simpl;auto).
  now rewrite <-H;simpl;simpl_perm.
  simpl_in;simpl_perm;tauto.
  now apply IHe;simpl;simpl_perm.
  simpl_in;simpl_perm;tauto.
  now apply IHe; rewrite mact_rm in H1;simpl_perm.
  now apply IHe;simpl;simpl_perm;rewrite <-mact_comp_app.
Qed.
Lemma apply_perm_has_type {e A} :
  forall p, (p⧓A) ⊨ (p • e) ->  A ⊨ e.
Proof.
  revert A;induction e;intros X q;simpl;try simpl_perm;intro h;
  try tauto.
  now rewrite<-(@mact_cong q);simpl.
  now rewrite<-(@mact_cong q);simpl.
  destruct l as ((x,l),m);unfold τ,apply_perm_var in *;simpl in *.
  now  rewrite<-(@mact_cong q);simpl.
  replace [q⋈a] with (q⧓[a]) by (simpl;auto).
  now rewrite<-(@mact_cong q);simpl.
  split;[apply (IHe1 _ q)|apply (IHe2 _ q)];tauto.
  split;[apply (IHe1 _ q)|apply (IHe2 _ q)];tauto.
  now apply (IHe _ q).
  simpl_in;simpl_perm;split;[|apply (IHe _ q)];tauto.
  simpl_in;split;[|apply (IHe _ q);rewrite mact_rm in h];
  simpl_perm;tauto.
  now apply (IHe _ (q++p));rewrite <-mact_comp_app;simpl_perm.
Qed.

(** [apply_perm] doesn't change untypedness. *)
Lemma untyped_apply_perm : forall e p, untyped e <-> untyped (p•e).
Proof. induction e;intro q;simpl;firstorder. Qed.

(** [apply_perm] works well with freshness. *)
Lemma fresh_apply_perm : forall e a p, a # e <-> (p⋈a) # p • e.
Proof.
  induction e;intros b q;simpl;try simpl_perm;
  try rewrite perm_bij in *;firstorder.
  now destruct l as ((x,l),m);unfold τ,apply_perm_var in *;
  simpl in *; simpl_in;simpl_perm.
  now destruct l as ((x,l),m);unfold τ,apply_perm_var in *;
  simpl in *; simpl_in;simpl_perm.
  now rewrite (IHe _ (q++p)),<-act_comp_app in H;simpl_perm.
  now rewrite (IHe _ (q++p)),<-act_comp_app;simpl_perm.
Qed.

(** The support of [p•e] is [p] applied to the support of [e]. *)
Lemma support_apply_perm : forall e p, support (p•e) = p ⧓ (support e).
Proof.
  induction e;intro q;simpl;auto.
  now destruct l as ((x,l),m);unfold τ,apply_perm_var in *;
  simpl in *; simpl_in;simpl_perm.
  now rewrite IHe1,IHe2,mact_app_distr.
  now rewrite IHe1,IHe2,mact_app_distr.
  now rewrite IHe,mact_rm;simpl_perm.
  now rewrite mact_nil,mact_comp_app.
Qed.
(** [apply_perm] commutes with [erase],[typed_vers] and [W].*)
Lemma erase_apply_perm : forall e p, p • ⎣e⎦ = ⎣p•e⎦.
Proof. induction e;intro q;simpl;firstorder congruence. Qed.

Lemma W_apply_perm : forall e l p, p • (W l e) = W (p⧓l) (p•e).
Proof. intro e;induction l;intro q;simpl;firstorder congruence. Qed.

Lemma typed_vers_apply_perm : forall e p, p • ⎡e⎤ = ⎡p•e⎤.
Proof.
  induction e;intro q;simpl;firstorder (try congruence).
  repeat rewrite W_apply_perm,support_apply_perm,<- mact_undup, mact_purge;
    rewrite IHe1,IHe2;simpl_perm;auto.
  repeat rewrite W_apply_perm,support_apply_perm,<- mact_undup, mact_purge;
    rewrite IHe1,IHe2;simpl_perm;auto.
  rewrite support_apply_perm;auto.
  case_ex a (support e);case_ex (q⋈a) (q⧓support e);
    try (simpl_in;simpl_perm;tauto);simpl;rewrite IHe;auto.
Qed.

(** [apply_perm] preserves [positive] and [str_positive]. *)
Lemma str_positive_apply_perm e p :
  str_positive e <-> str_positive (p•e).
Proof. revert p;induction e;firstorder. Qed.

Lemma positive_apply_perm e p :
  positive e <-> positive (p•e).
Proof.
  revert p;induction e;intros ?;simpl;
  repeat rewrite <-str_positive_apply_perm;tauto.
Qed.
Lemma rm0_inv_apply_perm : forall e p, rm0 e = e -> rm0 (p•e) = p•e.
Proof.
  intros;apply positive_rm0_id,positive_apply_perm,positive_rm0_id;auto.
Qed.

(** Simplification lemma. *)
Lemma eq0'_rm0_apply_perm : forall e p, eq0' (rm0 (p•e)) = eq0' (rm0 e).
Proof. now intros;rewrite<-eq0_rm0_u,<-eq0_rm0_u,eq0_apply_perm. Qed.

(** Technical lemma. *)
Lemma rm0_eq0_get_bot : forall e p,
    rm0 e = e -> reflect3 (e = 0)
                          (fun B => e = ⊥ (p⋆⧓B))
                          (get_bot e = inl true)
                          (get_bot (p • e)).
Proof.
  induction e;simpl;intros q h;
  try (apply r3_true||apply r3_false || apply r3_Some);auto.
  now simpl_perm.
Qed.  
  
(** [apply_perm] commutes wiht [rm0]. *)
Lemma rm0_apply_perm : forall e p, rm0 (p•e) = p•rm0 e.
Proof.
  induction e;intro q;simpl;firstorder (try congruence).
  - repeat rewrite eq0'_rm0_apply_perm; rewrite IHe1,IHe2.
    destruct (eq0' (rm0 e1));destruct (eq0' (rm0 e2));auto.
  - repeat rewrite eq0'_rm0_apply_perm; rewrite IHe1,IHe2.
    destruct (eq0' (rm0 e1));destruct (eq0' (rm0 e2));auto.
  - rewrite IHe;
    destruct (@rm0_eq0_get_bot (rm0 e) q (rm0_involutive e)).
    rewrite e0;simpl;auto.
    rewrite e0;simpl;simpl_perm;auto.
    rewrite e0;simpl;auto.
  - rewrite IHe;
    destruct (@rm0_eq0_get_bot (rm0 e) q (rm0_involutive e)).
    rewrite e0;simpl;auto.
    rewrite e0;simpl;rewrite mact_rm;simpl_perm;auto.
    rewrite e0;simpl;auto.
  - rewrite IHe;
    destruct (@rm0_eq0_get_bot (rm0 e) q (rm0_involutive e)).
    rewrite e0;simpl;auto.
    rewrite e0;simpl;simpl_perm;auto.
    rewrite e0;simpl;auto.
  - rewrite IHe;
    destruct (@rm0_eq0_get_bot (rm0 e) (q++p) (rm0_involutive e)).
    rewrite e0;simpl;auto.
    rewrite e0;simpl;simpl_perm;rewrite <-mact_comp_app;simpl_perm;auto.
    rewrite e0;simpl;auto.
Qed.

(** If [e] doesn't contain any [θ] constructors, *)
(** then applying the permutations [[(a,a)]] or [[]] doesn't change *)
(** the expression. *)
Lemma apply_perm_nil (e : expr) :
  no_perm e -> [] • e == e.
Proof. induction e;simpl;try rewrite mact_nil;firstorder try congruence.
       unfold eq_letter; destruct l as ((?,?),?);simpl; f_equal;
       apply mact_nil.
Qed.
Lemma apply_perm_id e a :
  no_perm e -> [a,a] • e == e.
Proof.
  intros P; rewrite apply_perm_equiv_perm.
  - now apply apply_perm_nil;auto.
  - apply perm_a_a_eq_nil.
  - firstorder.
Qed.
(** * Removing permutations*)
Fixpoint rm_perm e :=
  match e with
  | e + f => rm_perm e + rm_perm f
  | e · f => rm_perm e · rm_perm f
  | e ^* => (rm_perm e)^*
  | ν a e => ν a (rm_perm e)
  | ω a e => ω a (rm_perm e)
  | ‹ _ › e => rm_perm e
  | e => e
  end.
Hint Unfold rm_perm.
(** [rm_perm] reduces the size of expressions. *)
Lemma rm_perm_size e : size (rm_perm e) <= size e.
Proof. induction e;simpl;auto with arith;lia. Qed.
(** [rm_perm] doesn't change untypedness, [single] or [str_positive]. *)
Lemma untyped_rm_perm e : untyped e <-> untyped (rm_perm e).
Proof. induction e;simpl;tauto. Qed.
Lemma single_rm_perm e : single e <-> single (rm_perm e).
Proof. induction e;simpl;intuition. Qed.
Lemma multi_rm_perm e : multi e <-> multi (rm_perm e).
Proof. induction e;simpl;intuition. Qed.
Lemma str_positive_rm_perm : forall e, str_positive e <->
                                   str_positive (rm_perm e).
Proof. induction e;firstorder. Qed.
(** [rm_perm] preserves [positive]. *)
Lemma positive_rm_perm : forall e, positive e -> positive (rm_perm e).
Proof.
  induction e;simpl;repeat rewrite<- str_positive_rm_perm in *;
  firstorder. apply str_positive_positive in H;auto.
Qed.
(** The kernel of [rm_perm] is [no_perm]. *)
Lemma no_perm_rm_perm_id e : no_perm e <-> rm_perm e = e.
Proof.
  induction e;simpl;try tauto;
  try (split;[intros (h1 & h2);rewrite IHe1 in h1;rewrite IHe2 in h2;
              rewrite h1,h2
             |intro h;rewrite IHe1,IHe2;inversion h as ((h1 & h2));
              repeat rewrite h1,h2])
  || (split;[intros h;rewrite IHe in h;rewrite h
            |intro h;rewrite IHe;inversion h as (h');
             repeat rewrite h']);auto.
  split. tauto.
  intro h;assert (C:size e >= size (‹ p ›e)).
  rewrite <-h;apply rm_perm_size.
  simpl in C;lia.
Qed.
(** [rm_perm] produces expressions statisfying [no_perm]. *)
Lemma no_perm_rm_perm e : no_perm (rm_perm e).
Proof. induction e;simpl;intuition. Qed.
(** [rm_perm] and [apply_perm] *)
(** commute partially. *)
Lemma rm_perm_apply_perm_swap p q e :
  rm_perm ((p++q)•e) == p•(rm_perm (q•e)).
Proof.
  revert p q;induction e;intros p1 p2;simpl in *;auto;try tauto.
  now rewrite mact_comp_app.
  now rewrite mact_comp_app.
  now rewrite apply_perm_var_app.
  now rewrite act_comp_app.
  rewrite IHe,act_comp_app;firstorder.
  rewrite IHe,act_comp_app;firstorder.
  rewrite app_ass,IHe;firstorder.
Qed.

(** [rm_perm] preserves [==]. *)
Lemma eq_expr_rm_perm : Proper (eq_expr ==> eq_expr) rm_perm.
Proof. intro e;induction e;intro f;destruct f;simpl;intuition. Qed.

(** [fresh] and [has_type] behave well *)
(** with [rm_perm] and [apply_perm]. *)
Lemma fresh_rm_perm p a e : (a # e <-> p ⋈ a # rm_perm (p • e)).
Proof.
  revert p a;induction e;intros q b; simpl;simpl_in;try tauto.
  rewrite apply_perm_var_τ;simpl_in;simpl_perm;tauto.  
  rewrite perm_bij;tauto.
  rewrite (IHe1 q),(IHe2 q);tauto.
  rewrite (IHe1 q),(IHe2 q);tauto.
  rewrite (IHe q);tauto.
  rewrite (IHe q),perm_bij;tauto.
  rewrite (IHe q);tauto.
  rewrite (IHe (q++p)),<-act_comp_app;simpl_perm;tauto.
Qed.

Lemma has_type_rm_perm_apply_perm e p A :
  A ⊨ (p•e) -> A ⊨ rm_perm (p•e).
Proof.
  revert p A;induction e;intros q A T;simpl in *;auto;try tauto.
  split;apply IHe1||apply IHe2;tauto.
  split;apply IHe1||apply IHe2;tauto.
  split;try apply IHe;tauto.
  split;try apply IHe;tauto.
  rewrite mact_nil in T;apply IHe;tauto.
Qed. 

Ltac destr_posu e :=
  let hf := fresh "h" in
  let eqf := fresh "eq" in
  let gbf := fresh "gb" in
  let hyp := fresh "U" in
  assert (hyp: untyped e) by (now simpl in *;
                               auto using untyped_rm0,
                               untyped_apply_perm,untyped_erase);
  case (eq0'_u_spec hyp);
    [intros hf
    |intro hf;pose proof (eq0'_get_bot_positiveu hyp hf) as (eqf&gbf)];
    repeat rewrite hf in * || rewrite eqf in * || rewrite gbf in *;
    simpl;auto.
Ltac destr_post e A :=
  let hf := fresh "h" in
  let eqf := fresh "eq" in
  let gbf := fresh "gb" in
  let hyp := fresh "T" in
  assert (hyp: A ⊨ e) by (now simpl in *;
                           auto using typed_version_support_type,
                           has_type_apply_perm,typed_rm0);
  case (eq0'_t_spec hyp);
    [intros (? & hf)
    |intro hf;pose proof (eq0'_get_bot_positivet hyp hf) as (eqf&gbf)];
    repeat rewrite hf in * || rewrite eqf in * || rewrite gbf in *;
    simpl;auto.
