(** printing → %\ensuremath{\mapsto}% *)
(** printing << %\ensuremath{\preceq}% *)
(** printing ≡≡ %\ensuremath{\approxeq}% *)
(** printing ⊩ %\ensuremath{:}% *)
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

Require Import proofs rm0 removetype perm theories sg_to_mlt.

(** * Preorder on theories. *)
(** We say that [P,A] is smaller than [Q,B] if for every pair [e,f] of
expressions satisfying [P] we can find a pair [e',f'] such that : 
- both [e'] and [f'] satisfy [Q] and 
- [A ⊢ e ≡ f] if and only if [B ⊢ e' ≡ f']. *)
Definition included_th (S1 S2 : TyEq) : Prop :=
  let (P,A) := S1 in
  let (Q,B) := S2 in
  forall e f, P e -> P f ->
              exists e' f', Q e' /\ Q f'
                            /\ (A ⊢ e ≡ f <-> B ⊢ e' ≡ f').
Infix "<<" := included_th (at level 60).
Definition equiv_th S T := (included_th S T)/\(included_th T S).
Infix "≡≡" := equiv_th (at level 60).

Global Instance included_th_PreOrder : PreOrder included_th.
Proof.
  split.
  - intros (P&A); intros e f;exists e;exists f;tauto.
  - intros (P&A) (Q&B) (R&C) AB BC e f he hf.
    destruct (AB e f he hf) as (e1&f1&he1&hf1&h1).
    destruct (BC e1 f1 he1 hf1) as (e2&f2&h2).
    exists e2;exists f2;tauto.
Qed.
Instance equiv_th_Equivalence : Equivalence equiv_th.
Proof.
  split.
  - split;reflexivity.
  - now intros x y (h1&h2);split.
  - intros x y z (h1&h2) (h3&h4);split;etransitivity;eauto.
Qed.

(** If the theory [(P,A)] is smaller than [(Q,B)], and if the
equivalence relation [B ⊢ _≡_] is decidable for elements satifying
[Q], then the equivalence relation [A ⊢ _≡_] is decidable for elements
satifying [P]. *)
Lemma included_th_decidable :
  forall P A Q B,
    (P,A) << (Q,B) ->
    (forall e f, Q e -> Q f -> Decidable.decidable (B ⊢ e ≡ f)) ->
    forall e f, P e -> P f -> Decidable.decidable (A ⊢ e ≡ f).
Proof.
  unfold Decidable.decidable in *;intros P A Q B I D e f he hf.
  destruct (I e f he hf) as (e1 & f1 & he1 & hf1 & Eq);rewrite Eq in *.
  destruct (D e1 f1 he1 hf1) as [E|E];tauto.
Qed.

(** This notion will be instrumental in proving orderings. [φ] is a
transfer morphism between relation [P] with axioms [A] to relation [Q]
with axioms [B] if for every pair [e,f] related by [P], the pair 
[φ e,φ f] is related by [Q] and [A ⊢ e ≡ f] if and only if 
[B ⊢ φ e ≡ φ f]. *)
Definition transfer_fun (R1 R2 : relation expr * relation expr) φ : Prop :=
  let (P,A):= R1 in
  let (Q,B):= R2 in
    forall e f, P e f -> 
                Q (φ e) (φ f) /\
                (A ⊢ e ≡ f <-> B ⊢ (φ e) ≡ (φ f)).
Notation " φ ⊩ R →  S  " := (transfer_fun R S φ) (at level 50).

(** We may use this to define a stronger preorder on theories. *)
Definition is_model (T1 T2 : TyEq) :=
  let (P,A):= T1 in let (Q,B):= T2 in
                    exists φ, φ ⊩ (P^2,A) → (Q^2,B).
Infix "⊲" := is_model (at level 60).

(** Existence of such a morphism is a also a preorder. *)
Lemma transfer_fun_preorder : PreOrder is_model.
Proof.
  split.
  - intros (?,?);exists (fun x => x);intros ? ?;firstorder.
  - intros (?,?) (?,?) (?,?) (φ,h1) (ψ,h2);exists (ψ∘φ);intros e f.
    unfold Basics.compose;intro h;pose proof (h1 e f h) as (h3 & ->).
    pose proof (h2 _ _ h3) as (h4 & ->);firstorder.
Qed.

(** This preorder is stronger. *)
(** Indeed it is straightforward to show that if there is a transfer
morphism from [P^2] with axioms [A] to [Q^2] with axioms [B] then [P,A]
is smaller than [Q,B].*)
Lemma transfer_fun_included_th t1 t2 : t1 ⊲ t2 -> t1 << t2.
Proof. destruct t1 as (?&?); destruct t2 as (?&?);
       intros (φ&T) e f he hf; exists (φ e);exists (φ f);firstorder.
Qed.

(** ** Technical lemmas about transfer morphisms. *)
Global Instance transfer_fun_proper_str_impl :
  Proper (Basics.flip (@stronger expr) ==> (@eq_rel expr)
                      ==> (@stronger expr) ==> (@eq_rel expr)
                      ==> Logic.eq
                      ==> Basics.impl)
         (fun P A Q B => transfer_fun (P,A) (Q,B)).
Proof.
  autounfold; unfold Basics.flip, Proper, respectful, transfer_fun,Basics.impl.
  intros P P' E1 A A' E2 Q Q' E3 B B' E4 φ ? <- h e f h1;split.
  - now apply E3,h,E1.
  - split;intro h2.
    * apply (@stronger_ax_proof B);[firstorder|];apply h.
        now apply E1.
        apply (@stronger_ax_proof A');firstorder.
    * apply (@stronger_ax_proof A);[firstorder|]; apply h.
        now apply E1.
        apply (@stronger_ax_proof B');firstorder.
Qed.
Global Instance transfer_fun_proper_eq_impl :
  Proper ((@eq_rel expr) ==> (@eq_rel expr)
                      ==> (@eq_rel expr) ==> (@eq_rel expr)
                      ==> Logic.eq
                      ==> Basics.impl)
         (fun P A Q B => transfer_fun (P,A) (Q,B)).
Proof.
  autounfold; unfold Basics.flip, Proper, respectful, transfer_fun.
  intros P P' E1 A A' E2 Q Q' E3 B B' E4 φ ? <-;
         apply transfer_fun_proper_str_impl;firstorder.
Qed.
Global Instance transfer_fun_proper_eq_impl_rev :
  Proper ((@eq_rel expr) ==> (@eq_rel expr)
                      ==> (@eq_rel expr) ==> (@eq_rel expr)
                      ==> Logic.eq
                      ==> Basics.flip Basics.impl)
         (fun P A Q B => transfer_fun (P,A) (Q,B)).
Proof.
  autounfold; unfold Basics.flip, Proper, respectful, transfer_fun.
  intros P P' E1 A A' E2 Q Q' E3 B B' E4 φ ? <-;
         apply transfer_fun_proper_str_impl;firstorder.
Qed.
Global Instance transfer_fun_proper_iff :
  Proper ((@eq_rel expr) ==> (@eq_rel expr) ==> (@eq_rel expr) ==>
                        (@eq_rel expr) ==> Logic.eq ==> iff)
         (fun P A Q B => transfer_fun (P,A) (Q,B)).
Proof.
  autounfold; unfold Proper, respectful, transfer_fun.
  intros P P' E1 A A' E2 Q Q' E3 B B' E4 φ ? <- ;
    split;apply transfer_fun_proper_eq_impl;firstorder.
Qed.
Global Instance transfer_fun_proper_iff_1 A Q B φ :
  Proper ((@eq_rel expr) ==> Basics.flip Basics.impl)
         (fun P => φ ⊩ (P,A) → (Q,B)).
Proof. intros P P' E;apply transfer_fun_proper_eq_impl;firstorder. Qed.
Global Instance transfer_fun_proper_iff_2 P A B φ :
  Proper ((@eq_rel expr) ==>Basics.flip Basics.impl)
         (fun Q => φ ⊩ (P,A) → (Q,B)).
Proof. intros Q Q' E;apply transfer_fun_proper_eq_impl;firstorder. Qed.
Lemma TTrue_unit_l a A b B φ :
  (φ ⊩ (a,A) → (b,B)) <-> φ ⊩ (TTrue∧a,A) → (b,B).
Proof. unfold transfer_fun;firstorder. Qed.
Lemma TTrue_unit_r a A b B φ :
  (φ ⊩ (a,A) → (b,B)) <-> φ ⊩ (a,A) → (TTrue∧b,B).
Proof. unfold transfer_fun;firstorder. Qed.
Lemma incl_p_fun (p q : relation expr) φ P A Q B :
  (φ ⊩ (P,A) → (Q,B)) ->
  (forall x y, (p∧P) x y -> q (φ x) (φ y)) ->
  (φ ⊩ (p∧P,A) → (q∧Q,B)).
Proof. unfold transfer_fun. firstorder. Qed.
Lemma incl_p_fun_r (q : relation expr) φ P A Q B :
  (φ ⊩ (P,A) → (Q,B)) ->
  (forall x y, P x y -> q (φ x) (φ y)) ->
  (φ ⊩ (P,A) → (q∧Q,B)).
Proof. unfold transfer_fun. firstorder. Qed.
Lemma incl_p_fun_l (p : relation expr) φ P A Q B :
  (φ ⊩ (P,A) → (Q,B)) -> φ ⊩ (p∧P,A) → (Q,B).
Proof. unfold transfer_fun. firstorder. Qed.
Lemma augm_transfer (p : predicate expr) φ P A Q B :
  (φ ⊩ (P,A) → (Q,B)) ->
  (forall x, p x -> p (φ x)) ->
  (φ ⊩ (p^2∧P,A) → (p^2∧Q,B)).
Proof. unfold transfer_fun. firstorder. Qed.
Lemma augm_transfer' (p : predicate expr) φ A B :
  (φ ⊩ (TTrue,A) → (TTrue,B)) ->
  (forall x, p x -> p (φ x)) ->
  (φ ⊩ (p^2,A) → (p^2,B)).
Proof. unfold transfer_fun. firstorder. Qed.
Lemma augm_transfer2 (p : predicate expr) φ P A Q B :
  (φ ⊩ (P,A) → (Q,B)) ->
  (forall x, p (φ x)) ->
  (φ ⊩ (P,A) → (p^2∧Q,B)).
Proof. unfold transfer_fun. firstorder. Qed.

Ltac push_rel_down :=
  eapply transfer_fun_proper_iff_1;[|eapply transfer_fun_proper_iff_2];
  repeat rewrite <- rel_and_pred;try reflexivity.
Lemma technical e0 f0 (p q : expr->Prop) (a b r s : relation expr) φ :
  (φ ⊩ (r,a) → (s,b)) ->
  (forall e f, r e f \/ ~ p e \/ ~ p f \/ ~ (a ⊢ e≡f)) ->
  ~ (b ⊢ e0 ≡ f0) -> q e0 -> q f0 ->
  (forall e f, s e f -> q ^2 e f \/ (e == f) \/ ~ (b ⊢ e≡f)) ->
  (p,a) << (q,b).
Proof.
  intros tr h1 b0 q0 q0' h2 e f he hf.
  case (h1 e f) as [h|[h|[h|h]]];firstorder.
  destruct (tr e f h) as (hs&Eq).
  apply h2 in hs as [(h3 & h4)|[h3|h3]].
  firstorder.
  rewrite (eq_expr_eq _ h3) in Eq.
  firstorder.
  firstorder.
Qed.
Lemma pos_rev_rm0 {A B a} :
  (rm0 ⊩ (a,A) → (a,B)) -> rm0 ⊩ (po∧a,B) → (po∧a,A).
Proof. 
  intros I e f ((P1 & P2)&U).
  pose proof (I e f U) as (ha&E).
  set (P1':=P1);apply (positive_rm0_id) in P1'.
  set (P2':=P2);apply (positive_rm0_id) in P2'.
  rewrite P1',P2'. rewrite<-P1',<-P2' at 2.
  firstorder.
Qed.
Lemma str_pos_rev_rm0 {A B a} :
  (rm0 ⊩ (a,A) → (a,B)) ->
  rm0 ⊩ (sp∧a,B) → (sp∧a,A).
Proof. 
  intros I e f ((P1 & P2)&U).
  pose proof (I e f U) as (ha&E).
  set (P1':=P1);apply (str_positive_rm0_id) in P1'.
  set (P2':=P2);apply (str_positive_rm0_id) in P2'.
  rewrite P1',P2'. rewrite<-P1',<-P2' at 2.
  firstorder.
Qed.
Lemma po_to_str_rm0 r a s b :
  (rm0 ⊩ (r^2,a) → (po∧(s^2),b)) ->
  (forall e, r e \/ ~ r e) ->
  ((forall e, s e -> untyped e)
   \/ (forall A e f, b e f -> A ⊨ e <-> A ⊨ f)) ->
  (forall e f, b e f -> str_positive e <-> str_positive f) ->
  (exists e f, ~ (b ⊢ e ≡ f) /\ (str_positive⊗s)^2 e f) ->
  (r,a) << (str_positive⊗s,b).
Proof.
  intros h1 h2 h8 h4 (x&y&h5&h6&h7).
  eapply technical with (e0 := x) (f0:= y) (r := r^2)
                                  (s := po ∧ (s^2));eauto.
  - intros e f;unfold rel; case (h2 e);case (h2 f);tauto.
  - clear h1 x y h5 h6 h7 a r h2.
    unfold po,and_pred,and_rel,rel;intros e f ((h1&h2)&h5&h6).
    cut ((str_positive e /\ str_positive f) \/ 
         (e == f) \/ ~ (b ⊢ e ≡ f));[intuition|].
    rewrite pos_disj in *.
    apply eq_str_positive in h4.
    case h1 as [h1|[->|(A&->)]];simpl in *;try tauto;
    case h2 as [h2|[->|(B&->)]];simpl in *;try tauto.
    -- right;right;intro h;apply h4 in h;intuition.
    -- right;right;intro h;apply h4 in h;intuition.
    -- right;right;intro h;apply h4 in h;intuition.
    -- case h8 as [h8|h8].
       --- apply h8 in h6;simpl in h6;auto.
       --- right;right;intro h;apply (eq_type h8 B) in h;firstorder.
    -- right;right;intro h;apply h4 in h;intuition.
    -- case h8 as [h8|h8].
       --- apply h8 in h5;simpl in h5;auto.
       --- right;right;intro h;apply (eq_type h8 A) in h;firstorder.
    -- case (equiv_lst_b_spec A B);auto.
       intro n;right;right;intro;apply n.
       case h8 as [h8|h8].
       --- apply h8 in h5;simpl in h5;tauto.
       --- pose proof (eq_type h8 A H) as false;
         simpl in false;rewrite <-false;reflexivity.
Qed.

(** * Transfer morphisms. *)
Lemma NKA_NKA_p : rm0 ⊩ (np∧ut,NKA) → (np∧ut,NKA_p).
Proof.
  intros e f ((N1&N2)&(U1&U2)). 
  unfold ut, np, and_rel;repeat split_conj;
  auto using untyped_rm0, no_perm_rm0.
  apply untyped_equiv_rm0;auto.
  - apply NKA_untyped.
  - apply NKA_eq0.
  - firstorder.
  - apply NKA_p_rm0.
  - apply rm0_zero_eq_u.
Qed.
Lemma NKA_incl_NKA_p : rm0 ⊩ (np∧ut,NKA) → (po∧(np∧ut),NKA_p).
Proof. apply (augm_transfer2 NKA_NKA_p positive_rm0). Qed.
Lemma NKA_incl_NKA_p_nm :
  rm0 ⊩ (mlt∧(np∧ut),NKA) → (mlt∧(po∧(np∧ut)),NKA_p). 
Proof. apply (augm_transfer NKA_incl_NKA_p multi_rm0). Qed.
Lemma NKA_incl_NKA_p_ns :
  rm0 ⊩ (sg∧(np∧ut),NKA) → (sg∧(po∧(np∧ut)),NKA_p).  
Proof. apply (augm_transfer NKA_incl_NKA_p single_rm0). Qed.
  
Lemma NKA_p_incl_NKA : rm0 ⊩ (sp∧(np∧ut),NKA_p) → (np∧ut,NKA).
Proof.
  intros e f h; pose proof (str_pos_rev_rm0 NKA_NKA_p h);firstorder.
Qed.
Lemma NKA_p_incl_NKA_nm :
  rm0 ⊩ (mlt∧(sp∧(np∧ut)),NKA_p) → (mlt∧(np∧ut),NKA).
Proof. apply (augm_transfer NKA_p_incl_NKA multi_rm0). Qed.
Lemma NKA_p_incl_NKA_ns :
  rm0 ⊩ (sg∧(sp∧(np∧ut)),NKA_p) → (sg∧(np∧ut),NKA).
Proof. apply (augm_transfer NKA_p_incl_NKA single_rm0). Qed.
  
Lemma NKA_t_NKA_tp : rm0 ⊩ (np∧ty,NKA_t) → (np∧ty,NKA_tp).
Proof.
  intros e f ((N1&N2)&((A&T1)&B&T2)).
  unfold np,ty,typed,and_rel,rel;split_conj;
  eauto using typed_rm0, no_perm_rm0.
  apply typed_equiv_rm0 ;eauto.
  - apply NKA_t_has_type.
  - apply NKA_t_eq0.
  - firstorder.
  - apply NKA_tp_rm0.
  - apply rm0_zero_eq_t.
Qed.
Lemma NKA_t_incl_NKA_tp : rm0 ⊩ (np∧ty,NKA_t) → (po∧(np∧ty),NKA_tp).
Proof. apply (augm_transfer2 NKA_t_NKA_tp positive_rm0). Qed.
Lemma NKA_t_incl_NKA_tp_nm :
  rm0 ⊩ (mlt∧(np∧ty),NKA_t) → (mlt∧(po∧(np∧ty)),NKA_tp). 
Proof. apply (augm_transfer NKA_t_incl_NKA_tp multi_rm0). Qed.
Lemma NKA_t_incl_NKA_tp_ns :
  rm0 ⊩ (sg∧(np∧ty),NKA_t) → (sg∧(po∧(np∧ty)),NKA_tp). 
Proof. apply (augm_transfer NKA_t_incl_NKA_tp single_rm0). Qed.

Lemma NKA_tp_incl_NKA_t :
  rm0 ⊩ (sp∧(np∧ty),NKA_tp) → (np∧ty,NKA_t).
Proof.
  intros e f h; pose proof (str_pos_rev_rm0 NKA_t_NKA_tp h);
  unfold transfer_fun,po,ty in *;firstorder.
Qed.
Lemma NKA_tp_incl_NKA_tnm :
  rm0 ⊩ (mlt∧(sp∧(np∧ty)),NKA_tp) → (mlt∧(np∧ty),NKA_t).
Proof. apply (augm_transfer NKA_tp_incl_NKA_t multi_rm0). Qed.
Lemma NKA_tp_incl_NKA_tns :
  rm0 ⊩ (sg∧(sp∧(np∧ty)),NKA_tp) → (sg∧(np∧ty),NKA_t).
Proof. apply (augm_transfer NKA_tp_incl_NKA_t single_rm0). Qed.

Lemma NKA_θ_NKA_θp : rm0 ⊩ (ut,NKA_θ) → (ut,NKA_θp).
Proof.
  intros e f (U1&U2). 
  unfold ut;split_conj;auto using untyped_rm0.
  apply untyped_equiv_rm0;auto.
  - apply NKA_θ_untyped.
  - apply NKA_θ_eq0.
  - firstorder.
  - apply NKA_θp_rm0.
  - apply rm0_zero_eq_u.
Qed.
Lemma NKA_θ_incl_NKA_θp : rm0 ⊩ (ut,NKA_θ) → (po∧ut,NKA_θp).
Proof. apply (augm_transfer2 NKA_θ_NKA_θp positive_rm0). Qed.
Lemma NKA_θ_incl_NKA_θp_s : rm0 ⊩ (sg∧ut,NKA_θ) → (sg∧(po∧ut),NKA_θp).
Proof. apply (augm_transfer NKA_θ_incl_NKA_θp single_rm0). Qed.
Lemma NKA_θ_incl_NKA_θp_m :
  rm0 ⊩ (mlt∧ut,NKA_θ) → (mlt∧(po∧ut),NKA_θp).
Proof. apply (augm_transfer NKA_θ_incl_NKA_θp multi_rm0). Qed.

Lemma NKA_θp_incl_NKA_θ :
  rm0 ⊩ (sp∧ut,NKA_θp) → (ut,NKA_θ).
Proof.
  intros e f h; pose proof (str_pos_rev_rm0 NKA_θ_NKA_θp h);
  unfold transfer_fun,po,ut in *;firstorder.
Qed.
Lemma NKA_θp_incl_NKA_θ_s : rm0 ⊩ (sg∧(sp∧ut),NKA_θp) → (sg∧ut,NKA_θ).
Proof. apply (augm_transfer NKA_θp_incl_NKA_θ single_rm0). Qed.
Lemma NKA_θp_incl_NKA_θ_m :
  rm0 ⊩ (mlt∧(sp∧ut),NKA_θp) → (mlt∧ut,NKA_θ).
Proof. apply (augm_transfer NKA_θp_incl_NKA_θ multi_rm0). Qed.


Lemma NKA_θt_NKA_θtp : rm0 ⊩ (ty,NKA_θt) → (ty,NKA_θtp).
Proof.
  intros e f ((A&T1)&B&T2).
  unfold ty,rel,typed;split_conj;eauto using typed_rm0.
  apply typed_equiv_rm0;eauto.
  - apply NKA_θt_has_type.
  - apply NKA_θt_eq0.
  - firstorder.
  - apply NKA_θtp_rm0.
  - apply rm0_zero_eq_t. 
Qed.

Lemma NKA_θt_incl_NKA_θtp : rm0 ⊩ (ty,NKA_θt) → (po∧ty,NKA_θtp).
Proof. apply (augm_transfer2 NKA_θt_NKA_θtp positive_rm0). Qed.
Lemma NKA_θt_incl_NKA_θtp_s :
  rm0 ⊩ (sg∧ty,NKA_θt) → (sg∧(po∧ty),NKA_θtp).
Proof. apply (augm_transfer NKA_θt_incl_NKA_θtp single_rm0). Qed.
Lemma NKA_θt_incl_NKA_θtp_m :
  rm0 ⊩ (mlt∧ty,NKA_θt) → (mlt∧(po∧ty),NKA_θtp).
Proof. apply (augm_transfer NKA_θt_incl_NKA_θtp multi_rm0). Qed.

Lemma NKA_θtp_incl_NKA_θt :
  rm0 ⊩ (sp∧ty,NKA_θtp) → (ty,NKA_θt).
Proof.
  intros e f h; pose proof (str_pos_rev_rm0 NKA_θt_NKA_θtp h);
  unfold transfer_fun,po,ty in *;firstorder.
Qed.
Lemma NKA_θtp_incl_NKA_θt_s :
  rm0 ⊩ (sg∧(sp∧ty),NKA_θtp) → (sg∧ty,NKA_θt).
Proof. apply (augm_transfer NKA_θtp_incl_NKA_θt single_rm0). Qed.
Lemma NKA_θtp_incl_NKA_θt_m :
  rm0 ⊩ (mlt∧(sp∧ty),NKA_θtp) → (mlt∧ty,NKA_θt).
Proof. apply (augm_transfer NKA_θtp_incl_NKA_θt multi_rm0). Qed.

Lemma NKA_θp_NKA_θtp_fun :
  typed_vers ⊩ (Bs ut,NKA_θp) → (ty',NKA_θtp).
Proof.
  intros e f (B&U1&U2);split.
  exists (support e); split;[|rewrite B];
  apply typed_version_support_type; auto.
  split;intro π.
  - apply (@eq_typed_vers NKA_θp);try (repeat split;tauto).
    * apply NKAθ_typed_vers.
    * apply NKA_θp_untyped.
    * apply NKA_θp_support.
    * unfold NKA_θtp,stronger,or_rel;intuition.
  - rewrite <-(typed_version_erase U1),<-(typed_version_erase U2).
    apply (@eq_erase NKA_θtp).
    * apply NKAθ_erase.
    * apply NKA_θtp_has_type.
    * split;auto;exists (support e);split;auto;rewrite B;auto.
Qed.
Lemma NKA_θp_NKA_θtp_precise :
  typed_vers ⊩ (ut,NKA_θp) → (ty,NKA_θtp).
Proof.
  intros e f (U1&U2);pose proof (typed_version_support_type U1) as T1;
  pose proof (typed_version_support_type U2) as T2;split.
  - split;eauto.
  - case (equiv_lst_b_spec (support e) (support f));intro.
    -- apply (@NKA_θp_NKA_θtp_fun e f);split;auto.
    -- split;intro E;exfalso; apply n.
    * now apply ((eq_support NKA_θp_support) _ _ E) .
    * now apply (type_unicity T1);rewrite(eq_type NKA_θtp_has_type _ E).
Qed.
Lemma NKA_θtp_NKA_θp_fun :
  erase ⊩ (sp ∧ ty',NKA_θtp) →
  (sp ∧ ut,NKA_θp).
Proof.
  intros e f ((P1&P2)&(A&T1&T2));split.
  repeat split;auto using str_positive_erase, untyped_erase.
  apply incl_ty_ut;(try exists A);eauto using str_positive_positive.
  - apply NKA_θp_eq0.
  - apply NKAθ_erase.
  - apply NKAθ_typed_vers.
  - apply NKA_θtp_has_type.
  - apply NKA_θp_untyped.
  - apply NKA_θp_support.
  - firstorder.
Qed.
Lemma NKA_p_NKA_tp :
  typed_vers ⊩ (sp∧(np∧Bs ut),NKA_p) →
  (sp∧(np∧ty'), NKA_tp).
Proof.
  apply incl_p_fun;[|intros;repeat split;
                     apply str_positive_typed_vers;firstorder].
  intros e f ((n1&n2)&B & U1 & U2);split.
  repeat split;apply no_perm_typed_vers
               || apply single_typed_vers
               || (exists (support e);split;[|rewrite B]);auto.
  rewrite <-NKA_p'_NKA_p_eq, <-NKA_tp'_NKA_tp_eq;
    [|now firstorder;apply no_perm_typed_vers
                     || apply single_typed_vers
                     || eauto
     |firstorder].
  apply incl_ut_ty;eauto.
  - apply NKA_erase'.
  - apply NKA_typed_vers'.
  - apply NKA_tp'_has_type.
  - apply NKA_p'_untyped.
  - apply NKA_p'_support.
  - firstorder.
Qed.  
Lemma NKA_p_NKA_tp_precise :
  typed_vers ⊩ (sp∧(np∧ut),NKA_p) → (sp∧(np∧ty),NKA_tp).
Proof.
  intros e f ((S1&S2)&(P1&P2)&U1&U2);
  pose proof (typed_version_support_type U1) as T1;
  pose proof (typed_version_support_type U2) as T2;split.
  - now repeat split;eauto;
    apply str_positive_typed_vers||apply no_perm_typed_vers.
  - case (equiv_lst_b_spec (support e) (support f));intro.
    -- apply (@NKA_p_NKA_tp e f);firstorder.
    -- split;intro E;exfalso; apply n.
    * now apply ((eq_support NKA_p_support) _ _ E) .
    * now apply (type_unicity T1);rewrite(eq_type NKA_tp_has_type _ E).
Qed.
Lemma NKA_tp_NKA_p :
  erase ⊩ (sp∧(np∧ty'),NKA_tp) →
  (sp∧(np∧ut),NKA_p).
Proof.
  intros e f h;split.
  firstorder;eauto using untyped_erase,
  single_erase,no_perm_erase,str_positive_erase.
  rewrite <-NKA_p'_NKA_p_eq, <-NKA_tp'_NKA_tp_eq;
    [|firstorder| now firstorder;apply no_perm_erase
                                 || apply single_erase
                                 || apply untyped_erase
                                 || eauto].
  apply incl_ty_ut_np;try firstorder;
  try now apply str_positive_positive.
  - apply NKA_erase'.
  - apply NKA_typed_vers'.
  - apply NKA_tp'_has_type.
  - apply NKA_p'_untyped.
  - apply NKA_p'_support.
Qed.
Lemma NKA_p_NKA_tp_s:
  typed_vers
    ⊩ (sg∧(sp ∧ (np ∧ Bs ut)), NKA_p) → (sg∧(sp ∧ (np ∧ ty')), NKA_tp).
Proof. apply (augm_transfer NKA_p_NKA_tp single_typed_vers). Qed.
Lemma NKA_θp_NKA_θtp_s:
  typed_vers ⊩ (sg∧(Bs ut), NKA_θp) → (sg∧ty', NKA_θtp).
Proof. apply (augm_transfer NKA_θp_NKA_θtp_fun single_typed_vers). Qed.
Lemma NKA_p_NKA_tp_m:
  typed_vers
    ⊩ (mlt∧(sp∧(np∧Bs ut)),NKA_p) → (mlt∧(sp ∧ (np ∧ ty')), NKA_tp).
Proof. apply (augm_transfer NKA_p_NKA_tp multi_typed_vers). Qed.
Lemma NKA_θp_NKA_θtp_m:
  typed_vers ⊩ (mlt∧(Bs ut), NKA_θp) → (mlt∧ty', NKA_θtp).
Proof. apply (augm_transfer NKA_θp_NKA_θtp_fun multi_typed_vers). Qed.

Lemma NKA_tp_NKA_p_s :
  erase
    ⊩ (sg ∧ (sp ∧ (np ∧ ty')), NKA_tp) → (sg ∧ (sp ∧ (np ∧ ut)),NKA_p).
Proof. apply (augm_transfer NKA_tp_NKA_p single_erase). Qed.
Lemma NKA_θtp_NKA_θp_s:
  erase ⊩ (sg ∧ (sp ∧ ty'), NKA_θtp) → (sg ∧ (sp ∧ ut), NKA_θp).
Proof. apply (augm_transfer NKA_θtp_NKA_θp_fun single_erase). Qed.
Lemma NKA_tp_NKA_p_m:
  erase ⊩ (mlt∧(sp∧(np∧ty')),NKA_tp) → (mlt ∧ (sp ∧ (np ∧ ut)), NKA_p).
Proof. apply (augm_transfer NKA_tp_NKA_p multi_erase). Qed.
Lemma NKA_θtp_NKA_θp_m:
  erase ⊩ (mlt ∧ (sp ∧ ty'), NKA_θtp) → (mlt ∧ (sp ∧ ut), NKA_θp).
Proof. apply (augm_transfer NKA_θtp_NKA_θp_fun multi_erase). Qed. 

Lemma NKA_θ_NKA_fun :
  rm_perm ∘ (apply_perm [])⊩(sp∧ut,NKA_θp)→ (np∧(sp∧ut),NKA_p).
Proof.
  intros e f ((P1&P2)&(U1&U2));repeat split;
  unfold Basics.compose; repeat rewrite <-single_rm_perm
                         || rewrite <-single_apply_perm
                         || rewrite <-untyped_rm_perm
                         || rewrite <-untyped_apply_perm
                         || rewrite <-str_positive_rm_perm
                         || rewrite <-str_positive_apply_perm
                         || apply no_perm_rm_perm;auto;
  intro E.
  *  assert (Eq : [] ≌ []) by reflexivity.
     apply (eq_apply_perm NKA_θp_NKA_p_rm_perm_apply_perm Eq E).
  * transitivity (rm_perm ([]•e));
    [|symmetry;transitivity (rm_perm ([]•f));[|symmetry]].
      assert (e=0 \/ str_positive e) as [->|P]
           by (destruct e;simpl;auto);simpl;auto.
         etransitivity;[|eapply stronger_ax_proof;
                          [|apply eq_rm_perm_ut_pos]];
         simpl;firstorder;symmetry;eql;auto 10.
      assert (f=0 \/ str_positive f) as [->|P]
           by (destruct f;simpl;auto);simpl;auto.
         etransitivity;[|eapply stronger_ax_proof;
                          [|apply eq_rm_perm_ut_pos]];
         simpl;firstorder;symmetry;eql;auto 10.
      apply NKA'_NKA_θp;try apply switch_NKA'_NKA_p;auto;try split;
         apply positive_rm_perm,
         positive_apply_perm,str_positive_positive
         ||rewrite <-untyped_rm_perm,<-untyped_apply_perm
         ||rewrite <-single_rm_perm,<-single_apply_perm;try tauto. 
Qed.
Lemma NKA_θt_NKA_t_fun :
  rm_perm ∘ (apply_perm [])⊩(sp∧ty,NKA_θtp)→ (np∧(sp∧ty),NKA_tp).
Proof.
  intros e f ((P1&P2)&(A&T1)&B&T2);repeat split;
  unfold Basics.compose; repeat rewrite <-single_rm_perm
                         || rewrite <-single_apply_perm
                         || rewrite <-str_positive_rm_perm
                         || rewrite <-str_positive_apply_perm
                         || apply no_perm_rm_perm;auto;
  try intro E.
  * exists A;apply has_type_rm_perm_apply_perm, has_type_apply_perm;
    now simpl_perm.
  * exists B;apply has_type_rm_perm_apply_perm, has_type_apply_perm;
    now simpl_perm.
  *  assert (Eq : [] ≌ []) by reflexivity.
     apply (eq_apply_perm NKA_θtp_NKA_tp_rm_perm_apply_perm Eq E).
  * transitivity (rm_perm ([]•e));
    [|symmetry;transitivity (rm_perm ([]•f));[|symmetry]].
    - assert (e=0 \/ str_positive e) as [->|P]
          by (destruct e;simpl;auto);simpl;auto.
      etransitivity;[|eapply stronger_ax_proof;
                       [|apply eq_rm_perm_pos]];
      simpl;firstorder;symmetry;eql;auto 10.
    - assert (f=0 \/ str_positive f) as [->|P]
          by (destruct f;simpl;auto);simpl;auto.
      etransitivity;[|eapply stronger_ax_proof;
                       [|apply eq_rm_perm_pos]];
      simpl;firstorder;symmetry;eql;auto 10.
    - apply NKA_tp_NKA_θtp;auto.
      unfold ty,sp,np; repeat split
                       ||rewrite <-single_rm_perm
                       || rewrite <-single_apply_perm
                       || rewrite <-str_positive_rm_perm
                       || rewrite <-str_positive_apply_perm
                       || apply no_perm_rm_perm;auto;
      [exists A|exists B];
      apply has_type_rm_perm_apply_perm, has_type_apply_perm;
      now simpl_perm.
Qed.
Lemma NKA_θ_NKA_s :
  rm_perm ∘ (apply_perm [])
          ⊩(sg∧(sp∧ut),NKA_θp)→ (sg∧(np∧(sp∧ut)),NKA_p).
Proof. apply (augm_transfer NKA_θ_NKA_fun).
       intro;unfold Basics.compose;
       rewrite <-single_rm_perm,<-single_apply_perm;auto.
Qed.
Lemma NKA_θ_NKA_m :
  rm_perm ∘ (apply_perm [])
          ⊩(mlt∧(sp∧ut),NKA_θp)→ (mlt∧(np∧(sp∧ut)),NKA_p).
Proof. apply (augm_transfer NKA_θ_NKA_fun).
       intro;unfold Basics.compose;
       rewrite <-multi_rm_perm,<-multi_apply_perm;auto.
Qed.
Lemma NKA_NKA_θ_fun :
  (fun e => e) ⊩ (np∧(sp∧ut),NKA_p)→(sp∧ut,NKA_θp).
Proof.
  intros e f ((N1&N2)&(P1&P2)&U1&U2).
  unfold sp,sg, po, ut,rel,and_rel;split_conj;auto.
  split;intro π.
  apply NKA'_NKA_θp;auto using str_positive_positive;
  apply switch_NKA'_NKA_p;auto.
  assert (E : [] ≌ []) by reflexivity.
  eapply (eq_apply_perm NKA_θp_NKA_p_rm_perm_apply_perm E) in π.
  assert (NKA_p ⊢ rm_perm e ≡ rm_perm f) as π'.
  etransitivity;[symmetry|etransitivity;[apply π|]];
  apply eq_expr_eq,eq_expr_rm_perm,apply_perm_nil;auto.
  now rewrite no_perm_rm_perm_id in *;rewrite N1,N2 in π';eauto.
Qed.
Lemma NKA_t_NKA_θt_fun :
  (fun e => e) ⊩ (np∧(sp∧ty),NKA_tp)→(sp∧ty,NKA_θtp).
Proof.
  intros e f ((N1&N2)&(P1&P2)&(A&T1)&(B&T2)).
  unfold sp,sg, po, ty,rel,and_rel;split_conj;eauto.
  split;intro π.
  apply NKA_tp_NKA_θtp;auto using str_positive_positive.
  repeat split;eauto.
  assert (E : [] ≌ []) by reflexivity.
  eapply (eq_apply_perm NKA_θtp_NKA_tp_rm_perm_apply_perm E) in π.
  assert (NKA_tp ⊢ rm_perm e ≡ rm_perm f) as π'.
  etransitivity;[symmetry|etransitivity;[apply π|]];
  apply eq_expr_eq,eq_expr_rm_perm,apply_perm_nil;auto.
  now rewrite no_perm_rm_perm_id in *;rewrite N1,N2 in π';eauto.
Qed.
Lemma NKA_NKA_θ_s :
  (fun e => e) ⊩ (sg∧(np∧(sp∧ut)),NKA_p)→(sg∧(sp∧ut),NKA_θp).
Proof. apply (augm_transfer NKA_NKA_θ_fun);auto. Qed.
Lemma NKA_NKA_θ_m :
  (fun e => e) ⊩ (mlt∧(np∧(sp∧ut)),NKA_p)→(mlt∧(sp∧ut),NKA_θp).
Proof. apply (augm_transfer NKA_NKA_θ_fun);auto. Qed.

Lemma single_multi_ut :
  single_to_multi ⊩ (sg∧(sp∧ut),NKA_θp) →
  (mlt∧(sp∧ut),NKA_θp).
Proof.
  unfold sp,sg,po,mlt,ut,rel,and_rel;intros e f (S&P&U);repeat split;
  try apply mlt_single_to_multi
  || apply ut_ut_single_to_multi
  || apply poa_poa_single_to_multi;try tauto;intro E.
  apply Proper_ut_sg_eq;repeat split;tauto.
  rewrite <- ut_multi_to_single_single_to_multi;try tauto.
  rewrite <-(@ut_multi_to_single_single_to_multi f);try tauto.
  apply Proper_ut_img_eq;repeat split;auto;
  apply sg_image_single_to_multi || apply ut_ut_single_to_multi;tauto.
Qed.
Lemma single_multi_ty :
  single_to_multi ⊩ (sg∧(sp∧ty),NKA_θtp) →
  (mlt∧(sp∧ty),NKA_θtp).
Proof.
  unfold sp,sg,po,mlt,ty,rel,and_rel;intros e f (S&P&U);repeat split;
  try apply mlt_single_to_multi
  || apply ty_ty_single_to_multi
  || apply poa_poa_single_to_multi;try tauto;intro E.
  apply Proper_ty_sg_eq;repeat split;tauto.
  rewrite <-ty_multi_to_single_single_to_multi;try tauto.
  rewrite <-(@ty_multi_to_single_single_to_multi f);try tauto.
  apply Proper_ty_img_eq;repeat split;auto;
  apply sg_image_single_to_multi || apply ty_ty_single_to_multi;tauto.
Qed.
    
(** * Order between theories. *)

(** ** Expressions without explicit permutations.*)
Lemma NKAnspu_is_model_NKAnsu : NKAnspu ⊲ NKAnsu.
Proof.
  exists rm0;eapply transfer_fun_proper_iff_2;
  [|eapply transfer_fun_proper_iff_1;[|apply NKA_p_incl_NKA_ns]];
  firstorder.
Qed.
  
(** For untyped and typed expressions, zero may be eliminated. *)
Lemma NKAnsu_NKAnspu_eq : NKAnsu ≡≡ NKAnspu.
Proof.
  split.
  - transitivity (str_positive ⊗ no_perm ⊗ single ⊗ untyped, NKA_p).
    unfold NKAnsu;eapply po_to_str_rm0.
    -- eapply transfer_fun_proper_iff_2;
      [|eapply transfer_fun_proper_iff_1;
         [|apply incl_p_fun with (p:= np)(q:=np);auto;
           [apply incl_p_fun with (p:= sg)(q:=sg);auto;
            [apply NKA_incl_NKA_p|]|]]]; firstorder;
      auto using single_rm0, no_perm_rm0.
    -- intros e; case (decidable_no_perm e);case (decidable_single e);
       case (decidable_untyped e);firstorder.
    -- firstorder.
    -- apply NKA_p_str_positive.
    -- exists (vat α);eexists;split;[apply at_diff_un_NKA_p|];
       firstorder.
    -- firstorder.
  - apply transfer_fun_included_th;apply NKAnspu_is_model_NKAnsu.
Qed.
Lemma NKAnmpu_is_model_NKAnmu : NKAnmpu ⊲ NKAnmu.
Proof.
  exists rm0;eapply transfer_fun_proper_iff_2;
  [|eapply transfer_fun_proper_iff_1;[|apply NKA_p_incl_NKA_nm]];
  firstorder.
Qed.
Lemma NKAnmu_NKAnmpu_eq : NKAnmu ≡≡ NKAnmpu.
Proof.
  split.
  - transitivity (str_positive ⊗ no_perm ⊗ multi ⊗ untyped, NKA_p).
    unfold NKAnsu;eapply po_to_str_rm0.
    -- eapply transfer_fun_proper_iff_2;
      [|eapply transfer_fun_proper_iff_1;
         [|apply incl_p_fun with (p:= np)(q:=np);auto;
           [apply incl_p_fun with (p:= mlt)(q:=mlt);auto;
            [apply NKA_incl_NKA_p|]|]]]; firstorder;
      auto using multi_rm0, no_perm_rm0.
    -- intros e; case (decidable_no_perm e);case (decidable_multi e);
       case (decidable_untyped e);firstorder.
    -- firstorder.
    -- apply NKA_p_str_positive.
    -- exists (var λ);eexists;split;[apply le_diff_un_NKA_p|];
       try rewrite τ_λ;firstorder.
    -- firstorder.
  - apply transfer_fun_included_th,NKAnmpu_is_model_NKAnmu.
Qed.
Lemma NKAnspt_is_model_NKAnst : NKAnspt ⊲ NKAnst.
Proof.
  exists rm0;eapply transfer_fun_proper_iff_2;
  [|eapply transfer_fun_proper_iff_1;[|apply NKA_tp_incl_NKA_tns]];
  firstorder.
Qed.
Lemma NKAnst_NKAnspt_eq : NKAnst ≡≡ NKAnspt.
Proof.
  split.
  - transitivity (str_positive ⊗ no_perm ⊗ single ⊗ typed, NKA_tp).
    unfold NKAnst;eapply po_to_str_rm0.
    -- eapply transfer_fun_proper_iff_2;
      [|eapply transfer_fun_proper_iff_1;
         [|apply incl_p_fun with (p:= np)(q:=np);auto;
           [apply incl_p_fun with (p:= sg)(q:=sg);auto;
            [apply NKA_t_incl_NKA_tp|]|]]]; firstorder;
      auto using single_rm0, no_perm_rm0.
    -- intros e; case (decidable_no_perm e);case (decidable_single e);
       case (decidable_typed e);firstorder.
    -- right;apply NKA_tp_has_type.
    -- apply NKA_tp_str_positive.
    -- exists (id [α]);eexists;split;[apply id_a_diff_un_NKA_tp|];
       firstorder;eexists;eauto.
    -- firstorder.
  - apply transfer_fun_included_th,NKAnspt_is_model_NKAnst.
Qed.
Lemma NKAnmpt_is_model_NKAnmt : NKAnmpt ⊲ NKAnmt.
Proof.
  exists rm0;eapply transfer_fun_proper_iff_2;
  [|eapply transfer_fun_proper_iff_1;[|apply NKA_tp_incl_NKA_tnm]];
  firstorder.
Qed.
Lemma NKAnmt_NKAnmpt_eq : NKAnmt ≡≡ NKAnmpt.
Proof.
  split.
  - transitivity (str_positive ⊗ no_perm ⊗ multi ⊗ typed, NKA_tp).
    unfold NKAnst;eapply po_to_str_rm0.
    -- eapply transfer_fun_proper_iff_2;
      [|eapply transfer_fun_proper_iff_1;
         [|apply incl_p_fun with (p:= np)(q:=np);auto;
           [apply incl_p_fun with (p:= mlt)(q:=mlt);auto;
            [apply NKA_t_incl_NKA_tp|]|]]]; firstorder;
      auto using multi_rm0, no_perm_rm0.
    -- intros e; case (decidable_no_perm e);case (decidable_multi e);
       case (decidable_typed e);firstorder.
    -- right;apply NKA_tp_has_type.
    -- apply NKA_tp_str_positive.
    -- exists (id [α]);eexists;split;[apply id_a_diff_un_NKA_tp|];
       firstorder;eexists;eauto.
    -- firstorder.
  - apply transfer_fun_included_th,NKAnmpt_is_model_NKAnmt.
Qed.
(** Typed and untyped expressions are equivalent. *)
Lemma NKAnmpu_is_model_NKAnmpt : NKAnmpu ⊲ NKAnmpt.
Proof.
  exists typed_vers;eapply transfer_fun_proper_iff_2;
  [|eapply transfer_fun_proper_iff_1;
     [|apply (augm_transfer NKA_p_NKA_tp_precise multi_typed_vers)]];
  firstorder.
Qed.

Lemma NKAnspu_is_model_NKAnspt : NKAnspu ⊲ NKAnspt.
Proof.
  exists typed_vers;eapply transfer_fun_proper_iff_2;
  [|eapply transfer_fun_proper_iff_1;
     [|apply (augm_transfer NKA_p_NKA_tp_precise single_typed_vers)]];
  firstorder.
Qed.
Lemma NKAnspu_NKAnspt_eq : NKAnspu ≡≡ NKAnspt.
Proof.
  split.
  - apply transfer_fun_included_th,NKAnspu_is_model_NKAnspt.
  - eapply technical with (r:=sg ∧ (sp ∧ (np ∧ ty')))
                            (s:=sg ∧ (sp ∧ (np ∧ ut))).
    * apply incl_p_fun;[apply NKA_tp_NKA_p|].
      now intros x y ((?&?)&_);split;apply single_erase.
    * unfold  sp,po, np,sg,rel,and_rel, and_pred;intros e f.
       destruct (decidable_str_positive e);try tauto;
       destruct (decidable_str_positive f);try tauto;
       destruct (decidable_single e);try tauto;
       destruct (decidable_single f);try tauto;
       destruct (decidable_no_perm e);try tauto;
       destruct (decidable_no_perm f);try tauto.
       destruct (ty'_or_diff_NKA_tp e f);tauto.
    * apply (@at_diff_un_NKA_p α).
    * firstorder.
    * firstorder.
    * firstorder.
Qed.
Lemma NKAnmpu_NKAnmpt_eq : NKAnmpu ≡≡ NKAnmpt.
Proof.
  split.
  - apply transfer_fun_included_th,NKAnmpu_is_model_NKAnmpt.
  - eapply technical with (r:=mlt ∧ (sp ∧ (np ∧ ty')))
                            (s:=mlt ∧ (sp ∧ (np ∧ ut))).
    * apply incl_p_fun;[apply NKA_tp_NKA_p|].
      now intros x y ((?&?)&_);split;apply multi_erase.
    * unfold  sp, np,mlt,rel,and_rel, and_pred;intros e f.
       destruct (decidable_str_positive e);try tauto;
       destruct (decidable_str_positive f);try tauto;
       destruct (decidable_multi e);try tauto;
       destruct (decidable_multi f);try tauto;
       destruct (decidable_no_perm e);try tauto;
       destruct (decidable_no_perm f);try tauto.
       destruct (ty'_or_diff_NKA_tp e f);tauto.
    * apply (@le_diff_un_NKA_p λ);rewrite τ_λ;discriminate.
    * firstorder.
    * firstorder.
    * firstorder.
Qed.
(** ** For positive expressions, it is equivalent to have 
 explicit and applied permutations.  *)
Lemma NKAnspu_is_model_NKAspu : NKAnspu ⊲ NKAspu.
Proof.
  eexists;eapply transfer_fun_proper_iff_2;
  [|eapply transfer_fun_proper_iff_1;
     [|apply (@augm_transfer single);[apply NKA_NKA_θ_fun|]]];
  firstorder.
Qed.
Lemma NKAspu_is_model_NKAnspu : NKAspu ⊲ NKAnspu.
Proof.
  eexists;eapply transfer_fun_proper_iff_2;
  [|eapply transfer_fun_proper_iff_1;
     [|apply (@augm_transfer single);[apply NKA_θ_NKA_fun|]]];
  unfold Basics.compose; firstorder;
  rewrite<- single_rm_perm,<-single_apply_perm;auto.
Qed.
Lemma NKAnspu_NKAspu_eq : NKAnspu ≡≡ NKAspu.
Proof.
  split.
  - apply transfer_fun_included_th,NKAnspu_is_model_NKAspu.
  - apply transfer_fun_included_th,NKAspu_is_model_NKAnspu.
Qed.
Lemma NKAnmpu_is_model_NKAmpu : NKAnmpu ⊲ NKAmpu.
Proof.
  eexists;eapply transfer_fun_proper_iff_2;
  [|eapply transfer_fun_proper_iff_1;
     [|apply (@augm_transfer multi);[apply NKA_NKA_θ_fun|]]];
  firstorder.
Qed.
Lemma NKAmpu_is_model_NKAnmpu : NKAmpu ⊲ NKAnmpu.
Proof.
  eexists;eapply transfer_fun_proper_iff_2;
  [|eapply transfer_fun_proper_iff_1;
     [|apply (@augm_transfer multi);[apply NKA_θ_NKA_fun|]]];
  unfold Basics.compose; firstorder;
  rewrite<- multi_rm_perm,<-multi_apply_perm;auto.
Qed.
Lemma NKAnmpu_NKAmpu_eq : NKAnmpu ≡≡ NKAmpu.
Proof.
  split.
  - apply transfer_fun_included_th,NKAnmpu_is_model_NKAmpu.
  - apply transfer_fun_included_th,NKAmpu_is_model_NKAnmpu.
Qed.

Lemma NKAnspt_is_model_NKAspt : NKAnspt ⊲ NKAspt.
Proof.
  eexists;eapply transfer_fun_proper_iff_2;
  [|eapply transfer_fun_proper_iff_1;
     [|apply (@augm_transfer single);[apply NKA_t_NKA_θt_fun|]]];
  firstorder.
Qed.
Lemma NKAspt_is_model_NKAnspt : NKAspt ⊲ NKAnspt.
Proof.
  eexists;eapply transfer_fun_proper_iff_2;
  [|eapply transfer_fun_proper_iff_1;
     [|apply (@augm_transfer single);[apply NKA_θt_NKA_t_fun|]]];
  unfold Basics.compose; firstorder;
  rewrite<- single_rm_perm,<-single_apply_perm;auto.
Qed.
Lemma NKAnspt_NKAspt_eq : NKAnspt ≡≡ NKAspt.
Proof.
  split.
  - apply transfer_fun_included_th,NKAnspt_is_model_NKAspt.
  - apply transfer_fun_included_th,NKAspt_is_model_NKAnspt.
Qed.
Lemma NKAnmpt_is_model_NKAmpt : NKAnmpt ⊲ NKAmpt.
Proof.
  eexists;eapply transfer_fun_proper_iff_2;
  [|eapply transfer_fun_proper_iff_1;
     [|apply (@augm_transfer multi);[apply NKA_t_NKA_θt_fun|]]];
  firstorder.
Qed.
Lemma NKAmpt_is_model_NKAnmpt : NKAmpt ⊲ NKAnmpt.
Proof.
  eexists;eapply transfer_fun_proper_iff_2;
  [|eapply transfer_fun_proper_iff_1;
     [|apply (@augm_transfer multi);[apply NKA_θt_NKA_t_fun|]]];
  unfold Basics.compose; firstorder;
  rewrite<- multi_rm_perm,<-multi_apply_perm;auto.
Qed.
Lemma NKAnmpt_NKAmpt_eq : NKAnmpt ≡≡ NKAmpt.
Proof.
  split.
  - apply transfer_fun_included_th,NKAnmpt_is_model_NKAmpt.
  - apply transfer_fun_included_th,NKAmpt_is_model_NKAnmpt.
Qed.

(** ** Expression with explicit permutations over atoms. *)
(** Zero may be eliminated both in the untyped and typed setting. *)
Lemma NKAspu_is_model_NKAsu : NKAspu ⊲ NKAsu.
Proof.
  exists rm0;eapply transfer_fun_proper_iff_2;
  [|eapply transfer_fun_proper_iff_1;[|apply NKA_θp_incl_NKA_θ_s]];
  firstorder.
Qed.
Lemma NKAmpu_is_model_NKAmu : NKAmpu ⊲ NKAmu.
Proof.
  exists rm0;eapply transfer_fun_proper_iff_2;
  [|eapply transfer_fun_proper_iff_1;[|apply NKA_θp_incl_NKA_θ_m]];
  firstorder.
Qed.
Lemma NKAspt_is_model_NKAst : NKAspt ⊲ NKAst.
Proof.
  exists rm0;eapply transfer_fun_proper_iff_2;
  [|eapply transfer_fun_proper_iff_1;[|apply NKA_θtp_incl_NKA_θt_s]];
  firstorder.
Qed.
Lemma NKAmpt_is_model_NKAmt : NKAmpt ⊲ NKAmt.
Proof.
  exists rm0;eapply transfer_fun_proper_iff_2;
  [|eapply transfer_fun_proper_iff_1;[|apply NKA_θtp_incl_NKA_θt_m]];
  firstorder.
Qed.
Lemma NKAspu_is_model_NKAspt : NKAspu ⊲ NKAspt.
Proof.
  exists typed_vers;eapply transfer_fun_proper_iff_2;
  [|eapply transfer_fun_proper_iff_1;
     [|eapply (@augm_transfer (single⊗str_positive));
        [apply NKA_θp_NKA_θtp_precise|unfold and_pred;intros;split;
           apply single_typed_vers||apply str_positive_typed_vers]]];
  firstorder.
Qed.
Lemma NKAmpu_is_model_NKAmpt : NKAmpu ⊲ NKAmpt.
Proof.
  exists typed_vers;eapply transfer_fun_proper_iff_2;
  [|eapply transfer_fun_proper_iff_1;
     [|eapply (@augm_transfer (multi⊗str_positive));
        [apply NKA_θp_NKA_θtp_precise|unfold and_pred;intros;split;
           apply multi_typed_vers||apply str_positive_typed_vers]]];
  firstorder.
Qed.

Lemma NKAsu_NKAspu_eq : NKAsu ≡≡ NKAspu.
Proof.
  split.
  - transitivity (str_positive ⊗ single ⊗ untyped, NKA_θp).
    unfold NKAnsu;eapply po_to_str_rm0.
    -- eapply transfer_fun_proper_iff_2;
      [|eapply transfer_fun_proper_iff_1;
         [|apply incl_p_fun with (p:= sg)(q:=sg);auto;
           [apply NKA_θ_incl_NKA_θp|]]]; firstorder;
      auto using single_rm0, no_perm_rm0.
    -- intros e; case (decidable_single e);
       case (decidable_untyped e);firstorder.
    -- firstorder.
    -- apply NKA_θp_str_positive.
    -- exists (vat α);eexists;split;[apply at_diff_un_NKA_θp|];
       firstorder.
    -- firstorder.
  - apply transfer_fun_included_th,NKAspu_is_model_NKAsu.
Qed.
Lemma NKAst_NKAspt_eq : NKAst ≡≡ NKAspt.
Proof.
  split.
  - transitivity (str_positive ⊗ single ⊗ typed, NKA_θtp).
    unfold NKAst;eapply po_to_str_rm0.
    -- eapply transfer_fun_proper_iff_2;
      [|eapply transfer_fun_proper_iff_1;
         [|apply incl_p_fun with (p:= sg)(q:=sg);auto;
           [apply NKA_θt_incl_NKA_θtp|]]]; firstorder;
      auto using single_rm0, no_perm_rm0.
    -- intros e; case (decidable_single e);
       case (decidable_typed e);firstorder.
    -- right;apply NKA_θtp_has_type.
    -- apply NKA_θtp_str_positive.
    -- exists (id [α]);eexists;split;[apply id_a_diff_un_NKA_θtp|];
       firstorder;eexists;eauto.
    -- firstorder.
  - apply transfer_fun_included_th,NKAspt_is_model_NKAst.
Qed.
(** Typed and untyped expressions are equivalent. *)
Lemma NKAspu_NKAspt_eq : NKAspu ≡≡ NKAspt. 
Proof.
  split.
  - apply transfer_fun_included_th,NKAspu_is_model_NKAspt.
  - eapply technical with (r:=sg ∧ (sp ∧ ty'))
                            (s:=sg ∧ (sp ∧ ut)).
    * eapply (incl_p_fun NKA_θtp_NKA_θp_fun).
       intros; split;apply single_erase; firstorder.
    * unfold and_pred,sp,sg,rel,and_rel;intros e f;simpl;
       destruct (decidable_str_positive e);try tauto;
       destruct (decidable_str_positive f);try tauto;
       destruct (decidable_single e);try tauto;
       destruct (decidable_single f);try tauto;
       destruct (ty'_or_diff_NKA_θtp e f);try tauto.
    * apply (@at_diff_un_NKA_θp α).
    * repeat split;auto.
    * repeat split;auto.
    * firstorder.
Qed.

(** ** Expressions over letters (with explicit permutations). *)
(** In this theory zero may be eliminated as well. *)
Lemma NKAmu_NKAmpu_eq : NKAmu ≡≡ NKAmpu.
Proof.
  split.
  - transitivity (str_positive ⊗ multi ⊗ untyped, NKA_θp).
    unfold NKAmu;eapply po_to_str_rm0.
    -- eapply transfer_fun_proper_iff_2;
      [|eapply transfer_fun_proper_iff_1;
         [|apply incl_p_fun with (p:= mlt)(q:=mlt);auto;
           [apply NKA_θ_incl_NKA_θp|]]]; firstorder;
      auto using multi_rm0, no_perm_rm0.
    -- intros e; case (decidable_multi e);
       case (decidable_untyped e);firstorder.
    -- firstorder.
    -- apply NKA_θp_str_positive.
    -- exists (var λ);eexists;split;[apply le_diff_un_NKA_θp|];
       firstorder;rewrite τ_λ;firstorder.
    -- firstorder.
  - apply transfer_fun_included_th,NKAmpu_is_model_NKAmu. 
Qed.
Lemma NKAmt_NKAmpt_eq : NKAmt ≡≡ NKAmpt.
Proof.
  split.
  - transitivity (str_positive ⊗ multi ⊗ typed, NKA_θtp).
    unfold NKAst;eapply po_to_str_rm0.
    -- eapply transfer_fun_proper_iff_2;
      [|eapply transfer_fun_proper_iff_1;
         [|apply incl_p_fun with (p:= mlt)(q:=mlt);auto;
           [apply NKA_θt_incl_NKA_θtp|]]]; firstorder;
      auto using multi_rm0, no_perm_rm0.
    -- intros e; case (decidable_multi e);
       case (decidable_typed e);firstorder.
    -- right;apply NKA_θtp_has_type.
    -- apply NKA_θtp_str_positive.
    -- exists (id [α]);eexists;split;[apply id_a_diff_un_NKA_θtp|];
       firstorder;eexists;eauto.
    -- firstorder.
  - apply transfer_fun_included_th,NKAmpt_is_model_NKAmt.
Qed.
(** Typed and untyped versions of the theory are also equivalent. *)
Lemma NKAmpu_NKAmpt_eq : NKAmpu ≡≡ NKAmpt. 
Proof.
  split.
  - apply transfer_fun_included_th,NKAmpu_is_model_NKAmpt.
  - eapply technical with (r:=mlt ∧ (sp ∧ ty'))
                            (s:=mlt ∧ (sp ∧ ut)).
    * eapply (incl_p_fun NKA_θtp_NKA_θp_fun).
       intros; split;apply multi_erase; firstorder.
    * unfold and_pred,sp,mlt,rel,and_rel;intros e f;simpl;
       destruct (decidable_str_positive e);try tauto;
       destruct (decidable_str_positive f);try tauto;
       destruct (decidable_multi e);try tauto;
       destruct (decidable_multi f);try tauto;
       destruct (ty'_or_diff_NKA_θtp e f);try tauto.
    * apply (@le_diff_un_NKA_θp λ);rewrite τ_λ;firstorder.
    * repeat split;auto.
    * repeat split;auto.
    * firstorder.
Qed.
(** ** The theory of expressions over atoms may be encoded in that of expressions over letters. *)
Theorem ut_Single_to_Multi : NKAspu ⊲ NKAmpu.
Proof.
  eexists;eapply transfer_fun_proper_iff_2;
  [|eapply transfer_fun_proper_iff_1;[|apply single_multi_ut]];
  firstorder.
Qed.
Theorem ty_Single_to_Multi : NKAspt ⊲ NKAmpt.
Proof.
  eexists;eapply transfer_fun_proper_iff_2;
  [|eapply transfer_fun_proper_iff_1;[|apply single_multi_ty]];
  firstorder.
Qed.

