(** printing == %\ensuremath{\doteq}% *)
(** printing ⊨ %\ensuremath{\models}% *)
(** printing # %\ensuremath{\#}% *)
(** printing α %\ensuremath{\alpha}% *)
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

Require Export tools atoms atoms_ops equiv_lst expr.

(** Equality of expressions modulo equivalence of letters. *)
Reserved Notation " e == f " (at level 100).
Fixpoint eq_expr e f : Prop :=
  match (e,f) with
  | (0,0) | (1,1) => True
  | (var x,var y) => x ≃ y
  | (vat a,vat b) => a = b
  | (⊥ l,⊥ m) | (id l,id m) => l ≈ m
  | (e^*,f^* ) => (e == f)
  | (e+e',f+f') | (e·e',f·f') => (e == f)/\(e' == f')
  | (ν a e,ν b f) | (ω a e,ω b f) => (e == f)/\(a = b)
  | (‹p›e,‹q›f) => (e == f)/\(p ≌ q)
  | _ => False
  end
where " e == f " := (eq_expr e f).
Global Instance eq_expr_Equivalence : Equivalence eq_expr.
Proof.
  split.
  * intro e;induction e;simpl;auto.
  * intro e;induction e;intros f h;destruct f;simpl in *;firstorder.
  * intros e;induction e;intros f g h1 h2;destruct f;
    simpl in *;intuition;destruct g;simpl in *;intuition;
    try etransitivity;eauto.
Qed.

Lemma eq_expr_W l : Proper (eq_expr ==> eq_expr) (W l).
Proof.
  intro e;induction e;intro f;destruct f;simpl;intuition;
  induction l;simpl;firstorder.
Qed.
Lemma eq_expr_size : Proper (eq_expr ==> Logic.eq) size.
Proof. intro e;induction e;intro f;destruct f;simpl;intuition. Qed.
Lemma eq_expr_support : Proper (eq_expr ==> equiv_lst) support.
Proof.
  intro e;induction e;intro f;destruct f;simpl;intuition (subst;auto).
  - rewrite eq_letter_τ;eauto.
  - rewrite (IHe1 _ H0),(IHe2 _ H1);auto.
  - rewrite (IHe1 _ H0),(IHe2 _ H1);auto.
  - rewrite (IHe _ H0);auto.
  - rewrite (IHe _ H0),(equiv_perm_mact _ _ H1);auto.
Qed.
Lemma eq_expr_stage : Proper (eq_expr ==> equiv_lst) stage.
Proof.
  intro e;induction e;intro f;destruct f;simpl;intuition (subst;auto).
  - rewrite eq_letter_τ;eauto.
  - rewrite (IHe1 _ H0),(IHe2 _ H1);auto.
  - rewrite (IHe1 _ H0),(IHe2 _ H1);auto.
  - rewrite (IHe _ H0);auto.
  - rewrite (IHe _ H0);auto.
  - rewrite (IHe _ H0),(equiv_perm_mact _ _ H1);auto.
Qed.
Lemma eq_expr_eq0 : Proper (eq_expr ==> Logic.eq) eq0.
Proof.
  intro e;induction e;intro f;destruct f;simpl;intuition.
  - now rewrite (IHe1 _ H0), (IHe2 _ H1).
  - now rewrite (IHe1 _ H0), (IHe2 _ H1).
Qed.
Lemma eq_expr_eq0' : Proper (eq_expr ==> Logic.eq) eq0'.
Proof. intro e;induction e;intro f;destruct f;simpl;intuition. Qed.
Lemma eq_expr_get_bot e f :
  (e == f) ->
  (get_bot e = get_bot f) \/
   exists a b, get_bot e = inr a /\ get_bot f = inr b /\ a≈b.
Proof. revert f;induction e;intro f;destruct f;simpl;firstorder. Qed.
(** Boolean equality. *)
Fixpoint beq_expr e f : bool :=
  match (e,f) with
  | (0,0) | (1,1) => true
  | (var x,var y) => beq_letter x y
  | (vat a,vat b) => beq_atom a b
  | (⊥ l,⊥ m) | (id l,id m) => equiv_lst_b l m
  | (e^*,f^* ) => beq_expr e f
  | (e+e',f+f') | (e·e',f·f') => (beq_expr e f)&&(beq_expr e' f')
  | (ν a e,ν b f) | (ω a e,ω b f) => (beq_expr e f)&&(beq_atom a b)
  | (‹p›e,‹q›f) => (beq_expr e f)&&(bequiv_perm p q)
  | _ => false
  end.
Lemma beq_expr_spec e f : reflect (e==f) (beq_expr e f).
Proof.
  apply iff_reflect;revert f;induction e;intros g;destruct g;
  simpl;repeat simpl_bool;try firstorder congruence.
  - rewrite <-(reflect_iff _ _ (equiv_lst_b_spec _ _));
    firstorder congruence.
  - rewrite <-(reflect_iff _ _ (equiv_lst_b_spec _ _));
    firstorder congruence.
  - rewrite <-(reflect_iff _ _ (beq_letter_spec _ _));
    firstorder congruence.
  - rewrite <-(reflect_iff _ _ (beq_atom_refl _ _));
    firstorder congruence.
  - rewrite <-(reflect_iff _ _ (beq_atom_refl _ _)),<-IHe;
    firstorder congruence.
  - rewrite <-(reflect_iff _ _ (beq_atom_refl _ _)),<-IHe;
    firstorder congruence.
  - rewrite <-(reflect_iff _ _ (bequiv_perm_spec _ _)),<-IHe;
    firstorder congruence.
Qed.


(** * Typing jugements *)
Reserved Notation "A ⊨ e" (at level 30).
Fixpoint has_type A e : Prop :=
  match e with
  | ⊥ B | id B => A ≈ B
  | var a => A ≈ (τ a)
  | vat a => A ≈ [a]
  | (e + f) | (e · f) => A ⊨ e /\ A ⊨ f
  | (e^* ) => A ⊨ e
  | (‹p › e) => (p⋆ ⧓ A)⊨ e
  | (ν a e) => ~ In a A /\ (a::A) ⊨ e
  | (ω a e) => In a A /\ (rm a A)⊨ e
  | 0 | 1 => False
  end
where "A ⊨ e" := (has_type A e).
Definition typed e := (exists A, A ⊨ e).
Definition ty' e f := (exists A, A ⊨ e /\ A ⊨ f).
Definition ty := typed^2.
Hint Unfold has_type typed ty ty'.
Fixpoint typedb e :=
  match e with
  | ⊥ B | id B => true
  | var a => true
  | vat a => true
  | (e + f) | (e · f) =>
              typedb e && typedb f && (equiv_lst_b (stage e) (stage f))
  | (e^* ) | (‹_ › e) => typedb e
  | (ν a e) => inb a (stage e) && typedb e
  | (ω a e) => negb (inb a (stage e)) && typedb e
  | 0 | 1 => false
  end.


(** Key lemma for simplification. *)
Lemma ht_convert A B e :
  (A ≈ B) -> A ⊨ e -> B ⊨ e.
Proof.
  cut (forall e A, A ⊨ e -> forall B, (A ≈ B) -> B ⊨ e).
  intros h heq ha; eapply h;eauto.
  clear A B e;intros e;induction e;intros A ha C heq;simpl in *;eauto.
  split;[eapply IHe1|eapply IHe2];eauto;tauto.
  split;[eapply IHe1|eapply IHe2];eauto;tauto.
  split;[|eapply IHe;[apply ha|]];rewrite <-heq;tauto||auto.
  split;[|eapply IHe;[apply ha|]];rewrite <-heq;tauto||auto.
  eapply IHe;[apply ha|];rewrite <-heq;auto. 
Qed.
Hint Resolve ht_convert.
Global Instance ht_equiv e :
  Proper (equiv_lst ==> iff) (Basics.flip has_type e).
Proof. intros ? ? ?;split;intro; eapply ht_convert;eauto. Qed.
Global Instance ht_equiv_iff :
  Proper (equiv_lst ==> Logic.eq ==> iff) has_type.
Proof. intros ? ? ? ? ? ->;split;intro; eapply ht_convert;eauto. Qed.
Global Instance ht_bot_eq A :
  Proper (equiv_lst ==> iff) (fun l => A ⊨ (bot l)).
Proof. intros X Y F;split;intro h;simpl in *;eauto. Qed.
Global Instance ht_id_eq A :
  Proper (equiv_lst ==> iff) (fun l => A ⊨ (id l)).
Proof. intros X Y F;split;intro h;simpl in *; eauto. Qed.
Ltac prove_type := eapply ht_convert;[|now eauto].
(** If an expression can be typed, then its type is equivalent to its *)
(** stage. *)

Lemma stage_has_type e A : A ⊨ e -> A ≈ stage e.
Proof.
  revert A;induction e;intros A T; simpl in *;repeat simpl_hyp;
  try tauto.
  - rewrite <-(IHe1 _ h),<- (IHe2 _ h0),app_l_l;eauto.
  - rewrite <-IHe1,<-IHe2,app_l_l;eauto.
  - apply (IHe _ T).
  - rewrite <- (IHe _ h0);simpl.
    destruct (Atom_eq_dec a a) as [heq|heq];try tauto;clear heq.
    rewrite rm_id;auto.
  - rewrite <- (IHe _ h0);apply (rm_cons_eq' h).
  - rewrite <- (IHe _ T);simpl_perm;auto.
Qed.
Lemma typed_stage_has_type e : typed e -> stage e ⊨ e.
Proof. intros (A&T);rewrite <-(stage_has_type T);auto. Qed.

Lemma typedb_spec e : reflect (typed e) (typedb e).
Proof.
  unfold typed;induction e;simpl;try apply ReflectT||apply ReflectF;
  firstorder eauto with bool.
  - destruct IHe1 as [T|h];destruct IHe2 as [S|g];
    try apply typed_stage_has_type in T;
    try apply typed_stage_has_type in S;simpl;try apply ReflectF.
    * destruct (equiv_lst_b_spec (stage e1) (stage e2));simpl;
       try apply ReflectT||apply ReflectF.
       exists (stage e1);split;auto;rewrite e;auto.
       intros (A&h1&h2);apply n.
       rewrite <-(stage_has_type h1),<-(stage_has_type h2);auto.
    * intros (A&h1&h2);apply g;eauto.
    * intros (A&h1&h2);apply h;eauto.
    * intros (A&h1&h2);apply h;eauto.
  - destruct IHe1 as [T|h];destruct IHe2 as [S|g];
    try apply typed_stage_has_type in T;
    try apply typed_stage_has_type in S;simpl;try apply ReflectF.
    * destruct (equiv_lst_b_spec (stage e1) (stage e2));simpl;
       try apply ReflectT||apply ReflectF.
       exists (stage e1);split;auto;rewrite e;auto.
       intros (A&h1&h2);apply n.
       rewrite <-(stage_has_type h1),<-(stage_has_type h2);auto.
    * intros (A&h1&h2);apply g;eauto.
    * intros (A&h1&h2);apply h;eauto.
    * intros (A&h1&h2);apply h;eauto.
  - destruct IHe as [T|h];
    try apply typed_stage_has_type in T;simpl;try apply ReflectF.
    * case_ex a (stage e);simpl;try apply ReflectT||apply ReflectF.
       exists (rm a (stage e));split;simpl_in;
       intuition now rewrite <-(rm_cons_eq' h).
       now intros (A&h1&h2);apply h;rewrite <-(stage_has_type h2);left.
    * replace (_&&false) with false.
       apply ReflectF;intros (A&h1&h2);apply h;exists (a::A);auto.
       destruct (inb _ _);auto.
  - destruct IHe as [T|h];
    try apply typed_stage_has_type in T;simpl;try apply ReflectF.
    * case_ex a (stage e);simpl;try apply ReflectT||apply ReflectF.
       intros (A&h1&h2); assert (I:In a (rm a A))
         by (rewrite (stage_has_type h2);auto); simpl_in;tauto.
       exists (a :: (stage e));split;simpl_in;auto.
       intuition now rewrite (cons_rm_eq' h).
    * replace (_&&false) with false.
       apply ReflectF;intros (A&h1&h2);apply h;exists (rm a A);auto.
       destruct (inb _ _);auto.
  - destruct IHe as [T|h];try apply typed_stage_has_type in T;simpl;
    try apply ReflectF || apply ReflectT.
    * exists (p⧓stage e);simpl_perm;auto.
    * intros (A&T);apply h;eauto.
Qed.
Lemma decidable_typed e : typed e \/ ~ typed e.
Proof. destruct (reflect_dec _ _ (typedb_spec e));auto. Qed.
    
    
(** * Untyped expressions *)
Fixpoint untyped e : Prop :=
  match e with
  | var _ | vat _ | 1 | 0 => True
  | ⊥ _ | id _ | ω _ _ => False
  | ν _ e | ‹_› e | e ^* => untyped e
  | e + f | e · f => untyped e /\ untyped f
  end.
Definition ut := untyped^2.
Hint Unfold untyped ut.
Fixpoint untypedb e : bool :=
  match e with
  | var _ | vat _ | 1 | 0 => true
  | ⊥ _ | id _ | ω _ _ => false
  | ν _ e | ‹_› e | e ^* => untypedb e
  | e + f | e · f => untypedb e && untypedb f
  end.
Lemma untypedb_spec e : reflect (untyped e) (untypedb e).
Proof.
  apply iff_reflect.
  induction e;simpl;try simpl_bool;firstorder (auto with bool).
Qed.
Lemma decidable_untyped e : untyped e \/ ~ untyped e.
Proof. destruct (reflect_dec _ _ (untypedb_spec e));tauto. Qed.

(** * Strictly positive expressions *)
(** [str_positive] expressions do not contain [0] or [⊥] constructors. *)
Fixpoint str_positive e : Prop :=
  match e with
  | var _ | vat _ | 1 | id _ => True
  | ⊥ _ | 0 => False
  | ω _ e | ν _ e |‹ _ › e | e ^* => str_positive e
  | e + f | e · f => str_positive e /\ str_positive f
  end.
Fixpoint str_positiveb e : bool :=
  match e with
  | var _ | vat _ | 1 | id _ => true
  | ⊥ _ | 0 => false
  | ω _ e | ν _ e |‹ _ › e | e ^* => str_positiveb e
  | e + f | e · f => str_positiveb e && str_positiveb f
  end.
Definition sp := str_positive^2.
Lemma str_positiveb_spec e : reflect (str_positive e) (str_positiveb e).
Proof.
  apply iff_reflect.
  induction e;simpl;try simpl_bool;firstorder (auto with bool).
Qed.
Lemma decidable_str_positive :
  forall e, str_positive e \/ ~ str_positive e.
Proof. induction e;firstorder. Qed.

(** [positive] expressions are either equal to [0] or [⊥], or are *)
(** [str_positive]. *)
Definition positive e : Prop :=
  match e with
  | ⊥ _ | 0 => True
  | _ => str_positive e
  end.
Definition positiveb e : bool :=
  match e with
  | ⊥ _ | 0 => true
  | _ => str_positiveb e
  end.
(** [positiveb] is a boolean function testing the predicate *)
(** [positive].  *)
Lemma positiveb_spec e : reflect (positive e) (positiveb e).
Proof.
  pose proof (reflect_iff _ _ (str_positiveb_spec e)).
  apply iff_reflect;induction e;simpl;firstorder auto with bool.
Qed.
(** This entails the decidability of the predicate [positive]. *)
Lemma decidable_positive :
  forall e, positive e \/ ~ positive e.
Proof.
 intro e; destruct (decidable_str_positive e);destruct e;firstorder.
Qed.
Definition po := positive^2.
Hint Unfold positive str_positive po.
(** [str_positive] entails [positive]. *)
Lemma str_positive_positive e : str_positive e -> positive e.
Proof. induction e;simpl; tauto. Qed.

(** An expression is positive if and only if it is either strictly *)
(** positive or equal to [0] or some [⊥ A]. *)
Lemma pos_disj e :
  positive e <-> (str_positive e \/ e = 0 \/ (exists A, e = ⊥ A)).
Proof. destruct e;simpl;firstorder;eauto;discriminate. Qed.


(** * Freshness *)

Reserved Notation "a # e" (at level 75).
Reserved Notation "a #' e" (at level 75).
Fixpoint fresh a e :=
  match e with
  | var x => ~ In a (τ x)
  | vat x => a <> x
  | 0 | 1 | ⊥ _ | id _ => True
  | e + f | e · f => a # e /\ a # f
  | e ^* | ω _ e => a # e
  | ν b e => a = b \/ a # e
  |‹ p › e => (p⋆⋈a) # e
  end
where "a # e" := (fresh a e).
Hint Unfold fresh.
Fixpoint freshb a e :=
  match e with
  | var x => negb (inb a (τ x))
  | vat x => negb (beq_atom a x)
  | 0 | 1 | ⊥ _ | id _ => true
  | e + f | e · f => (a #' e) && (a #' f)
  | e ^* | ω _ e => a #' e
  | ν b e => (beq_atom a b) || (a #' e)
  |‹ p › e => (p⋆⋈a) #' e
  end
where "a #' e" := (freshb a e).
Lemma freshb_spec a e : reflect (a # e) (a #' e).
Proof.
  apply iff_reflect;revert a;induction e;intro b;simpl;
  repeat simpl_bool;try now firstorder (auto with bool).
  - case_ex b (τ l);firstorder (auto with bool).
  - case_eq_at b a;simpl_beq; firstorder (auto with bool).
  - rewrite <-(reflect_iff _ _ (beq_atom_refl _ _)),IHe;tauto.
Qed.

(** * Expressions with variables over Letters. *)
Fixpoint multi e : Prop :=
  match e with
  | var _ | un | zero | id _ | bot _ => True
  | plus e f | times e f => multi e /\ multi f
  | star e | ν _ e | ω _ e | θ _ e => multi e
  | vat _ => False
  end.
Definition mlt := multi^2.
Hint Unfold multi mlt.
Fixpoint multib e :=
  match e with
  | var _ | un | zero | id _ | bot _ => true
  | plus e f | times e f => multib e && multib f
  | star e | ν _ e | ω _ e | θ _ e => multib e
  | vat _ => false
  end.
Lemma multib_spec e : reflect (multi e) (multib e).
Proof.
  apply iff_reflect.
  induction e;simpl;try simpl_bool;firstorder (auto with bool).
Qed.
Lemma decidable_multi : forall e, multi e \/ ~ multi e.
Proof. induction e;firstorder. Qed.

(** * Expressions with variables over Atoms *)
Fixpoint single e : Prop :=
  match e with
  | vat _ | un | zero | id _ | bot _ => True
  | plus e f | times e f => single e /\ single f
  | star e | ν _ e | ω _ e | θ _ e => single e
  | var _ => False
  end.
Definition sg := single^2.
Hint Unfold single sg.
Fixpoint singleb e :=
  match e with
  | vat _ | un | zero | id _ | bot _ => true
  | plus e f | times e f => singleb e && singleb f
  | star e | ν _ e | ω _ e | θ _ e => singleb e
  | var _ => false
  end.
Lemma singleb_spec e : reflect (single e) (singleb e).
Proof.
  apply iff_reflect.
  induction e;simpl;try simpl_bool;firstorder (auto with bool).
Qed.
Lemma decidable_single : forall e, single e \/ ~ single e.
Proof. induction e;firstorder. Qed.

(** Atomic expressions have a support smaller than their size. *)
Lemma single_support_le_size e : single e -> length (support e) <= size e.
Proof. induction e;simpl;intuition;
       try rewrite app_length || rewrite mact_length || rewrite rm_length;lia.
Qed.

(** * Expressions not using permutations *)
Fixpoint no_perm e : Prop :=
  match e with
  | var _ | vat _ | un | zero | id _ | bot _ => True
  |‹ _ › _ => False
  | plus e f | times e f => no_perm e /\ no_perm f
  | star e | ν _ e | ω _ e => no_perm e
  end.
Definition np := no_perm^2.
Hint Unfold no_perm np.
Fixpoint no_permb e :=
  match e with
  | var _ | vat _ | un | zero | id _ | bot _ => true
  |‹ _ › _ => false
  | plus e f | times e f => no_permb e && no_permb f
  | star e | ν _ e | ω _ e => no_permb e
  end.
Lemma no_permb_spec e : reflect (no_perm e) (no_permb e).
Proof.
  apply iff_reflect.
  induction e;simpl;try simpl_bool;firstorder (auto with bool).
Qed.
Lemma decidable_no_perm : forall e, no_perm e \/ ~ no_perm e.
Proof. induction e;firstorder. Qed.


(** lemmas about [eq0'] and [get_bot] when the argument is untyped *)
Lemma eq0'_u_spec {e} : untyped e -> reflect (e = 0) (eq0' e).
Proof. destruct e;simpl;intro U;constructor; firstorder congruence. Qed.
Lemma get_bot_u_spec {e} :
  untyped e ->
  reflect3 (e = 0) (fun B => False) (e<>0) (get_bot e).
Proof.
  destruct e;simpl;constructor; firstorder congruence.
Qed.
Lemma eq0'_get_bot_positiveu e :
  untyped e -> ~ (e = 0) ->
  eq0' e = false /\ get_bot e = inl true.
Proof.
  intro U;case (eq0'_u_spec U);[intros ->;tauto|intros _ ].
  case (get_bot_u_spec U);[intros ->| |];intros;tauto.
Qed.
(** lemmas about [eq0'] and [get_bot] when the argument is typed *)
Lemma eq0'_t_spec {e A} : A ⊨ e ->
  reflect (exists B, e=⊥ B) (eq0' e).
Proof.
  destruct e;simpl; intro T; constructor;eauto; tauto
  || (intros (?&h);discriminate).
Qed.
Lemma get_bot_t_spec {e A} : A ⊨ e ->
  reflect3 (False) (fun B => e = ⊥ B) (eq0' e = false) (get_bot e).
Proof.
  destruct e;simpl;constructor; firstorder congruence.
Qed.
Lemma eq0'_get_bot_positivet e A :
  A ⊨ e ->
  ~ (exists B : list Atom, e = ⊥ B) ->
  eq0' e = false /\ get_bot e = inl true.
Proof.
  intro U;case (eq0'_t_spec U);[intros (?&->) h;exfalso;apply h;eauto
                               |intros _ ].
  case (get_bot_t_spec U);[|intros ? -> h;exfalso;apply h;eauto |];
  intros;tauto.
Qed.

(** * Facts about expressions. *)

(** All types associated with a given expression are equivalent. *)
Lemma type_unicity e :
  forall A B, A ⊨ e -> B ⊨ e -> (A ≈ B).
Proof.
  intros;transitivity (stage e);[|symmetry];apply stage_has_type;auto.
Qed.
(** Technical lemma about types. *)
Lemma type_eq_cons :
  forall e a A B, (a::A) ⊨ e -> (a::B) ⊨ e ->
                  ~ In a A -> ~ In a B -> A ≈ B.
Proof.
  intros e a B C h1 h2 h3 h4;rewrite<- (cons_rm_eq' h3);
  rewrite<- (cons_rm_eq' h4);eapply rm_cong,type_unicity;eauto.
Qed.

Ltac solve_no_dup :=
  apply NoDup_nil
  || apply τ_spec
  || apply NoDup_cons
  || apply NoDup_app
  || apply NoDup_purge
  || apply NoDup_inter
  || apply NoDup_rm
  || apply NoDup_mact
  || apply NoDup_undup.

(** An atom is fresh for [e] if and only if it is not free in [e]. *)
Lemma fresh_support a e : a # e <-> (~ (In a (support e))).
Proof.
  revert a;induction e;intros α;simpl;simpl_in;
  split;intro h; firstorder subst.
  * intros [->|?];tauto.
  * now intro h;apply H.
  * tauto.
  * now rewrite IHe;case_eq_at α a;auto.
Qed.

(** * Properties of [W] *)

(** If [l] is without duplication and if [e] can be typed by a
list disjoint from [l], then [W l e] can be typed. *)
Lemma W_type :
  forall l A e, (A -- l) ⊨ e -> l ⊆ A -> NoDup l -> A ⊨ W l e.
Proof.
  intro l;induction l; intros A e T I D.
  now rewrite purge_nil in T.
  apply NoDup_cons_iff in D as (D1 & D2);rewrite purge_cons in T;
  simpl in *;repeat split; [ |apply IHl;eauto].
  now apply I;left.
  intros x h;simpl_in;split;[apply I;right|intros ->];tauto.
Qed.

(** If [W l e] can be typed, then so can [e]. *)
Lemma type_W' : forall l A e, A ⊨ W l e -> (A -- l) ⊨ e.
Proof.
  induction l;intros A e h;simpl in *.
  now rewrite purge_nil.
  now rewrite purge_cons;intuition.
Qed.
(** [W] perserves [no_perm],[single],[multi] and [str_positive]. *)
Lemma no_perm_W : forall e l, no_perm e -> no_perm (W l e).
Proof. induction l;simpl;intuition. Qed.
Lemma single_W : forall e l, single e -> single (W l e).
Proof. induction l;simpl;intuition. Qed.
Lemma multi_W : forall e l, multi e -> multi (W l e).
Proof. induction l;simpl;intuition. Qed.
Lemma str_positive_W :
  forall e l, str_positive e -> str_positive (W l e).
Proof.  induction l;simpl;intuition. Qed.
