(** printing ⊢ %\ensuremath{\vdash}% *)
(** printing ≡ %\ensuremath{\equiv}% *)
(** printing ≤ %\ensuremath{\leqslant}% *)
(** printing =* %\ensuremath{=^\star}% *)
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

Require Export expr predicates morphisms.

Reserved Notation "e =* f" (at level 80).

(** Kleene algebra axioms with no constants. *)
Inductive ka_eq : relation expr :=
| ka_tt e f g : (e · (f · g)) =* ((e · f) · g)

| ka_pp e f g : (e + (f + g)) =* ((e + f) + g)
| ka_pc e f : (e + f) =* (f + e)
| ka_pi e : (e + e) =* e
                    
| ka_pt e f g : (e · (f + g)) =* ((e · f) + (e · g))
| ka_tp e f g : ((e + f) · g) =* ((e · g) + (f · g))

where "e =* f" := (ka_eq e f).
(** Kleene algebra entailments. *)
Inductive ka_impl : expr -> relation expr :=
| ka_s3 e f g : ka_impl (f + e·g) (e^*·f) g
| ka_s4 e f g : ka_impl (f + g·e) (f·e^* ) g.

(** Kleene algebra axioms for untyped constants. *)
Inductive ka_eq_u : relation expr :=
| kau_t1 e : ka_eq_u (e · 1) e
| kau_1t e : ka_eq_u (1 · e) e
| kau_s1 e : ka_eq_u (1 + e · (e^*) + e ^*) (e^*)
| kau_s2 e : ka_eq_u (1 + (e^*) · e + e^*) (e^*).
(** Kleene algebra axioms for typed constants. *)
Inductive ka_eq_t : relation expr :=
| kat_t1 e A : A ⊨ e -> ka_eq_t (e · id A) e
| kat_1t e A : A ⊨ e -> ka_eq_t (id A · e) e
| kat_s1 e A : A ⊨ e -> ka_eq_t (id A + e · (e^*) + e ^*) (e^*)
| kat_s2 e A : A ⊨ e -> ka_eq_t (id A + (e^*) · e + e^*) (e^*).
                                   
Hint Constructors ka_eq ka_impl ka_eq_u ka_eq_t.

Reserved Notation " ax ⊢ e ≡ f" (at level 80).
Reserved Notation " ax ⊢ e ≤ f" (at level 80).

Inductive eq (ax : relation expr) : relation expr :=
(** Structural rules *)
| eq_sym e f : (ax ⊢ e ≡ f) -> ax ⊢ f ≡ e
| eq_trans e f g : ax ⊢ e ≡ f -> ax ⊢ f ≡ g -> ax ⊢ e ≡ g

(** Congruence *)
| eq_zero : ax ⊢ 0 ≡ 0
| eq_un : ax ⊢ 1 ≡ 1
| eq_bot A B : A ≈ B -> ax ⊢ (⊥ A) ≡ (⊥ B)
| eq_id A B : A ≈ B -> ax ⊢ (id A) ≡ (id B)
| eq_var x y : x ≃ y -> ax ⊢ (var x) ≡ (var y)
| eq_vat a : ax ⊢ (vat a) ≡ (vat a)
| eq_plus e f g h : ax ⊢ e ≡ f -> ax ⊢ g ≡ h -> ax ⊢ (e+g) ≡ (f+h)
| eq_times e f g h : ax ⊢ e ≡ f -> ax ⊢ g ≡ h -> ax ⊢ (e·g) ≡ (f·h)
| eq_star e f : ax ⊢ e ≡ f -> ax ⊢ (e^* ) ≡ (f^* )
| eq_ν a e f: ax ⊢ e ≡ f -> ax ⊢ (ν a e) ≡ (ν a f)
| eq_θ p q e f: p≌q -> ax ⊢ e ≡ f -> ax ⊢ (‹ p › e) ≡ (‹ q › f)
| eq_ω a e f: ax ⊢ e ≡ f -> ax ⊢ (ω a e) ≡ (ω a f)

(** Kleene algebra without constants *)           
| eq_ka e f : e =* f -> ax ⊢ e ≡ f
| eq_ka_impl e f g : ax ⊢ e ≤ g -> ka_impl e f g -> ax ⊢ f ≤ g
                     
(** Additionnal axioms *)
| eq_ax e f : ax e f -> ax ⊢ e ≡ f
where " ax ⊢ e ≡ f " := (eq ax e f)
and " ax ⊢ e ≤ f " := (eq ax (e+f) f).
Hint Constructors eq.
(** For any axiomatic system the [eq] relation is reflexive. *)
Lemma eq_refl ax e : ax ⊢ e ≡ e.
Proof. induction e;eauto. Qed.
Hint Resolve eq_refl.
(** For any axiomatic system the [eq] relation an equivalence relation. *)
Instance eq_Equiv ax : Equivalence (eq ax).
Proof. split;eauto. Qed.
Lemma eq_expr_eq ax e f : (e == f) -> ax ⊢ e ≡ f.
Proof.
  revert f;induction e;intros f h;destruct f;simpl in *;intuition;
  now subst; eauto.
Qed.
Hint Resolve eq_expr_eq.

Global Instance stronger_ax_proof :
  Proper ((@stronger expr) ==> inclusion expr) eq.
Proof. intros ax1 ax2 S e f E;induction E;eauto. Qed.
Lemma stronger_ax_eq a b : a ⇒ b -> a ⇒ eq b.
Proof. intros h e f A;apply eq_ax,h,A. Qed.
Lemma stronger_eq_or_rel_l :
  forall A B, inclusion expr (eq A) (eq (A∨B)).
Proof. intros A B;apply stronger_ax_proof,or_rel_stronger_l. Qed.
Lemma stronger_eq_or_rel_r :
  forall A B, inclusion expr (eq B) (eq (A∨B)).
Proof. intros A B;apply stronger_ax_proof,or_rel_stronger_r. Qed.
Ltac eql := apply stronger_eq_or_rel_l.
Ltac eqr := apply stronger_eq_or_rel_r.

Ltac simpl_or_rel :=
  let rec aux l acc :=
      match l with
      | ?A∨(?B∨?C) => (aux ((A∨B)∨C) acc)
      | ?A∨?B =>
        match acc with
        | FFalse => aux A B
        | ?C => aux A (C∨B)
        end
      | ?A => 
        match acc with
        | FFalse => apply (@stronger_ax_proof A) 
        | ?C => apply (@stronger_ax_proof (C∨A))
        end
      end
  in
  let e := fresh "e" in
  let f := fresh "f" in
  match goal with
  | _:_ |- ?ax ⊢ _ ≡ _ => aux ax (fun _ _ : expr => False);
      [intros e f;unfold or_rel;tauto|]
  end.
Ltac prove_ax :=
  apply eq_ax;repeat (now eauto)||(now right; eauto)||left.
Goal (forall A B C e f, A∨B∨C ⊢ e ≡ f -> A∨(B∨C) ⊢ e ≡ f).
  intros. simpl_or_rel. induction H;eauto.
  destruct H as [[H|H]|H];prove_ax.
Qed.

(** For any axiomatic system all operators are morphisms with respect to [eq]. *)
Global Instance ν_eq ax a :
  Proper (eq ax ==> eq ax) (ν a).
Proof. intros ? ? ?;auto. Qed.
Global Instance ω_eq ax a :
  Proper (eq ax ==> eq ax) (ω a).
Proof. intros ? ? ?;auto. Qed.
Global Instance plus_eq ax :
  Proper (eq ax ==> eq ax ==> eq ax) plus.
Proof. intros ? ? ? ? ?;auto. Qed.
Global Instance times_eq ax :
  Proper (eq ax ==> eq ax ==> eq ax) times.
Proof. intros ? ? ? ? ?;auto. Qed.
Global Instance θ_eq ax :
  Proper (equiv_perm ==> eq ax ==> eq ax) θ.
Proof. autounfold;firstorder auto. Qed.
Global Instance star_eq ax:
  Proper (eq ax ==> eq ax) star.
Proof. intros ? ? ?;auto. Qed.
Global Instance bot_equiv_lst_eq ax:
  Proper (equiv_lst ==> eq ax) bot.
Proof. intros ? ? ?;auto. Qed.
Global Instance id_equiv_lst_eq ax:
  Proper (equiv_lst ==> eq ax) id.
Proof. intros ? ? ?;auto. Qed.

Notation inf ax := (fun x y => ax ⊢ x ≤ y).
(** The relation ≤ is antisymmetric with respect to ≡. *)
Instance inf_antisym ax : Antisymmetric expr (eq ax) (inf ax).
Proof.
  intros x y h1 h2;rewrite <-h1,<-h2;eauto.
Qed.


(** Kleene algebra axioms for 0. *)
Inductive zero_ka_u : relation expr :=
| zero_u_p0 e : zero_ka_u (e + 0) e
| zero_u_t0 e : untyped e -> zero_ka_u (e · 0) 0
| zero_u_0t e : untyped e -> zero_ka_u (0 · e) 0.

(** Kleene algebra axioms for ⊥. *)
Inductive zero_ka_t : relation expr :=
| zero_t_p0 e A : A ⊨ e -> zero_ka_t (e + ⊥ A) e
| zero_t_t0 e A : A ⊨ e -> zero_ka_t (e · ⊥ A) (⊥ A)
| zero_t_0t e A : A ⊨ e -> zero_ka_t (⊥ A · e) (⊥ A).

Inductive zero_perm_u : relation expr :=
| eq_θ_0 p : zero_perm_u (‹ p ›0) 0.

Inductive zero_perm_t : relation expr :=
| eq_θ_bot A B p : (B ≈ p ⧓ A) -> zero_perm_t (‹ p ›(⊥ A)) (⊥ B).
Inductive zero_nom_u : relation expr :=
| equ_ν0 a : zero_nom_u (ν a 0) 0. (** derivable from eq_nom_untyped, because [a#0] *)
Inductive zero_nom_t : relation expr :=
| eq_ω_bot a A : ~(In a A) -> zero_nom_t (ω a (⊥ A)) (⊥ (a::A))
| eqt_ν0 A a : In a A -> zero_nom_t (ν a (⊥ A)) (⊥ (rm a A)).
Hint Constructors zero_ka_u zero_ka_t zero_perm_u zero_perm_t zero_nom_t.

Notation zero_eq_t :=  (zero_ka_t ∨ zero_perm_t ∨ zero_nom_t).
Notation zero_eq_u :=  (zero_ka_u ∨ zero_perm_u ∨ zero_nom_u).

(** Permutations. *)
Inductive eq_perm : relation expr :=
| eq_θ_p p e f : eq_perm (‹ p › (e+f)) (‹ p ›e + ‹ p ›f)
| eq_θ_t p e f : eq_perm (‹ p › (e·f)) (‹ p ›e · ‹ p ›f)
| eq_θ_s p e :  eq_perm (‹ p › (e^* )) ((‹ p ›e)^* )
| eq_θ_ν p a e : eq_perm (‹ p › (ν a e)) (ν (p⋈a) (‹ p › e))
| eq_θ_θ p' p e : eq_perm (‹ p › (‹ p' › e)) (‹p++p'› e).

Inductive eq_perm_ω : relation expr :=
| eq_θ_ω p a e : eq_perm_ω (‹ p › (ω a e)) (ω (p⋈a) (‹ p › e)).

Inductive eq_perm_elim : relation expr:=
| eq_θ_1 p : eq_perm_elim (‹ p ›1) 1
| eq_θ_at p a : eq_perm_elim (‹ p ›(vat a)) (vat (p ⋈ a))
| eq_θ_var p a : eq_perm_elim (‹ p ›(var a)) (var (apply_perm_var p a))
| eq_θ_id A B p : (B ≈ p ⧓ A) -> eq_perm_elim (‹ p ›(id A)) (id B)
(* | eq_θ_var p q x : (p ⧓ (τ x) = q ⧓ (τ x)) -> *)
(*                    eq_perm_elim (‹ p ›var x) (‹ q ›var x) *)
| eq_nil_e e : eq_perm_elim (‹[]› e) e.

(** Distributivity laws of ν,ω. *)       
Inductive eq_ν_smpl : relation expr :=          
| eq_νp a e f : eq_ν_smpl (ν a (e+f)) (ν a e + ν a f)
| eq_νab a b e : eq_ν_smpl (ν a (ν b e)) (ν b (ν a e)).
Inductive eq_ω_smpl : relation expr :=          
| eq_ω_id a A : ~(In a A) -> eq_ω_smpl (ω a (id A)) (id (a::A))
| eq_ω_p e f a : eq_ω_smpl (ω a (e + f)) (ω a e + ω a f)
| eq_ω_t e f a : eq_ω_smpl (ω a (e · f)) (ω a e · ω a f)
| eq_ω_s e a : eq_ω_smpl (ω a (e^*)) ((ω a e)^*)
| eq_ωω e a b : eq_ω_smpl (ω a (ω b e)) (ω b (ω a e)).


Hint Constructors eq_perm eq_perm_elim eq_ν_smpl eq_ω_smpl eq_perm_ω.

(** Untyped nominal axioms. *)
Inductive eq_nom_untyped_p : relation expr :=
| equ_θν' a b e :
    a # e -> eq_nom_untyped_p (ν b e) (ν a ( ‹[a,b]›e )).
Inductive eq_nom_untyped_np : relation expr :=
| equ_θν a b e :
    a # e -> eq_nom_untyped_np (ν b e) (ν a ( [a,b]•e )).

Inductive eq_nom_untyped : relation expr :=
| equ_νω1 a e :
    a # e -> eq_nom_untyped (ν a e) e
| equ_ν3 a e f :
    a # e -> eq_nom_untyped ((ν a f) · e) (ν a (f · e))
| equ_ν4 a e f :
    a # e -> eq_nom_untyped (e · (ν a f)) (ν a (e · f)).

(** Typed nominal axioms. *)
Inductive eq_nom_typed_p : relation expr :=
| eqt_θν' A a b e : A ⊨ ν a e -> ~ In b A ->
                    eq_nom_typed_p (ν a e) (ν b (‹[b,a]›e)).
Inductive eq_nom_typed_np : relation expr :=
| eqt_θν A a b e : A ⊨ ν a e -> ~ In b A ->
                    eq_nom_typed_np (ν a e) (ν b ([b,a]•e)).
Inductive eq_nom_typed : relation expr :=
(* | eqt_ν1 A a : In a A -> *)
(*              eq_nom_typed (ν a (id A)) (id (rm a A)) *)
| eqt_νω1 A a e :
    A ⊨ ν a (ω a e) ->
    eq_nom_typed (ν a (ω a e)) e
| eqt_νω2 e a b :
    a <> b -> eq_nom_typed (ν a (ω b e)) (ω b (ν a e))
| eqt_ν3 a e f :
    eq_nom_typed ((ν a f) · e) (ν a (f · ω a e))
| eqt_ν4 a e f :
    eq_nom_typed (e · (ν a f)) (ν a ((ω a e) · f)).
Hint Constructors eq_nom_typed eq_nom_untyped.

Lemma eqt_ν1 A a :
  In a A -> 
  eq_nom_typed ∨ eq_ω_smpl ⊢ (ν a (id A)) ≡ (id (rm a A)).
Proof.
  intro I.
  symmetry;etransitivity.
  symmetry;eql;apply eq_ax,(@eqt_νω1 (rm a A) a).
  simpl; unfold equiv_lst; intuition; repeat simpl_beq; simpl_in;firstorder.
  setoid_replace (id A) with (id (a::rm a A)).
  eqr;apply eq_ν,eq_ax,eq_ω_id;simpl_in;tauto.
  apply eq_id; intro x;repeat simpl_in;case_eq_at x a;firstorder.
Qed.
  
Lemma stronger_Ty_zero_ka_t : zero_ka_t ⇒ ty' ∧ zero_ka_t.
Proof. intros ? ? E;unfold ty';destruct E;split;simpl;eauto. Qed.
Lemma stronger_Ty_ka_eq_t : ka_eq_t ⇒ ty'∧ ka_eq_t.
Proof. intros ? ? E;unfold ty';destruct E;split;simpl;auto;
       exists A;auto. Qed.

Lemma one_star_is_one_u : ka_eq_u ⊢ 1^* ≡ 1.
Proof.
  apply inf_antisym.
  - transitivity (1·1^*+1);auto. 
    eapply eq_ka_impl.
    2:apply ka_s4;eauto.
    transitivity (1+1+1);auto.
    transitivity (1+1);auto.
  - transitivity (1+1·1^*+1^* );auto.
    transitivity (1+(1^*+1^* ));auto.
    transitivity (1+1^*+1^* );auto.
Qed.
Lemma one_star_is_one_t A: ka_eq_t ⊢ (id A)^* ≡ (id A).
Proof.
  apply inf_antisym.
  - transitivity (id A·id A^*+id A);auto.
    * transitivity (id A^*+id A);auto.
       symmetry;apply eq_plus;auto. 
    * eapply eq_ka_impl.
       2:apply ka_s4;eauto.
       transitivity (id A+id A+id A);auto.
       repeat apply eq_plus;auto.
       transitivity (id A+id A);auto.
  - transitivity (id A+(id A^*+id A^* ));auto.
    transitivity (id A+id A^*+id A^* );auto.
    transitivity (id A+id A·id A^*+id A^* );auto.
    repeat apply eq_plus;auto.
Qed.

Lemma zero_nom_u_derivable : zero_nom_u ⇒ eq_nom_untyped.
Proof. intros e f H;destruct H;apply equ_νω1;simpl;auto. Qed.

Lemma derivable_axioms (ax1 ax2 : relation expr) :
  (forall e f, ax1 e f -> ax2 ⊢ e ≡ f) ->
  (forall e f, (ax1∨ax2) ⊢ e ≡ f -> ax2 ⊢ e ≡ f).
Proof.
  intros hyp e f E;induction E;eauto.
  destruct H as [H|H];auto.
Qed.


(** Type preservation. *)
Create HintDb propertype.
Lemma eq_nom_typed_np_type A :
    Proper (eq_nom_typed_np ==> iff) (has_type A).
Proof.
  intros e f E;destruct E;split; simpl in *;
  intro T;simpl_hyp;auto;
  try match goal with| h:_≈_ |- _ =>rewrite h||rewrite<- h end;
  simpl in *;repeat simpl_beq;simpl in *;simpl_in;split_conj;
  auto;try tauto.
  - rewrite (@type_eq_cons e a _ _ h0 h2);auto.
  - apply has_type_apply_perm;simpl;simpl_beq.
    case_eq_at b a;try simpl_beq;simpl;[rewrite mact_id1
                                       |rewrite mact_id2];auto.
    rewrite (@type_eq_cons e a _ _ h0 h2);auto.
  - assert (h0': [b,a]⋆⧓(b::A)⊨e)
      by (now apply apply_perm_has_type with (p:=[b, a]);
           simpl_perm).
    simpl in *;simpl_beq.
    case_eq_at a b;try simpl_beq;simpl in *.
    * rewrite mact_id1 in h0';tauto.
    * intro I. cut (In b (a::A0));[simpl;tauto|].
       rewrite (@type_eq_cons e a _ _ h2 h0');simpl_in;simpl;
       repeat simpl_beq;auto.
  - assert (h0': [b,a]⋆⧓(b::A)⊨e)
      by (now apply apply_perm_has_type  with (p:=[b, a]);
           simpl_perm).
    simpl in *;simpl_beq.
    case_eq_at a b;try simpl_beq;simpl in *;[rewrite mact_id1 in h0'
                                            |rewrite mact_id2 in h0'];
    auto.
    intro I. cut (In b (a::A0));[simpl;tauto|].
    rewrite (@type_eq_cons e a _ _ h2 h0');simpl_in;simpl;
    repeat simpl_beq;auto.
Qed.
Hint Resolve eq_nom_typed_np_type : propertype.
Lemma eq_nom_typed_p_type A :
    Proper (eq_nom_typed_p ==> iff) (has_type A).
Proof.
  intros e f E;destruct E;split; simpl in *;
  intro T;simpl_hyp;auto;
  try match goal with| h:_≈_ |- _ =>rewrite h||rewrite<- h end;
  simpl in *;repeat simpl_beq;simpl in *;simpl_in;split_conj;
  auto;try tauto.
  - rewrite (@type_eq_cons e a _ _ h0 h2);auto.
  - case_eq_at b a;try simpl_beq;simpl;[rewrite mact_id1
                                       |rewrite mact_id2];auto.
    rewrite (@type_eq_cons e a _ _ h0 h2);auto.
  - case_eq_at a b;try simpl_beq;simpl in *.
    * rewrite mact_id1 in h0;tauto.
    * intro I. cut (In b (a::A0));[simpl;tauto|].
       rewrite (@type_eq_cons e a _ _ h2 h0);simpl_in;simpl;
       repeat simpl_beq;auto.
  - case_eq_at a b;try simpl_beq;simpl in *;[rewrite mact_id1 in h0
                                            |rewrite mact_id2 in h0];
    auto.
    intro I. cut (In b (a::A0));[simpl;tauto|].
    rewrite (@type_eq_cons e a _ _ h2 h0);simpl_in;simpl;
    repeat simpl_beq;auto.
Qed.
Hint Resolve eq_nom_typed_p_type : propertype.

Lemma eq_nom_typed_type A :
    Proper (eq_nom_typed ==> iff) (has_type A).
Proof.
  intros e f E;destruct E;split; simpl in *;
  intro T;simpl_hyp;auto;
  try match goal with| h:_≈_ |- _ =>rewrite h||rewrite<- h end;
  simpl in *;repeat simpl_beq;simpl in *;simpl_in;split_conj;
  auto;try tauto.
  -rewrite rm_id in h2;tauto.
  - rewrite (type_unicity T h2);rewrite rm_id;tauto.
  - prove_type;rewrite rm_id;auto.
    rewrite (type_unicity T h2);rewrite rm_id;tauto.
  - rewrite rm_id;tauto.
  - rewrite rm_id in h3;tauto.
  - rewrite rm_id;tauto.
  - rewrite rm_id in h3;tauto.
Qed.
Hint Resolve eq_nom_typed_type : propertype.
Lemma eq_perm_elim_type A :
    Proper (eq_perm_elim ==> iff) (has_type A).
Proof.
  intros e f E;destruct E;split;simpl;intro T;
  repeat simpl_perm;auto;
  try match goal with
      | h:_≈_ |- _ =>
        rewrite h||rewrite<- h
      end;simpl;simpl_in;repeat simpl_perm;split_conj;
  auto;try tauto.
  - transitivity (p⧓[a]);auto. now rewrite<- T;simpl_perm.
  - now rewrite apply_perm_var_τ,<-T;simpl_perm.
  - now rewrite apply_perm_var_τ;simpl_perm.
  - now rewrite<- T;simpl_perm.
  - now rewrite H;simpl_perm.
Qed.
Hint Resolve eq_perm_elim_type : propertype.
Lemma eq_perm_type A :
    Proper (eq_perm ==> iff) (has_type A).
Proof.
  intros e f E;destruct E;split;simpl;intro T;
  simpl_in;repeat (simpl in *;simpl_perm);auto;try tauto.
  - rewrite <-mact_comp_app. assumption.
  - rewrite <-mact_comp_app in T;assumption.
Qed.
Hint Resolve eq_perm_type : propertype.
Lemma eq_perm_ω_type A :
    Proper (eq_perm_ω ==> iff) (has_type A).
Proof.
  intros e f E;destruct E;split;simpl;intro T;
  simpl_in;repeat (simpl in *;simpl_perm);auto;try tauto.
  - rewrite mact_rm in T;simpl_perm;tauto.
  - rewrite mact_rm;simpl_perm;tauto.
Qed.
Hint Resolve eq_perm_ω_type : propertype.
Lemma eq_ν_smpl_type A :
  Proper (eq_ν_smpl ==> iff) (has_type A).
Proof.
  intros e f E;destruct E;split;simpl;intro T;
  simpl_in;repeat (simpl in *;simpl_perm);auto;
  try tauto.
  - rewrite cons_comm;intuition eauto.
  - rewrite cons_comm;intuition eauto.
Qed.
Hint Resolve eq_ν_smpl_type : propertype.
Lemma eq_ω_smpl_type A :
  Proper (eq_ω_smpl ==> iff) (has_type A).
Proof.
  intros e f E;destruct E;split;simpl;intro T;
  simpl_in;repeat (simpl in *;simpl_perm);auto;
  try tauto.
  - destruct T as (I& T);now rewrite <- T;auto.
  - rewrite T; simpl;simpl_beq;simpl; rewrite rm_id;auto.
  - rewrite rm_comm; intuition congruence.
  - rewrite rm_comm; intuition congruence.
Qed.
Hint Resolve eq_ω_smpl_type : propertype.
Lemma zero_ka_t_type A :
  Proper (zero_ka_t ==> iff) (has_type A).
Proof.
  intros e f E;destruct E;split;simpl;intro T;
  simpl_in;repeat (simpl in *;simpl_perm);auto;
  intuition ((eapply type_unicity;eauto)||(prove_type;eauto)).
Qed.
Hint Resolve zero_ka_t_type : propertype.

Lemma ka_eq_t_type A :
  Proper (ka_eq_t ==> iff) ( has_type A).
Proof.
  intros e f E;destruct E;split;simpl;intro T;
  simpl_in;repeat (simpl in *;simpl_perm);
  intuition ((eapply type_unicity;eauto)||(prove_type;eauto)).
Qed.
Hint Resolve ka_eq_t_type : propertype.
Lemma ka_eq_type A :
  Proper (ka_eq ==> iff) ( has_type A).
Proof. intros e f E;destruct E;simpl;firstorder. Qed.

Hint Resolve ka_eq_type : propertype.
Lemma zero_perm_t_type A :
  Proper (zero_perm_t ==> iff) ( has_type A).
Proof.
  intros e f E. destruct E as [X Y p E].
  rewrite (ht_bot_eq _ E);simpl.
  now rewrite map_act_inv_eq;simpl_perm. 
Qed.
Hint Resolve zero_perm_t_type : propertype.

Lemma zero_nom_t_type A :
  Proper (zero_nom_t ==> iff) ( has_type A).
Proof.
  intros e f E;destruct E as [a X E|X a E].
  simpl;split;[intros(?&e);rewrite <-e|intros e;rewrite e;split;[left|]];auto.
  simpl;split;[intros(?&e);rewrite <-e|intros e;rewrite e;split;[simpl_in|]];auto;tauto.
Qed.
Hint Resolve zero_nom_t_type : propertype.

Lemma eq_type ax:
  (forall A, Proper (ax ==> iff) ( has_type A)) ->
  forall A, Proper (eq ax ==> iff) ( has_type A).
Proof.
  intros hax A e f E;revert A;induction E;simpl;intro X;
  try (now firstorder;auto).
  - now destruct x as ((?,?),?); destruct y as ((?,?),?);
    unfold eq_letter, τ in *;simpl in *;inversion H;subst.
  - now rewrite (equiv_perm_mact _ _ (equiv_perm_inverse H)).
  - now apply ka_eq_type.
  - rewrite<- IHE;destruct H;simpl in *;tauto.
Qed.

Lemma eq_nom_untyped_untyped :
  Proper (eq_nom_untyped ==> iff) untyped.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Hint Resolve eq_nom_untyped_untyped : propertype.

Lemma eq_nom_untyped_np_untyped :
  Proper (eq_nom_untyped_np ==> iff) untyped.
Proof. intros e f E. destruct E;simpl. rewrite<- untyped_apply_perm.
       reflexivity.
Qed.
Hint Resolve eq_nom_untyped_np_untyped : propertype.
Lemma eq_nom_untyped_p_untyped :
  Proper (eq_nom_untyped_p ==> iff) untyped.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Hint Resolve eq_nom_untyped_p_untyped : propertype.
Lemma eq_perm_elim_untyped : Proper (eq_perm_elim ==> iff) untyped.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Hint Resolve eq_perm_elim_untyped : propertype.
Lemma eq_perm_untyped : Proper (eq_perm ==> iff) untyped.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Hint Resolve eq_perm_untyped : propertype.
Lemma eq_perm_ω_untyped : Proper (eq_perm_ω ==> iff) untyped.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Hint Resolve eq_perm_ω_untyped : propertype.
Lemma eq_ν_smpl_untyped : Proper (eq_ν_smpl ==> iff) untyped.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Hint Resolve eq_ν_smpl_untyped : propertype.
Lemma eq_ω_smpl_untyped : Proper (eq_ω_smpl ==> iff) untyped.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Hint Resolve eq_ω_smpl_untyped : propertype.
Lemma zero_ka_u_untyped : Proper (zero_ka_u ==> iff) untyped.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Hint Resolve zero_ka_u_untyped : propertype.
Lemma ka_eq_u_untyped : Proper (ka_eq_u ==> iff) untyped.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Hint Resolve ka_eq_u_untyped : propertype.
Lemma ka_eq_untyped : Proper (ka_eq ==> iff) untyped.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Hint Resolve ka_eq_untyped : propertype.
Lemma zero_perm_u_untyped : Proper (zero_perm_u ==> iff) untyped.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Hint Resolve zero_perm_u_untyped : propertype.
Lemma zero_nom_u_untyped : Proper (zero_nom_u ==> iff) untyped.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Hint Resolve zero_nom_u_untyped : propertype.
Lemma eq_untyped ax :
  (Proper (ax ==> iff) untyped) ->
  Proper (eq ax ==> iff) untyped.
Proof.
  intros hax e f E;induction E;simpl;
  try (now firstorder;auto).
  now apply ka_eq_untyped.
  rewrite<- IHE;destruct H;simpl in *;tauto.
Qed.

Lemma proper_eq : forall (ax1 ax2 : relation expr) φ,
    Proper (ax1 ==> ax2) φ ->
    Proper (ax1 ==> eq ax2) φ.
Proof. intros ax1 ax2 φ P e f E;apply eq_ax,P,E. Qed.

Lemma remain_Ty ax :
  (forall A, Proper (ax ==> iff) ( has_type A)) ->
  forall e f A,  A ⊨ e -> ax ⊢ e ≡ f -> ty'∧ ax ⊢ e ≡ f.
Proof.
  intros hyp e f A T E;revert A T.
  pose proof (eq_type hyp) as Typ.
  induction E;intros X T;auto.
  - symmetry;apply (IHE X);firstorder.
  - transitivity f;[apply (IHE1 X)|apply (IHE2 X)];firstorder.
  - destruct T as (T1&T2);rewrite (IHE1 X), (IHE2 X);firstorder.
  - destruct T as (T1&T2);rewrite (IHE1 X), (IHE2 X);firstorder.
  - rewrite (IHE X);firstorder.
  - rewrite (IHE (a::X));firstorder.
  - rewrite (IHE (p⋆⧓X));firstorder.
  - rewrite (IHE (rm a X));firstorder.
  - destruct T as (T1&T2);apply (@eq_ka_impl _ e);auto.
    apply (IHE X);split;auto.
    destruct H;destruct T1;firstorder.
  - apply eq_ax;split;[exists X|];firstorder.
Qed.

Lemma remain_Ut ax :
  (Proper (ax ==> iff) untyped) ->
  forall e f,  untyped e -> ax ⊢ e ≡ f -> ut ∧ ax ⊢ e ≡ f.
Proof.
  intros hyp e f U E;revert U.
  pose proof (eq_untyped hyp) as Unt.
  induction E;intros U;auto.
  - symmetry;apply IHE;firstorder.
  - transitivity f;[apply IHE1|apply IHE2];firstorder.
  - destruct U as (U1&U2);rewrite IHE1, IHE2;firstorder.
  - destruct U as (U1&U2);rewrite IHE1, IHE2;firstorder.
  - rewrite IHE;firstorder.
  - destruct U as (U1&U2);apply (@eq_ka_impl _ e);auto.
    apply IHE;split;auto.
    destruct H;destruct U1;firstorder.
  - apply eq_ax;split;firstorder.
Qed.


Lemma eq_nom_typed_single : Proper (eq_nom_typed ==> iff) single.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma eq_nom_typed_np_single : Proper (eq_nom_typed_np ==> iff) single.
Proof. intros e f E. destruct E;simpl;now rewrite <-single_apply_perm.
Qed.
Lemma eq_nom_typed_p_single : Proper (eq_nom_typed_p ==> iff) single.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma eq_nom_untyped_single : Proper (eq_nom_untyped ==> iff) single.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma eq_nom_untyped_np_single :
  Proper (eq_nom_untyped_np ==> iff) single.
Proof. intros e f E. destruct E;simpl;now rewrite <-single_apply_perm.
Qed.
Lemma eq_nom_untyped_p_single :
  Proper (eq_nom_untyped_p ==> iff) single.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma eq_ω_smpl_single : Proper (eq_ω_smpl ==> iff) single.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma eq_ν_smpl_single : Proper (eq_ν_smpl ==> iff) single.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma eq_perm_elim_single : Proper (eq_perm_elim ==> iff) single.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma eq_perm_ω_single : Proper (eq_perm_ω ==> iff) single.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma eq_perm_single : Proper (eq_perm ==> iff) single.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma zero_nom_t_single : Proper (zero_nom_t ==> iff) single.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma zero_nom_u_single : Proper (zero_nom_u ==> iff) single.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma zero_perm_t_single : Proper (zero_perm_t ==> iff) single.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma zero_perm_u_single : Proper (zero_perm_u ==> iff) single.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma ka_eq_t_single : Proper (ka_eq_t ==> iff) single.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma ka_eq_u_single : Proper (ka_eq_u ==> iff) single.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma ka_eq_single : Proper (ka_eq ==> iff) single.
Proof. intros e f E. destruct E;simpl;tauto. Qed.

Lemma eq_single ax :
  (Proper (ax ==> iff) single) ->
  (Proper (eq ax ==> iff) single).
Proof.
  intros hyp e f E;induction E;simpl;firstorder;
  destruct H;firstorder.
Qed.  
  
Lemma remain_Sg ax :
  (Proper (ax ==> iff) single) ->
  forall e f,  single e -> ax ⊢ e ≡ f -> sg ∧ ax ⊢ e ≡ f.
Proof.
  intros hyp e f U E;revert U.
  pose proof (eq_single hyp) as Unt.
  induction E;intros U;auto.
  - symmetry;apply IHE;firstorder.
  - transitivity f;[apply IHE1|apply IHE2];firstorder.
  - destruct U as (U1&U2);rewrite IHE1, IHE2;firstorder.
  - destruct U as (U1&U2);rewrite IHE1, IHE2;firstorder.
  - destruct U as (U1&U2);apply (@eq_ka_impl _ e);auto.
    apply IHE;split;auto.
    destruct H;destruct U1;firstorder.
  - apply eq_ax;split;firstorder.
Qed.


Lemma eq_nom_typed_multi : Proper (eq_nom_typed ==> iff) multi.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma eq_nom_typed_np_multi : Proper (eq_nom_typed_np ==> iff) multi.
Proof. intros e f E. destruct E;simpl;now rewrite <-multi_apply_perm.
Qed.
Lemma eq_nom_typed_p_multi : Proper (eq_nom_typed_p ==> iff) multi.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma eq_nom_untyped_multi : Proper (eq_nom_untyped ==> iff) multi.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma eq_nom_untyped_np_multi :
  Proper (eq_nom_untyped_np ==> iff) multi.
Proof. intros e f E. destruct E;simpl;now rewrite <-multi_apply_perm.
Qed.
Lemma eq_nom_untyped_p_multi :
  Proper (eq_nom_untyped_p ==> iff) multi.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma eq_ω_smpl_multi : Proper (eq_ω_smpl ==> iff) multi.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma eq_ν_smpl_multi : Proper (eq_ν_smpl ==> iff) multi.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma eq_perm_elim_multi : Proper (eq_perm_elim ==> iff) multi.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma eq_perm_ω_multi : Proper (eq_perm_ω ==> iff) multi.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma eq_perm_multi : Proper (eq_perm ==> iff) multi.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma zero_nom_t_multi : Proper (zero_nom_t ==> iff) multi.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma zero_nom_u_multi : Proper (zero_nom_u ==> iff) multi.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma zero_perm_t_multi : Proper (zero_perm_t ==> iff) multi.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma zero_perm_u_multi : Proper (zero_perm_u ==> iff) multi.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma ka_eq_t_multi : Proper (ka_eq_t ==> iff) multi.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma ka_eq_u_multi : Proper (ka_eq_u ==> iff) multi.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma ka_eq_multi : Proper (ka_eq ==> iff) multi.
Proof. intros e f E. destruct E;simpl;tauto. Qed.

Lemma eq_multi ax :
  (Proper (ax ==> iff) multi) ->
  (Proper (eq ax ==> iff) multi).
Proof.
  intros hyp e f E;induction E;simpl;firstorder;
  destruct H;firstorder.
Qed.  
  
Lemma remain_Mlt ax :
  (Proper (ax ==> iff) multi) ->
  forall e f,  multi e -> ax ⊢ e ≡ f -> mlt ∧ ax ⊢ e ≡ f.
Proof.
  intros hyp e f U E;revert U.
  pose proof (eq_multi hyp) as Unt.
  induction E;intros U;auto.
  - symmetry;apply IHE;firstorder.
  - transitivity f;[apply IHE1|apply IHE2];firstorder.
  - destruct U as (U1&U2);rewrite IHE1, IHE2;firstorder.
  - destruct U as (U1&U2);rewrite IHE1, IHE2;firstorder.
  - destruct U as (U1&U2);apply (@eq_ka_impl _ e);auto.
    apply IHE;split;auto.
    destruct H;destruct U1;firstorder.
  - apply eq_ax;split;firstorder.
Qed.

Lemma eq_nom_typed_no_perm : Proper (eq_nom_typed ==> iff) no_perm.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma eq_nom_untyped_no_perm : Proper (eq_nom_untyped ==> iff) no_perm.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma eq_nom_typed_np_no_perm :
  Proper (eq_nom_typed_np ==> iff)  no_perm.
Proof. intros e f E; destruct E;simpl;apply no_perm_apply_perm. Qed.
Lemma eq_nom_untyped_np_no_perm :
  Proper (eq_nom_untyped_np ==> iff) no_perm.
Proof. intros e f E; destruct E;simpl;apply no_perm_apply_perm. Qed.
Lemma eq_ω_smpl_no_perm : Proper (eq_ω_smpl ==> iff) no_perm.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma eq_ν_smpl_no_perm : Proper (eq_ν_smpl ==> iff) no_perm.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma eq_perm_ω_no_perm : Proper (eq_perm_ω ==> iff) no_perm.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma eq_perm_no_perm : Proper (eq_perm ==> iff) no_perm.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma zero_nom_t_no_perm : Proper (zero_nom_t ==> iff) no_perm.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma zero_nom_u_no_perm : Proper (zero_nom_u ==> iff) no_perm.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma ka_eq_t_no_perm : Proper (ka_eq_t ==> iff) no_perm.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma ka_eq_u_no_perm : Proper (ka_eq_u ==> iff) no_perm.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma ka_eq_no_perm : Proper (ka_eq ==> iff) no_perm.
Proof. intros e f E. destruct E;simpl;tauto. Qed.

Lemma eq_no_perm ax :
  (Proper (ax ==> iff) no_perm) ->
  (Proper (eq ax ==> iff) no_perm).
Proof.
  intros hyp e f E;induction E;simpl;firstorder;
  destruct H;firstorder.
Qed.
Lemma remain_Np ax :
  (Proper (ax ==> iff) no_perm) ->
  forall e f,  no_perm e -> ax ⊢ e ≡ f -> np ∧ ax ⊢ e ≡ f.
Proof.
  intros hyp e f U E;revert U.
  pose proof (eq_no_perm hyp) as Unt.
  induction E;intros U;auto.
  - symmetry;apply IHE;firstorder.
  - transitivity f;[apply IHE1|apply IHE2];firstorder.
  - destruct U as (U1&U2);rewrite IHE1, IHE2;firstorder.
  - destruct U as (U1&U2);rewrite IHE1, IHE2;firstorder.
  - simpl in U;tauto.
  - destruct U as (U1&U2);apply (@eq_ka_impl _ e);auto.
    apply IHE;split;auto.
    destruct H;destruct U1;firstorder.
  - apply eq_ax;split;firstorder.
Qed.

Lemma eq_nom_typed_str_positive : Proper (eq_nom_typed ==> iff) str_positive.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma eq_nom_untyped_str_positive : Proper (eq_nom_untyped ==> iff) str_positive.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma eq_nom_typed_np_str_positive :
  Proper (eq_nom_typed_np ==> iff)  str_positive.
Proof. intros e f E; destruct E;simpl;apply str_positive_apply_perm. Qed.
Lemma eq_nom_untyped_np_str_positive :
  Proper (eq_nom_untyped_np ==> iff) str_positive.
Proof. intros e f E; destruct E;simpl;apply str_positive_apply_perm. Qed.
Lemma eq_ω_smpl_str_positive : Proper (eq_ω_smpl ==> iff) str_positive.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma eq_ν_smpl_str_positive : Proper (eq_ν_smpl ==> iff) str_positive.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma eq_perm_ω_str_positive : Proper (eq_perm_ω ==> iff) str_positive.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma eq_perm_str_positive : Proper (eq_perm ==> iff) str_positive.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma zero_nom_t_str_positive : Proper (zero_nom_t ==> iff) str_positive.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma zero_nom_u_str_positive : Proper (zero_nom_u ==> iff) str_positive.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma ka_eq_t_str_positive : Proper (ka_eq_t ==> iff) str_positive.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma ka_eq_u_str_positive : Proper (ka_eq_u ==> iff) str_positive.
Proof. intros e f E. destruct E;simpl;tauto. Qed.
Lemma ka_eq_str_positive : Proper (ka_eq ==> iff) str_positive.
Proof. intros e f E. destruct E;simpl;tauto. Qed.

Lemma eq_str_positive ax :
  (Proper (ax ==> iff) str_positive) ->
  (Proper (eq ax ==> iff) str_positive).
Proof.
  intros hyp e f E;induction E;simpl;firstorder;
  destruct H;firstorder.
Qed.  
  
Lemma remain_Sp ax :
  (Proper (ax ==> iff) str_positive) ->
  forall e f, str_positive e -> ax ⊢ e ≡ f -> po ∧ ax ⊢ e ≡ f.
Proof.
  intros hyp e f U E;revert U.
  pose proof (eq_str_positive hyp) as Unt.
  induction E;intros U;auto.
  - symmetry;apply IHE;firstorder.
  - transitivity f;[apply IHE1|apply IHE2];firstorder.
  - destruct U as (U1&U2);rewrite IHE1, IHE2;firstorder.
  - destruct U as (U1&U2);rewrite IHE1, IHE2;firstorder.
  - destruct U as (U1&U2);apply (@eq_ka_impl _ e);auto.
    apply IHE;split;auto.
    destruct H;destruct U1;firstorder.
  - apply eq_ax;split;firstorder;
    apply str_positive_positive;firstorder.
Qed.



Hint Unfold or_rel.
Lemma zero_star_is_one_u : (zero_ka_u ∨ ka_eq_u) ⊢ 0^* ≡ 1.
Proof.
  apply inf_antisym.
  - transitivity (1·0^*+1);auto 10.
    eapply eq_ka_impl,ka_s4.
    transitivity ((1+0)+1);auto 10.
    transitivity (1+1);auto.
  - transitivity (1+0+0^* );auto 10.
    transitivity (1+0·0^*+0^* );auto 10.
Qed.
Lemma zero_star_is_one_t A : zero_ka_t ∨ ka_eq_t ⊢ (⊥ A)^* ≡ (id A).
Proof.
  apply inf_antisym.
  - transitivity ((id A)·(⊥ A)^*+(id A));auto.
    symmetry;eqr;apply eq_plus;auto 10.
    eapply eq_ka_impl,ka_s4.
    transitivity ((id A)+(⊥ A)+(id A));auto.
    * repeat apply eq_plus;auto 10.
    * transitivity ((id A)+(id A));auto 10.
  - transitivity ((id A)+(⊥ A)·(⊥ A)^*+(⊥ A)^* );auto 10.
    repeat apply eq_plus;auto.  
    transitivity ((id A)+(⊥ A));auto 10.
Qed.
