(** printing ∘ %\ensuremath{\circ}% *)
(** printing ⇒ %\ensuremath{\Rightarrow}% *)
(** printing ⇔ %\ensuremath{\Leftrightarrow}% *)
(** printing ∧ %\ensuremath{\wedge}% *)
(** printing ∨ %\ensuremath{\vee}% *)
(** printing ↔ %\ensuremath{\leftrightarrow}% *)
(** printing ^2 %\ensuremath{ {}^2}% *)
(** printing ⊗ %\ensuremath{\otimes}% *)
(** printing φ %\ensuremath{\phi}% *)
Require Export Eqdep Setoid Morphisms.
Require Export Arith List Psatz Bool.
Export ListNotations.

Hint Resolve in_eq in_cons.
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.
(** This module contains simple notations, definitions, tactics and *)
(** lemmas, independant of the rest of the developpement. *)


Notation " [ x , y ] " := (cons (@pair _ _ x y) nil) : list_scope.
Infix "∘" := Basics.compose (at level 20).
Hint Unfold Basics.compose.

(** Simple arithmetic facts. *)
Global Instance proper_plus_le : Proper (le ==> le ==> le) plus.
Proof. intros n m I k l J;lia. Qed.
Global Instance proper_S_le : Proper (le ==> le) S.
Proof. intros n m I;lia. Qed.

Ltac simpl_bool :=
  let h1 := fresh "h" in
  (match goal with
     | h: _ <> true |- _ => apply Bool.not_true_is_false in h
     | h: _ <> false |- _ => apply Bool.not_false_is_true in h
     | h: (_&&_)%bool = false |- _ =>
       apply Bool.andb_false_iff in h as [h|h]
     | h: (_&&_)%bool = true |- _ =>
       apply Bool.andb_true_iff in h as (h&h1)
     | h: (_||_)%bool = false |- _ =>
       apply Bool.andb_false_iff in h as (h&h1)
     | h: (_||_)%bool = true |- _ =>
       apply Bool.andb_true_iff in h as [h|h1]
     | h: negb _ = false |- _ =>
       apply Bool.negb_false_iff in h
     | h: negb _ = true |- _ =>
       apply Bool.negb_true_iff in h
   end)
    || rewrite Bool.andb_false_iff
    || rewrite Bool.andb_true_iff
    || rewrite Bool.orb_false_iff
    || rewrite Bool.orb_true_iff
    || rewrite Bool.negb_false_iff
    || rewrite Bool.negb_true_iff.

(** Variation of the [Bool.reflect] function, for type [bool+A] *)
(** instead of [bool].  *)
Inductive reflect3 {A : Set} (P :Prop) (Q : A -> Prop) (R : Prop) : bool + A -> Set :=
| r3_false : P -> reflect3 P Q R (inl false)
| r3_Some a : Q a -> reflect3 P Q R (inr a)
| r3_true : R -> reflect3 P Q R (inl true).

Ltac split_conj :=
  repeat
    match goal with
    | _:_ |- _ /\ _ => split
    | |- _ /\ _ => split
    end.

Ltac simpl_hyp :=
  repeat
    match goal with
    | h: _ /\ _ |- _ =>
      let h1 := fresh "h" in
      let h2 := fresh "h" in
      destruct h as (h1 & h2)
    end.

Section s.
  (** * Predicates and relations. *)
  Hypothesis T : Set.
  (** ** Relational operators. *)
  (** Top relation. *)
  Notation TTrue := (fun _ _ : T => True).
  (** Empty relation. *)
  Notation FFalse := (fun _ _ : T => False).
  (** Relational entailment. *)
  Definition stronger : relation (relation T) :=
    fun ax1 ax2 => (forall e f : T, ax1 e f -> ax2 e f).
  (** Relational union. *)
  Definition or_rel (ax1 ax2 : relation T) :=
    (fun e f => ax1 e f \/ ax2 e f).
  (** Relational intersection. *)
  Definition and_rel (p P : relation T) :=
    (fun e f : T => p e f /\ P e f).
  (** Relational equivalence. *)
  Definition eq_rel (P Q :relation T) := forall e f, P e f <-> Q e f.

  (** Notations. *)
  Infix "⇒" := stronger (at level 70).
  Infix "∨" := or_rel (at level 60).
  Infix "∧" := and_rel (at level 40).
  Infix "⇔" := eq_rel (at level 60).

  (** ** Predicates operators. *)
  (** A predicate over [T] maps every element of [T] to a proposition. *)
  Definition predicate T := T -> Prop.
  (** Predicate conjunction. *)
  Definition and_pred (A B : predicate T) (e : T) : Prop := A e /\ B e.
  (** Coercion from predicates to relations. *)
  Definition rel : predicate T -> relation T :=
    fun (A : predicate T) (e f : T) => A e /\ A f.
  (** Predicate equivalence. *)
  Definition eq_pred (P Q : predicate T) := forall e, P e <-> Q e.

  (** Notations. *)
  Infix "⊗" := and_pred (at level 60, right associativity).
  Notation " A ^2 " := (rel A) (at level 65).
  Infix "↔" := eq_pred (at level 60).

  Hint Unfold or_rel stronger and_pred rel and_rel eq_rel eq_pred.
  (** ** General properties of the operators. *)
  (** Relational entailment is a preorder. *)
  Global Instance stronger_Preorder : PreOrder stronger.
  Proof.
    split.
    - intros A e f;tauto.
    - intros A B C AB BC e f;intuition.
  Qed.
  (** Relational equivalence is an equivalence relation. *)
  Global Instance eq_rel_Equivalence : Equivalence eq_rel.
  Proof. split;intros ? ?;unfold eq_rel; firstorder.
         rewrite <- H0,<-H;auto.
         rewrite H,H0;auto.
  Qed.
  (** Predicate equivalence is an equivalence relation. *)
  Global Instance eq_pred_Equivalence : Equivalence eq_pred.
  Proof. firstorder. Qed.
  (** The relations obtained from equivalent predicates are *)
  (** equivalent. *)
  Global Instance rel_eq_pred_eq_rel :
    Proper (eq_pred ==> eq_rel) rel.
  Proof. firstorder. Qed.
  (* (** [⇒] is antisymmetric with respect to [⇔].*) *)
  (* Global Instance stronger_antisym : Antisymmetric eq_rel stronger. *)
  (* Proof. intros a b h1 h2;split;auto. Qed. *)
  (** [⊗] is a morphism for [↔]. *)
  Global Instance and_pred_eq_pred_eq_pred_eq_pred :
    Proper (eq_pred ==> eq_pred ==> eq_pred) and_pred.
  Proof. firstorder. Qed.
  (** [∧] is a morphism for [⇔]. *)
  Global Instance and_rel_eq_rel_eq_rel_eq_rel :
    Proper (eq_rel ==> eq_rel ==> eq_rel) and_rel.
  Proof. unfold eq_rel,and_rel;intros A B E C D F;firstorder. Qed.
  (** [∨] is a morphism for [⇔]. *)
  Global Instance or_rel_eq_rel_eq_rel_eq_rel :
    Proper (eq_rel ==> eq_rel ==> eq_rel) or_rel.
  Proof. unfold eq_rel,and_rel;intros A B E C D F;firstorder. Qed.
  (** [∧] is a morphism for [⇒]. *)
  Global Instance and_rel_stronger_stronger_stronger :
    Proper (stronger ==> stronger ==> stronger) and_rel.
  Proof. unfold stronger,and_rel;intros A B E C D F;firstorder. Qed.
  (** [∨] is a morphism for [⇒]. *)
  Global Instance or_rel_stronger_stronger_stronger :
    Proper (stronger ==> stronger ==> stronger) or_rel.
  Proof. unfold stronger,and_rel;intros A B E C D F;firstorder. Qed.

  (** ** Simple tautologies. *)
  Lemma FFalse_spec : forall A, FFalse ⇒ A.
  Proof. intros A e f ;firstorder. Qed.
  Lemma TTrue_spec : forall A, A ⇒ TTrue.
  Proof. intros A e f ;firstorder. Qed.
  Lemma or_rel_stronger_l A B : A ⇒ A ∨ B.
  Proof. now intros e f h;left. Qed.
  Lemma or_rel_stronger_r A B : B ⇒ A ∨ B.
  Proof. now intros e f h;right. Qed.
  Lemma assoc_or_rel1 A B C : A ∨ (B ∨ C) ⇒ (A ∨ B) ∨ C.
  Proof. intros e f;unfold or_rel;tauto. Qed.
  Lemma or_and_distr {A B C} : (A ∧ B) ∨ (A ∧ C) ⇒ A ∧ (B ∨ C).
  Proof. autounfold;intuition. Qed.
  Lemma and_or_distr {A B C} : A ∧ (B ∨ C) ⇒ (A ∧ B) ∨ (A ∧ C).
  Proof. autounfold;intuition. Qed.
  Lemma and_stronger_l {A B} : A ∧ B ⇒ B.
  Proof. autounfold;intuition. Qed.
  Lemma rel_and_pred {A B} : ((A^2) ∧ (B^2)) ⇔ ((A ⊗ B)^2).
  Proof. firstorder. Qed.
End s.
Notation TTrue := (fun _ _ => True).
Notation FFalse := (fun _ _ => False).
Infix "⇒" := stronger (at level 70).
Infix "∨" := or_rel (at level 60).
Infix "∧" := and_rel (at level 40).
Infix "⇔" := eq_rel (at level 60).
Infix "⊗" := and_pred (at level 60, right associativity).
Notation " A ^2 " := (rel A) (at level 65).
Infix "↔" := eq_pred (at level 60).
Hint Unfold or_rel stronger and_pred rel and_rel eq_rel eq_pred.

(** ** Lemmas aimed at proving that some function [φ] is a morphism. *)
Lemma proper_stronger_l {S:Set} {T} (A B:relation S) (C:relation T) φ :
  B ⇒ A -> Proper (A ==> C) φ -> Proper (B ==> C) φ.
Proof. unfold stronger,Proper,respectful;firstorder. Qed.

Lemma proper_stronger_r {S} {T:Set} (A:relation S) (B C:relation T) φ :
  B ⇒ C -> Proper (A ==> B) φ -> Proper (A ==> C) φ.
Proof. unfold stronger,Proper,respectful;firstorder. Qed.

Lemma proper_or_split {S:Set} {T} (A B:relation S) (C:relation T)  φ :
  Proper ((A ∨ B) ==> C) φ <-> Proper (A ==> C) φ /\ Proper (B ==> C) φ.
Proof. autounfold;firstorder. Qed.

Lemma proper_and_split {S} {T:Set} (A:relation S) (B C:relation T)  φ :
  Proper (A ==> (B ∧ C)) φ <-> Proper (A ==> B) φ /\ Proper (A ==> C) φ.
Proof. autounfold;firstorder. Qed.

Ltac split_proper :=
  repeat (try apply (proper_stronger_l (@and_or_distr _ _ _ _));
           rewrite proper_or_split;split).

